/*
    Posadis - A DNS Server
    Domain-manipulation functions
    Copyright (C) 2001  Meilof Veeningen <meilof@users.sourceforge.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <config.h>
#include <domainfn.h>

void bitstringcpy(char *res, int resbit, char *src, int srcbit, int nbits) {
    int x, y;

    /* filter out bits we don't want */
    res[(resbit+nbits)/8] >>= (8-((resbit%8) + nbits)) % 8;
    res[(resbit+nbits)/8] <<= (8-((resbit%8) + nbits)) % 8;

    for (x = 0; x < nbits; x++) {
        y = bitisset(src, srcbit+x);
        bitsetval(res, resbit + x, y);
    }
}


int bitstringcmp(char *val1, int val1bit, char *val2, int val2bit, int nbits) {
    int x, y, z;

    for (x = 0; x < nbits; x++) {
        y = bitisset(val1, val1bit + x);
        z = bitisset(val2, val2bit + x);

        if (y != z) return y - z;
    }

    return 0;
}

#ifdef HAVE_BITSTRING

#include "domainfn.h"
#include "configuration.h"
#include <include/unistring.h>

#define MAX_DOMAIN_RECURSION 10

void *memdup(void *src, int len) { /* i don't know why this isn't a standard c function... */
    void *res;
    res = malloc(len);
    memcpy(res, src, len);
    return res;
}

inline int domlabellen(char *label) {
    switch (domlabeltype(*label)) {
        case dlt_common:
            /* the length/type byte itself, plus the name */
            return domlabelrest(*label) + 1;
        case dlt_edns0:
            if (domlabelrest(*label) == dlt_bitstring) {
                /* the type byte, length byte, and data */
                return bitstr_len(label[1]) + 2;
            }
    }
    return -1;
}

inline int domnitems(char *label) {
    switch (domlabeltype(*label)) {
        case dlt_common:
            return 1;
        case dlt_edns0:
            if (domlabelrest(*label) == dlt_bitstring)
                return bitstr_count(label[1]);
    }
    return -1;
}

/* returns the length in chars of the domain name */

int domlen(_domain domain, int maxbufflen, char *compr, int comprlen) {
    int len = 0, x;

    if (domain == NULL) return -1;

    if (compr != NULL) return comprdomlen(domain, maxbufflen, compr, comprlen);

    while (*domain != '\0') {
        if (domlabeltype(*domain) == dlt_compr) {
            /* compressed data. we only need to include this and the the next
               byte, which is the length byte */
            return len + 2;
        }

        x = domlabellen(domain);
        if (x < 0) return x;
        len += x;
        if ((maxbufflen && len >= maxbufflen) || len > DOM_MAX) return -1;
        domain += x;
    }

    return len + 1; /* count the \0 byte as well */
}

/* returns the length in chars of the domain name: the compressed variant */
int comprdomlen(_domain domain, int maxbufflen, char *compr, int comprlen) {
    int len = 0, x, recurse = 0;
    char *cur = domain;

    while (*cur != '\0') {
        if (domlabeltype(*domain) == dlt_compr) {
            if (++recurse >= MAX_DOMAIN_RECURSION) return -1;
            x = (*cur & 63) * 256 + *(cur + 1);
            if (x >= comprlen || &compr[x] >= cur) return -1;
            cur = &compr[x];
            domain = cur;
            maxbufflen = comprlen - x;
        } else {
            x = domlabellen(domain);
            if (x < 0) return x;
            len += x;
            if (cur + len + 1 - domain >= maxbufflen || len > DOM_MAX) return -1;
            cur += x;
        }
    }

    return len + 1; /* count the \0 byte as well */
}

/* returns the number of labels of the domain name (including the empty root label) */
int domlabelcount(_domain domain) {
    int count = 0, x;

    while (*domain != '\0') {
        x = domlabellen(domain);
        if (x < 0) return x;
        count += domnitems(domain);
        domain += x;
    }
    return count + 1; /* count the \0 byte as well */
}

/* fetches a number of bits from a bitstring label, and copies them to the
   current bitstring label (doesn't write to multiple bitstring labels)

   after the op, src points to the next label that contains data */
int dom_bsfetchcopy(char *res, int respos, char*& src, int& srcpos, int nbits) {
    int x;

    while (nbits > 0) {
        x = domnitems(src);

        if ((x - srcpos) > nbits) {
            /* this item contains more than we can handle */
            bitstringcpy(res + 2, respos, src + 2, srcpos, nbits);
            srcpos += nbits;
            return 0;
        } else {
            /* do copy all items in this bsl */
            bitstringcpy(res + 2, respos, src + 2, srcpos, x - srcpos);
            nbits -= x - srcpos;
            x = domlabellen(src);
            if (x < 0) return -1;
            src += x;
            srcpos = 0;
        }
    }

    /* apparently, we're done. we only get here in the special case when the
       last bitstring label we copied contained exactly the number of bs's we
       required. coming here does _not_ mean an error. */

    return 0;
}

/* version which does not edit anything */
int dom_bsfetchcopy2(char *res, int respos, char* src, int srcpos, int nbits) {
    return dom_bsfetchcopy(res, respos, src, srcpos, nbits);
}


/* creates a bitstring label at target

    [in, out] target - pointer to res domain. upon completion, points at
                       position right after the bs label(s)
    [in]      src    - pointer to source bitstring label header
    [in]      srcpos - the bit in the bitstring label to begin with
    [in]      x      - the number of bits to copy.

   returns 0 on success, <0 on errors
*/

int dom_bscreate(char*& target, char *src, int srcpos, int x) {
    int y, ret;

    while (x > 0) {
        /* nitems to be copied to this label */
        y = x % 256; if (y == 0) y = 256;

        target[0] = dlt_edns0 | dlt_bitstring;
        target[1] = bitstr_val(y);

        ret = dom_bsfetchcopy(target, 0, src, srcpos, y);
        if (ret < 0) return -1;

        ret = domlabellen(target);
        if (ret < 0) return -1;
        target += ret;

        x -= y;
    }
    return 0;
}



/* returns the label specified with index (0 being the first, domlabelcount()-1 the last */
int domlabel(char *target, _domain domain, int index) {
    int t;
    char byte;

    while (index >= 0) {
        t = domnitems(domain);
        if (index < t) {
            /* it's in this item */
            if (domlabeltype(*domain) == dlt_common) {
                /* common label: copy it all along */
                target[0] = '@';
                memcpy(target + 1, domain + 1, *domain);
                target[*domain + 1] = '\0';
                return 0;
            } else if (domlabeltype(*domain) == dlt_edns0) {
                if (domlabelrest(*domain) == dlt_bitstring) {
                    target[0] = '?';
                    bitstringcpy(&byte, 0, domain + 2, t - index - 1, 1);
                    if (byte & 128)
                        target[1] = '1';
                    else
                        target[1] = '0';
                    target[2] = '\0';
                    return 0;
                } else
                    return -1;
            } else return -1;
        } else {
            index -= t;
            t = domlabellen(domain);
            if (t < 0) return t;
            domain += t;
        }
    }

    /* basically, we shouldn't be able to come here */
    return -1;

}

/* returns the domain name portion beginning with the label specified with index */
int domfrom(char *target, _domain domain, int index) {
    int t;

    while (index >= 0) {
        t = domnitems(domain);
        if (index < t) {
            /* it's in this item */
            if (domlabeltype(*domain) == dlt_common) {
                /* common label: copy it */
                memcpy(target, domain, *domain + 1);
                target += *domain + 1;
            } else if (domlabeltype(*domain) == dlt_edns0) {
                if (domlabelrest(*domain) == dlt_bitstring) {
                    /* bitstring label */
                    target[0] = dlt_edns0 | dlt_bitstring;
                    target[1] = bitstr_val(t - index);
                    bitstringcpy(target + 2, 0, domain + 2, 0, t - index);
                    target += domlabellen(target);
                } else return -1;
            } else return -1;
            /* copy additional items as well */
            t = domlabellen(domain);
            if (t < 0) return -1;
            if (*domain == '\0') return 0;
            domain += t;
            memcpy(target, domain, domlen(domain));
            return 0;
        } else {
            index -= t;
            t = domlabellen(domain);
            if (t < 0) return t;
            domain += t;
        }
    }

    /* we shouldn't be able to come here */
    return -1;
}


/* no boundary checking is nessecary in here, since a part of a 255max-dname
   can never get any larger */
int domto(char *target, _domain domain, int index) {
    char *dom_start = domain;
    int t;

    while (index >= 0) {
        t = domnitems(domain);
        if (index < t) {
            /* it's in this item */
            if (domlabeltype(*domain) == dlt_common ||
                index == t - 1) {
                /* luckily, we can just copy it */
                memcpy(target, dom_start, domain - dom_start);
                target[domain - dom_start] = '\0';
                return 0;
            } else if (domlabeltype(*domain) == dlt_edns0 &&
                       domlabelrest(*domain) == dlt_bitstring) {
                /* bitstring label, we need to copy only one part */

                /* at first, copy what we are sure is right */
                memcpy(target, dom_start, domain - dom_start);
                target += domain - dom_start;

                if (index) {
                    target[0] = dlt_edns0 | dlt_bitstring;
                    target[1] = index;
                           
                    bitstringcpy(target + 2, 0, domain + 2, t - index, index);
                    target += bitstr_len(index) + 2;
                }
                *target = '\0';
                return 0;
            } else return -1;
        } else {
            index -= t;
            t = domlabellen(domain);
            if (t < 0) return t;
            domain += t;
        }
    }

    /* we shouldn't be able to come here */
    return -1;
}

#define restormax(x, y) ( (x % y == 0) ? y : x % y )
#define min(x, y) ( (x < y) ? x : y )

/* cats src to target */
bool domcat(_domain target, _domain src) {
    int t, x;
    char *domain = target, *lala = NULL, *tmp;

    /* check whether we can do a simple bitstring concaternation */
    if (domlabeltype(*src) == dlt_edns0 && domlabelrest(*src) == dlt_bitstring) {
        /* find out whether the target ends with a bsl */
        while (*domain) {
            if (domlabeltype(*domain) == dlt_edns0 &&
                domlabelrest(*domain) == dlt_bitstring) {
                lala = domain;
                domain += domlabellen(domain);
            } else {
                lala = NULL;
            }
        }
        if (lala) {
            /* last label is bitstring */
            x = bitstr_count(lala[1]) + bitstr_count(src[1]);
            if (x <= 256 &&
                domlen(target) + (bitstr_len(x) - bitstr_len(lala[1])) <= DOM_MAX) {
                /* let's do it! */
                tmp = (char *)memdup(lala, domlabellen(lala));
                bitstringcpy(lala + 2, 0, src + 2, 0, bitstr_count(src[1]));
                bitstringcpy(lala + 2, bitstr_count(src[1]), tmp + 2, 0, bitstr_count(tmp[1]));
                free(tmp);
                lala[1] = bitstr_val(x);
                lala[domlabellen(lala)] = '\0';
                src += domlabellen(src);
            }
        }
    }
                
    t = domlen(target) - 1;
    x = domlen(src);
    if (t + x > DOM_MAX) return false;
    memcpy(&target[t], src, x);
    return true;
}

bool domcpy(_domain target, _domain src) {
    int x = domlen(src);
    if (x <= 0 || x > 255) return false;
    memcpy(target, src, x);
    return true;
}

/* returns true if the second domain name is child/self of the first one */
bool domisparent(_domain parent, _domain child) {
    char dom[256];
    int x, y;

    x = domlabelcount(child);
    y = domlabelcount(parent);

    if (y > x) return false;

    y = domfrom(dom, child, x - y);
    if (y < 0) return false;

    return domcmp(dom, parent);
}



/* converts a domain name to a human-readable string */
int domtostring(char *target, _domain src) {
    int x, y, val;
    char hex[2] = { 0, 0 };
    char tmp[4];
    
    if (*src == '\0') {
        strcpy(target, ".");
        return 0;
    }

    target[0] = '\0';
    
    while (*src != '\0') {
        if (domlabeltype(*src) == dlt_common) {
            /* common data */
            x = strlen(target);
            if (x + *src + 1 >= 256) return -1;
            memcpy(target + x, src + 1, *src);
            target[x + *src] = '.';
            target[x + *src + 1] = '\0';
            src += *src + 1;
        } else if (domlabeltype(*src) == dlt_edns0 && domlabelrest(*src) == dlt_bitstring) {
            val = bitstr_count(src[1]);
            if (strlen(target) + ( (val + 3) / 4 ) + 9 >= 256) {
              /* remove dot */
              target[strlen(target) - 1] = '\0';
              return 0;
            }
            strcat(target, "\\[x");
            for (x = 0; x < val; x += 4) {
              y = 0;
              if (x < val && bitisset(src + 2, x)) y += 8;
              if (x + 1 < val && bitisset(src + 2, x + 1)) y += 4;
              if (x + 2 < val && bitisset(src + 2, x + 2)) y += 2;
              if (x + 3 < val && bitisset(src + 2, x + 3)) y += 1;
              if (y > 9) hex[0] = 'a' + (y - 10); else hex[0] = '0' + y;
              strcat(target, hex);
            }
            sprintf(tmp, "%d", val);
            strcat(target, "/");
            strcat(target, tmp);
            strcat(target, "].");
            src += domlabellen(src);
        } else return -1;
    }
    return 0;
}

/* like strdup, but instead operates on a domain name (last two arguments are optional and are used for decompression) */
_domain domdup(_domain domain, int maxbufflen, char *compr, int comprlen) {
    char data[DOM_MAX], tmp[34];
    int pos = 0, reclevel = 0, x;
    char *cur = domain;

    if (compr == NULL) return (_domain) memdup(domain, domlen(domain));

    while (*cur) {
        if (domlabeltype(*cur) == dlt_compr) {
            /* compressed data */
            if (++reclevel >= MAX_DOMAIN_RECURSION) return NULL;
            x = domlabelrest(*cur) * 256 + cur[1];
            if (x >= comprlen || compr + x >= domain) return NULL;
            domain = compr + x;
            maxbufflen = comprlen - x;
            cur = domain;
        } else if (domlabeltype(*cur) == dlt_common) {
            /* common label */
            if (cur + *cur + 2 > compr + comprlen || *cur + pos + 1 > DOM_MAX) return NULL;
            data[pos++] = *cur;
            memcpy(&data[pos], cur + 1, *cur);
            pos += *cur;
            cur += *cur + 1;
        } else if (domlabeltype(*cur) == dlt_edns0 &&
                   domlabelrest(*cur) == dlt_bitstring) {
            /* bitstring label */
            x = domlabellen(cur);
            if (cur + x  + 1 > compr + comprlen) return NULL;
            memcpy(tmp, cur, x);
            tmp[x] = '\0';
            data[pos] = '\0';
            domcat(data, tmp);
            pos = domlen(data) - 1;
            cur += x;
        } else return NULL;

        if (cur >= compr + comprlen) return NULL;

    }

    if (pos >= DOM_MAX) return NULL;
    data[pos] = '\0';
    return domdup(data);
}

bool domcmp(_domain dom1, _domain dom2) {
    int x, y, dom2_bitpos = 0;

    x = domlen(dom1);
    if (x != domlen(dom2)) return false;

    while (*dom1) {
        if (domlabeltype(*dom1) != domlabeltype(*dom2)) return false;

        if (domlabeltype(*dom1) == dlt_common) {
            if (strncmpi(dom1 + 1, dom2 + 1, *dom1)) return false;
            dom2_bitpos = 0;
            dom1 += *dom1 + 1;
            dom2 += *dom2 + 1;
        } else if (domlabeltype(*dom1) == dlt_edns0 &&
                   domlabelrest(*dom1) == dlt_bitstring) {
            /* compare the complete dom1 label with parts from the second */           
            y = bitstr_count(dom1[1]);
            while (y > 0) {
                x = bitstr_count(dom2[1]);
                if (y >= x) {
                    /* we also need another dom2 bsl */
                    if (bitstringcmp(dom1 + 2, bitstr_count(dom1[1]) - y, dom2 + 2, dom2_bitpos, x)) return false;
                    dom2_bitpos = 0;
                    dom2 += domlabellen(dom2);
                    y -= x;
                } else {
                    /* we don't need another: instead, we leave the bsl */
                    if (bitstringcmp(dom1 + 2, bitstr_count(dom1[1]) - y, dom2 + 2, dom2_bitpos, y)) return false;
                    dom2_bitpos += y;
                    break;
                }
            }
            dom1 += domlabellen(dom1);
        } else return false;
    }

    return true;
}

// returns true if the non-compressed d1 is the same as the (perhaps compressed) d2
bool advdomcmp(_domain dom1, int ix, char *compr, int comprlen) {
    char *ptr1 = dom1, *ptr2 = &compr[ix];
    int pos;

    while (*ptr1) {
        while (domlabeltype(*ptr2) == dlt_compr) {
            /* dereference second one */
            pos = domlabelrest(ptr2[0]) * 256 + ptr2[1];
            if (pos > comprlen || compr + pos >= ptr2) return false;
            ptr2 = &compr[pos];
        }

        if (*ptr1 != *ptr2) return false;

        pos = domlabellen(ptr1);
        if (pos < 0 || ptr2 + pos + 1 > compr + comprlen) return false;

        if (domlabeltype(*ptr1) == dlt_common) {
            /* common label */
            if (strncmpi(ptr1 + 1, ptr2 + 1, *ptr1)) return false;
        } else if (domlabeltype(*ptr1) == dlt_edns0 &&
                   domlabelrest(*ptr1) == dlt_bitstring) {
            if (ptr1[1] != ptr2[1]) return false;
            if (memcmp(ptr1 + 2, ptr2 + 2, bitstr_len(ptr1[1]))) return false;
        } else return false;

        ptr1 += pos;
        ptr2 += pos;
    }

    if (*ptr2) return false; else return true;
}

int domlabelfromname(_domain res, char *src) {
    switch (src[0]) {
        case '@':
            /* common label */
            strcpy(res + 1, src + 1);
            res[0] = strlen(src + 1);
            break;
        case '?':
            /* bitstring label */
            res[0] = dlt_edns0 | dlt_bitstring;
            res[1] = '\1';
            if (src[1] == '1') res[2] = 128; else res[2] = 0;
            res[3] = '\0';
            break;
    }
    return 0;
}


#endif /* ifdef BITSTRING */
