/*
    Posadis - A DNS Server
    Simple domain-manipulation functions
    Copyright (C) 2001  Meilof Veeningen <meilof@users.sourceforge.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <config.h>

#ifndef HAVE_BITSTRING

#include "domainfn.h"
#include <include/unistring.h>

#define MAX_DOMAIN_RECURSION 10

void *memdup(void *src, int len) { /* i don't know why this isn't a standard c function... */
    void *res;
    res = malloc(len);
    memcpy(res, src, len);
    return res;
}

/* returns the length in chars of the domain name */

int domlen(_domain domain, int maxbufflen, char *compr, int comprlen) {
	if (domain == NULL) return -1;
    if (compr != NULL) return comprdomlen(domain, maxbufflen, compr, comprlen);
    int len = 1;
    while (*domain != '\0') {
        if (*domain > 63) return len + 1;
        len += *domain + 1;
        if (maxbufflen && len > maxbufflen) return -1;
        domain += *domain + 1;
		if (len > DOM_MAX) return -1;
    }
    return len;
}

/* returns the length in chars of the domain name: the compressed variant */
int comprdomlen(_domain domain, int maxbufflen, char *compr, int comprlen) {
    int len = 1, comprpos, recurse = 0; char *cur = domain;
    while (*cur != '\0') {
        if (*cur > 63) {
            if (*cur & 192 == 192) {
                if (++recurse >= MAX_DOMAIN_RECURSION) return -1;
                comprpos = (*cur&63)*256 + *(cur+1);
                if (comprpos >= comprlen || &compr[comprpos] >= cur) return -1;
                cur = &compr[comprpos];
                domain = cur;
                maxbufflen = comprlen - comprpos;
            } else {
                return -1;
            }
        } else {
            len += *cur + 1;
            if (cur + len + 1 - domain >= maxbufflen) return -1;
            cur += *cur + 1;
        }
    }
    if (len > DOM_MAX) return -1;
    return len;
}

/* returns the number of labels of the domain name (without the empty root label) */
int domlabelcount(_domain domain) {
    int count = 1;
    while (*domain != '\0') {
        domain += *domain + 1;
        count++;
    }
    return count;
}

/* returns the label specified with index (0 being the first, domlabelcount()-1 the last */
int domlabel(char *target, _domain domain, int index) {
    int t;
    for (t = 0; t < index; t++) domain += *domain + 1;
    memcpy(target, domain + 1, *domain);
    target[*domain] = 0;
    return *domain;
}

/* returns the domain name portion beginning with the label specified with index */
int domfrom(char *target, _domain domain, int index) {
    for (int t = 0; t < index; t++) domain += *domain + 1;
    memcpy(target, domain, domlen(domain));
    return 0;
}

int domto(char *target, _domain domain, int index) {
	char *begin = domain;
    int t;
    for (t = 0; t < index; t++) domain += *domain + 1;
    memcpy(target, begin, domain - begin);
    target[domain-begin] = '\0';
    return 0;
}

/* cats src to target */
bool domcat(_domain target, _domain src) {
    if (domlen(target) + domlen(src) > DOM_MAX + 1) return false;
    strcat(target, src);
    return true;
}

bool domcpy(_domain target, _domain src) {
    if (domlen(src) > DOM_MAX) return false;
    strcpy(target, src);
    return true;
}

/* returns true if the second domain name is a child of the first one */
bool domisparent(_domain parent, _domain child) {
    int lenp, lenc;
    lenp = domlen(parent); lenc = domlen(child);
    return (lenp<=lenc && strncmpi(parent, &child[lenc-lenp], lenp)==0);
}

/* converts a domain name to a human-readable string */
int domtostring(char *target, _domain src) {
	if (*src == '\0') {
		strcpy(target, ".");
		return 0;
	}
    while (*src != '\0') {
        memcpy(target, src + 1, *src);
        target += *src;
        *(target++) = '.';
        src += *src + 1;
    }
    *target = '\0';
    return 0;
}

/* like strdup, but then operates on a domain name (last two arguments are optional and are used for decompression */
_domain domdup(_domain domain, int maxbufflen, char *compr, int comprlen) {
    char data[DOM_MAX]; char *cur = domain; char len; int left, pos = 0, recurse = 0;
    if (compr == NULL) return strdup(domain);
    left = 0;
    while ((len=*cur) != 0) {
        if (len < 64) { /* normal label */
            if (left + len > comprlen || len + pos + 1> DOM_MAX) return NULL;
            data[pos++] = len;
            memcpy(&data[pos], ++cur, len);
            pos += len;
            if (cur + len - domain >= maxbufflen) return NULL;
            cur += len;
            continue;
        }
        if ((len & 192) == 192) { /* compressed data */
            if (++recurse >= MAX_DOMAIN_RECURSION) return NULL;
            left = (*cur&63)*256 + *(cur+1);
            if (left >= comprlen || &compr[left] >= domain) return NULL; /* an incorrect domain! */
            domain = &compr[left];
            maxbufflen = comprlen - left;
            cur = domain;
            continue;
        }
        return NULL;
    }
    data[pos] = '\0';
    return domdup(data);
}

bool domcmp(_domain dom1, _domain dom2) {
    if (strcmp(dom1, dom2) == 0) return true; else return false;
}

// returns true if the non-compressed d1 is the same as the (perhaps compressed) d2
bool advdomcmp(_domain dom1, int ix, char *compr, int comprlen) {
	char *ptr1 = dom1, *ptr2 = &compr[ix];
	int pos;
	
	while (*ptr1) {
		if (*ptr1 > 63) return false; /* error: first one also compressed */
		while (*ptr2 > 63) {
			/* dereference second one */
			if (*ptr2 & 192 == 192) {
				pos = (*ptr2 & 63) * 256 + *(ptr2 + 1);
				if (pos > comprlen) return false;
				ptr2 = &compr[ix];
			} else return false; /* we don't recognize this */
		}
		if (*ptr1 != *ptr2 || memcmp(ptr1, ptr2, *ptr1 + 1) != 0) return false;
		ptr1 += *ptr1 + 1;
		ptr2 += *ptr2 + 1;
	}
  if (*ptr2 == '\0')
    return true;
  else
    return false;
}	

int domlabelfromname(_domain res, char *src) {
    int x = strlen(src);
    res[0] = x;
    strcpy(res + 1, src);
    return 0;
}



#endif /* ifnfdef HAVE_BITSTRING */
