/*
    Posadis - A DNS Server
    Functions for reading configuration files
    Copyright (C) 2001  Meilof Veeningen <meilof@users.sourceforge.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <config.h>

#include "include/unistring.h"
#include "include/unifile.h"
#include "domainfn.h"
#include "lexfn.h"

posadis_infile::posadis_infile() {
	linenum = 1;
	file = NULL;
	name = NULL;
	errcode = ERR_NOPE;
	in_line = false;
	in_parentheses = false;
}

posadis_infile::~posadis_infile() {
	close();
	free(name);
}

bool posadis_infile::open(char *_filename) {
	char *chr;
	/* find name portion */
	chr = &_filename[strlen(_filename) - 1];
	while (*(--chr - 1) != '\\' && *(chr - 1) != '/' && chr >= _filename);	
	name = strdup(chr);
	
	file = fopen(_filename, "r");
	if (!file) return false;
	return true;
}

bool posadis_infile::close() {
	if (file) fclose(file);
    free(name); name = NULL;
	file = NULL;
	return true;
}

bool posadis_infile::is_open() {
	if (file) return true; else return false;
}

int posadis_infile::read_entry(char *buff, int bufflen) {
	int c;
	int pos = 0;
	bool in_quotes = false;
	enum {
		MODE_READING, /* just reading to buff */
		MODE_WAITNEXTCHAR, /* waiting for the next character */
		MODE_WAITNEXTSPACE, /* waiting for end of entry, without writing */
	} state = MODE_READING;

	c = fgetc(file);
	if (c == EOF) { errcode = ERR_EOF; return -1; }
	if (c == '\n' || c == '\r') { errcode = ERR_EOL; return -1; }
	if (c == '\t' || c == ' ') {
		buff[0] = '\0';
		state = MODE_WAITNEXTSPACE;
	}
	ungetc(c, file);

	in_line = true;

	while (true) {
		if (pos >= bufflen - 1) state = MODE_WAITNEXTSPACE;
		c = fgetc(file);

		switch (c) {
			case EOF: /* end-of-file */
				if (in_quotes) {
					errcode = ERR_EOF_IN_QUOTES;
					return -2;
				} else if (in_parentheses) {
					errcode = ERR_EOF_IN_PARENTHESES;
					return -2;
				}
				if (state == MODE_READING) buff[pos] = '\0';
				return 0;
			case '\n': /* end-of-line */
				linenum++;
			case '\r':
				if (!in_quotes) {
					if (state == MODE_READING) buff[pos] = '\0';
					if (in_parentheses) {
						state = MODE_WAITNEXTCHAR;
						continue;
					} else {
						in_line = false;
						if (c == '\n') linenum--;
						ungetc(c, file);
						return 0;
					}
 				}
				break;
			case '\\': /* escaped char */
				if (state == MODE_READING || state == MODE_WAITNEXTSPACE) {
					c = fgetc(file);
					if (c == EOF) {
						errcode = ERR_EOF_IN_UNESCAPE;
						return -2;
					}
					if (isdigit(c)) { /* we can expect a '\xxx' ascii */
						char numtmp[4];
						numtmp[0] = c;

						c = fgetc(file);
						if (c == EOF) {
							errcode = ERR_EOF_IN_UNESCAPE;
							return -2;
						}
						numtmp[1] = c;

						c = fgetc(file);
						if (c == EOF) {
							errcode = ERR_EOF_IN_UNESCAPE;
							return -2;
						}
						numtmp[2] = c;
						numtmp[3] = '\0';

						sscanf(numtmp, "%i", &c);
					} else {
						if (c == '.') {
			   				buff[pos++] = '\\';
   							if (pos >= bufflen - 1) {
   								state = MODE_WAITNEXTSPACE;
   								continue;
   							}
						}
					}
					break;
				} else {
					return pos;
				}
				break;
			case ' ': /* whitespace */
			case '\t':
				if (!in_quotes) {
					/* this is the end of the entry */
					buff[pos] = '\0';
					state = MODE_WAITNEXTCHAR;
					continue;
				}
				break;
			case ';': /* comments 'till end of line */
				if (!in_quotes) {
					/* go to end of line */
					while ( (c = fgetc(file)) != '\r' && c != '\n' && c != EOF);
					ungetc(c, file);
					continue;
				}
				break;
			case '(': /* begin multi-line */
				if (!in_quotes) {
					if (in_parentheses) {
						errcode = ERR_UNMATCHED_PARENTHESES;
						return -2;
					} else {
						in_parentheses = true;
					}
					if (state == MODE_READING) {
						buff[pos] = '\0';
						state = MODE_WAITNEXTCHAR;
					}
					continue;
				}
				break;
			case ')': /* end of multi-line */
				if (!in_quotes) {
					if (!in_parentheses) {
						errcode = ERR_UNMATCHED_PARENTHESES;
						return -2;
					} else {
						in_parentheses = false;
					}
					if (state == MODE_READING) {
						buff[pos] = '\0';
						state = MODE_WAITNEXTCHAR;
					}
					continue;
				}
				break;
			case '"': /* quotes */
				if (in_quotes) {
					in_quotes = false;
				} else {
					if (state == MODE_WAITNEXTCHAR) {
						ungetc(c, file);
						return pos;
					} else {
						in_quotes = true;
					}
				}
				continue;
		}
		/* all non-special chars, and all chars from case statements that were
		   not 'continue'ed, get handled in here. */
   		if (state == MODE_WAITNEXTCHAR) {
   			ungetc(c, file);
   			return pos;
   		} else {
   			if (state == MODE_READING) {
   				buff[pos++] = c;
   			}
		}
	}
}

bool posadis_infile::is_eol() {
	if (is_eof()) return true;
	
	int c = fgetc(file);
	ungetc(c, file);
	if (c == '\n' || c == '\r')
		return true;
	else
		return false;
}

bool posadis_infile::is_eof() {
	if (feof(file) != 0)
		return true;
	else
		return false;
}

bool posadis_infile::start_line() {
	int c = 0;

	if (in_line) {
		/* we're in a line, so go on until we reach a '\n' or '\r' */
		while ( !feof(file) && (c = fgetc(file)) != '\n' && c != '\r' );
		if (!feof(file)) ungetc(c, file);
	}

	/* browse through the newline sequence */
	while ( !feof(file) && (c = fgetc(file)) == '\n' || c == '\r') 
		if (c == '\n') linenum++;
					
	if (!feof(file)) ungetc(c, file);
	if (c == '\n') linenum--;
	in_line = true;
	return true;
}

char posadis_infile::preview_char() {
    int c = fgetc(file);
    ungetc(c, file);
    if (c == '\n' || c == '\r' || c == ' ' || c == '\t' || c == EOF)
        return '\0';
    else
        return (char)c;
}

char* posadis_infile::geterrmsg() {
	char msg[128];

	switch(errcode) {
	    case ERR_NOPE:
	        strcpy(msg, "<no error>");
	        break;
		case ERR_EOF_IN_QUOTES:
			strcpy(msg, "End Of File in quotes!");
			break;
		case ERR_EOF_IN_PARENTHESES:
			strcpy(msg, "End of File in parentheses!");
			break;
		case ERR_EOF_IN_UNESCAPE:
			strcpy(msg, "End of File after escape character (\\)!");
			break;
		case ERR_UNMATCHED_PARENTHESES:
			strcpy(msg, "Unmatched parenthese found!");
			break;
		case ERR_EOL:
			strcpy(msg, "Unexpected End of Line!");
			break;
		case ERR_EOF:
			strcpy(msg, "Unexpected End of File!");
			break;
		default:
			sprintf(msg, "<general formatting error %d>", errcode);
			break;
	}
	return strdup(msg);
}

int hextoint(char val) {
   	if (val >= '0' && val <= '9') {
   		return  val - '0';
   	} else if (val >= 'a' && val <= 'f') {
 		return val - 'a' + 10;
 	} else if (val >= 'A' && val <= 'F') {
 		return val - 'A' + 10;
 	} else return -1;
}


/* writes a long to a char buffer */
void write_long(char *res, int val) {
	res[0] = val / 16777216;
  res[1] = val / 65536;
  res[2] = val / 256;
  res[3] = val;
}

/* converts an email adress to a domain name
   (may be a <domain-name> or a true email address) */

int txt_to_email(char *target, char *src, char *origin) {
  char tmp[255], *cptr;
  int ret;

  if (src[0] != '@' && (cptr = strchr(src, '@')) != NULL) {
    /* contains a '@', so assume it's an email address */
    *cptr++ = '\0';

    /* do name portion */
    ret = strlen(src);
    if (ret > 63) return -1;
    target[0] = ret;
    memcpy(&target[1], src, ret);

    /* domain portion */
    ret = txt_to_dname(tmp, cptr);
    if (ret < 0 || target[0] + ret >= 255) return -1;
    memcpy(&target[target[0] + 1], tmp, ret);
    return target[0] + ret + 1;
  } else {
    return txt_to_dname(target, src, origin);
  }
}

/* converts a textual representation for a domain name to an rfc <domain-name> */
int txt_to_dname(char *target, char *src, char *origin) {
    int pos = 0;
    unsigned int len;
    char *ptr;
#ifdef HAVE_BITSTRING
    char *ptr2;
    int binpos, z;
    char curlabel[132];
#endif
    if (src[0] == '@' && src[1] == '\0') {
        /* the origin */
        if (!origin) return -1;
        len = domlen(origin);
        if (len < 0) return len;
        memcpy(target, origin, len);
        return len;
    }

    if (!*src) return -1;

    if (*src == '.' && src[1] == '\0') {
        /* the root */
        target[0] = '\0';
        return 1;
    }

    while (*src) {
#ifdef HAVE_BITSTRING
        if (src[0] == '[') {
            if ( (ptr = strchr(src, ']')) == NULL ||
                 (ptr[1] && ptr[1] != '.')) return -1;

            *ptr = '\0';
            if (strlen(src + 1) > 131) return -1;
            strcpy(curlabel, src + 1);
            *ptr++ = ']';

            if ( (ptr2 = strchr(curlabel, '/')) != NULL ) {
                /* explicit length given */
                len = atoi(ptr2 + 1);
                if (len <= 0) return -1;
                *ptr2 = '\0';
            } else len = 0;

            if (strchr(curlabel, '.')) {
                /* seems an ip address to me... */
                char tmpip[4];
                if (len == 0) len = 32;
                if (len > 32 || len <= 0 ||
                    pos + bitstr_len(len) + 3 > DOM_MAX) return -1;
                if (txt_to_ip(tmpip, curlabel) < 0) return -1;
                target[pos] = dlt_edns0 | dlt_bitstring;
                target[pos + 1] = bitstr_val(len);
                bitstringcpy(target + pos + 2, 0, tmpip, 0, len);
            } else if (strchr(curlabel, ':')) {
                /* seems an ipv6 address to me... */
                char tmpip[16];
                if (len == 0) len = 128;
                if (len > 128 || len <= 0 ||
                    pos + bitstr_len(len) + 3 > DOM_MAX) return -1;
                if (txt_to_ipv6(tmpip, curlabel) < 0) return -1;
                target[pos] = dlt_edns0 | dlt_bitstring;
                target[pos + 1] = bitstr_val(len);
                bitstringcpy(target + pos + 2, 0, tmpip, 0, len);
            } else if (curlabel[0] == 'b') {
                /* binary bitstring label */
                if (len) {
                    if (len != strlen(curlabel + 1)) return -1;
                } else
                    len = strlen(curlabel + 1);
                
                if (len > 256 || len <= 0 ||
                    pos + bitstr_len(len) + 3 > DOM_MAX) return -1;

                target[pos] = dlt_edns0 | dlt_bitstring;
                target[pos + 1] = bitstr_val(len);
                memset(target + pos + 2, 0, bitstr_len(len));

                ptr2 = curlabel + 1;
                binpos = 0;
                while (*ptr2) {
                    if (*ptr2 == '1')
                        bitset(target + pos + 2, binpos);
                    else if (*ptr2 != '0') return -1;
                    binpos++; ptr2++;
                }
            } else if (curlabel[0] == 'o') {
                /* octal bitstring label */
                if (len) {
                    if ( (len - 1) / 3 != (strlen(curlabel + 1) - 1) ) return -1;
                } else
                    len = strlen(curlabel + 1) * 3;
                
                if (len > 256 || len <= 0 ||
                    pos + bitstr_len(len) + 3 > DOM_MAX) return -1;
                
                target[pos] = dlt_edns0 | dlt_bitstring;
                target[pos + 1] = bitstr_val(len);
                memset(target + pos + 2, 0, bitstr_len(len));

                ptr2 = curlabel + 1;
                binpos = 0;
                while (*ptr2) {
                    if (*ptr2 >= '0' && *ptr2 <= '7') {
                        z = *ptr2 - '0';
                        bitsetval(target + pos + 2, binpos, z & 4); binpos++;
                        bitsetval(target + pos + 2, binpos, z & 2); binpos++;
                        bitsetval(target + pos + 2, binpos, z & 1); binpos++;
                    } else return -1;
                    ptr2++;
                }
            } else if (curlabel[0] == 'x') {
                /* hexadecimal label */

                if (len) {
                    if ( (len - 1) / 4 != (strlen(curlabel + 1) - 1) ) return -1;
                } else
                    len = strlen(curlabel + 1) * 4;

                if (len > 256 || len <= 0 ||
                    pos + bitstr_len(len) + 3 > DOM_MAX) return -1;

                target[pos] = dlt_edns0 | dlt_bitstring;
                target[pos + 1] = bitstr_val(len);
                memset(target + pos + 2, 0, bitstr_len(len));

                ptr2 = curlabel + 1;
                binpos = 0;
                while (*ptr2) {
                    z = hextoint(*ptr2);
                    if (z < 0) return -1;
                    bitsetval(target + pos + 2, binpos, z & 8); binpos++;
                    bitsetval(target + pos + 2, binpos, z & 4); binpos++;
                    bitsetval(target + pos + 2, binpos, z & 2); binpos++;
                    bitsetval(target + pos + 2, binpos, z & 1); binpos++;
                    ptr2++;
                }
            } else return -1;

            pos += bitstr_len(len) + 2;
            if (*ptr != '.') ptr = NULL;
        } else {
#endif
            ptr = strchr(src, '.');
            if (ptr) *ptr = '\0';

            /* common label */
            len = strlen(src);

            if (len > 63) return -1;
            if (pos + len + 2 > 255) return -1;
            target[pos] = len;
            memcpy(target + pos + 1, src, len);
            pos += len + 1;
#ifdef HAVE_BITSTRING
        }
#endif
        if (ptr) {
            *ptr = '.';
            src = ptr + 1;
        } else break;
    }

    target[pos] = '\0';

    if (src[strlen(src)-1] != '.') {
        /* relative domain name */
        if (origin)
            if (!domcat(target, origin)) return -1;
    }

    return domlen(target);
}




/* converts a text from buffer to human-readable text */
/* note: wouldn't this potentially cause buffer overrun?! */
int txt_to_text(char *target, char *src) {
    int pos = 0;
    while(*src != '\0') {
        if (*src == '\xff') src++;
        *target = *src;
        target++; src++; pos++;
    }
    *target = '\0';
    return pos;
}
/* converts a buffer to a long numeric value, with postfix (e.g. 68k->68000) support */
long txt_to_long(char *buff) {
    long val = 0, tmpval = 0;
    bool neg = false;
    if (*buff == '-') { neg = true; buff++; }
    while (1) {
        if (*buff >= '0' && *buff <= '9') {
            tmpval *= 10; tmpval += *buff - '0';
        } else {
            if (*buff == '\0') {
              val += tmpval;
              return neg ? -val : val;
            }
            if (*buff == 'K') tmpval *= 1024;
            else if (*buff == 'M') tmpval *= 1048576;
            else if (*buff == 'G') tmpval *= 1073741824;
            else if (*buff == 's');
            else if (*buff == 'm') tmpval *= 60;
            else if (*buff == 'h') tmpval *= 3600;
            else if (*buff == 'd') tmpval *= 86400;
            else if (*buff == 'w') tmpval *= 604800;
            else if (*buff == 'y') tmpval *= 31536000;
            else return -2;

            val += tmpval;
            tmpval = 0;
        }
        buff++;
    }
}
/* converts a buffer to an 32-bit ip number */
int txt_to_ip(char ip[4], char *buff) {
    int p = 0, tmp = 0, node = 0;
    while (buff[p] != '\0') {
        if (isdigit(buff[p])) {
            node *= 10;
            node += buff[p] - '0';
            if (node > 255) return -1;
        } else {
            if (buff[p] == '.') {
                if (tmp >= 3) return -1;
                ip[tmp++] = (char)node;
                node = 0;
            } else {
                return -1;
            }
        }
        p++;
    }
    if (tmp == 3)
        ip[tmp] = (char)node;
    else
        return -1;
    return 0;
}

/* converts a buffer to a 128-bit ipv6 number */
int txt_to_ipv6(char ipv6[16], char *buff) {
	int multigroup_pos = -1;
	int pos = -1;
	int node = 0; /* the pair we're working on */
	int nodeval = -1; /* node value */
	int nodestart = 0; /* start pos of node */
	int chval;
	int x;

	while (buff[++pos] != '\0') {
		if (buff[pos] == '.') {
			/* an imbedded ipv4 */
			if (node + 2 > 8) return -1;
			x = txt_to_ip(&ipv6[node*2], &buff[nodestart]);
			if (x < 0) return x;
			node++;
			if (node == 8) break;
			nodeval = -1;
			nodestart = pos;
			break;
		}
		if (buff[pos] == ':') {
			/* write the previous node */
			if (pos) {
				if (nodeval == -1) return -1;
				ipv6[node*2] = nodeval / 256;
				ipv6[node*2 + 1] = nodeval;
			} else {
				if (buff[pos + 1] != ':') return -1;
				node--;
			}
   			if (buff[pos + 1] == ':') {
   				/* we're having a multigroup indicator here */
   				multigroup_pos = node + 1;
   				pos++;
   			}
			node++;
			if (node > 7) return -1;
			nodeval = -1;
			nodestart = pos + 1;
  		} else {
            chval = hextoint(buff[pos]);
            if (chval < 0) return -1;
			if (nodeval == -1)
				nodeval = chval;
			else
				nodeval = (nodeval * 16) + chval;
			if (nodeval > 65535) return -1;
		}
	}

	if (nodeval != -1) {
		ipv6[node*2] = nodeval / 256;
		ipv6[node*2 + 1] = nodeval;
	} else {
		if (buff[pos - 1] == ':' && buff[pos - 2] != ':') return -1;
	}

	if (multigroup_pos != -1) {
		for (x = 15; x >= 14 - 2 * (node - multigroup_pos); x--)
			ipv6[x] = ipv6[x - 2 * (7 - node)];
		memset(&ipv6[multigroup_pos * 2], 0, 14 - (2*node));
	} else {
		if (node < 7) return -1;
	}
	return 0;
}
