/*
    Posadis - A DNS Server
    System socket functions for platform independence
    Copyright (C) 2001  Meilof Veeningen <meilof@users.sourceforge.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "syssocket.h"
#include "dns.h"
#include "include/unistring.h"
#include "config.h"
#include <errno.h>

#ifdef __APPLE__
#define socklen_t int
#endif

#ifdef _WIN32
#define ioctl ioctlsocket
#endif

int struct_pf(_addr *addr) {
    if (addr->s_family == AF_INET) return PF_INET;
#ifdef HAVE_IPV6
    else  if (addr->s_family == AF_INET6) return PF_INET6;
#endif
    return -1;
}

int struct_len(_addr *addr) {
    if (addr->s_family == AF_INET) return sizeof(sockaddr_in);
#ifdef HAVE_IPV6
    else if (addr->s_family == AF_INET6) return sizeof(sockaddr_in6);
#endif
    return -1;
}

int initsocklib() {
#ifdef _WIN32
    WSADATA info;
#  ifdef HAVE_IPV6
    return WSAStartup(MAKEWORD(2, 2), &info);
#  else
    return WSAStartup(MAKEWORD(1, 1), &info);
#  endif
#else /* ifdef _WIN32 */
    return 0;
#endif
}

int closesocklib() {
#ifdef _WIN32
    return WSACleanup();
#else /* ifdef _WIN32 */
    return 0;
#endif
}

#ifdef _WIN32 /* win32 seems to require this... */
#define _Saddr S_un.S_addr
#else
#define _Saddr s_addr
#endif

int tcpcreateserver(_addr* socketaddr) {
    int sockid; 
    int one = 1;

    if ((sockid = socket(struct_pf(socketaddr), SOCK_STREAM, IPPROTO_TCP)) < 0) return -1;
    if (bind(sockid, (sockaddr *)socketaddr, struct_len(socketaddr)) < 0) return -1;
    setsockopt(sockid, SOL_SOCKET, SO_REUSEADDR, (char *)&one, sizeof(one));
    if (listen(sockid, 0) < 0) return -1;

    return sockid;
}

int udpcreateserver(_addr* socketaddr) {
    int sockid;

    if ((sockid = socket(struct_pf(socketaddr), SOCK_DGRAM, IPPROTO_UDP)) < 0) return -1;
    if (bind(sockid, (sockaddr *)(socketaddr), struct_len(socketaddr)) < 0) return -1;

    return sockid;
}

int tcpopen(_addr *addr) {
	int sockid;

	if ((sockid = socket(struct_pf(addr), SOCK_STREAM, IPPROTO_TCP)) < 0) return -1;
	if (connect(sockid, (sockaddr *)addr, struct_len(addr)) < 0) return -1;
	return sockid;
}

int tcpaccept(int socket, _addr *askaddr) {
    _addr querier;
    socklen_t len = sizeof(_addr);
    int ret = accept(socket, (sockaddr *)&querier, &len);
    if (ret >= 0)
        if (askaddr) memcpy(askaddr, &querier, len);

    return ret;
}

int udpread(int socket, char *data, _addr *addr) {
    socklen_t addr_size = sizeof(_addr);
    return recvfrom(socket, data, UDP_MSG_SIZE, 0, (sockaddr *) addr, &addr_size);
}

int udpsend(int socket, _addr *addr, char *msg, int msglen) {
    return sendto(socket, msg, msglen, 0, (sockaddr *)addr, struct_len(addr));
}

int tcpsend(int socket, char *buff, int bufflen) {
    return send(socket, buff, bufflen, 0);
}

int tcpread(int socket, char *buff, int bufflen) {
    return recv(socket, buff, bufflen, 0);
}

int tcpclose(int socket) {
    return close(socket);
}

int udpclose(int socket) {
    return close(socket);
}

/* functions for implementations of watchsets */
fd_set cset_in, cset_out, cset_err, set_in, set_out, set_err;
bool watchset_error = false; /* whether the latest call was erroneous */

int watchset_init() {
    FD_ZERO(&set_in);
    FD_ZERO(&set_out);
    FD_ZERO(&set_err);
    return 0;
}

int watchset_add(int sockid) {
    FD_SET(sockid, &set_in);
    FD_SET(sockid, &set_out);
    FD_SET(sockid, &set_err);
    return 0;
}

int watchset_remove(int sockid) {
    FD_CLR((unsigned)sockid, &set_in);
    FD_CLR((unsigned)sockid, &set_out);
    FD_CLR((unsigned)sockid, &set_err);
    return 0;
}

int watchset_wait(int maxtime) {
    struct timeval tv;
    int ret;
    tv.tv_sec = maxtime; tv.tv_usec = 0;
    memcpy(&cset_in, &set_in, sizeof(fd_set));
    memcpy(&cset_out, &set_out, sizeof(fd_set));
    memcpy(&cset_err, &set_err, sizeof(fd_set));
    ret = select(FD_SETSIZE - 1, &cset_in, NULL, &cset_err, &tv);
    tv.tv_sec = 0; tv.tv_usec = 0;
    select(FD_SETSIZE - 1, NULL, &cset_out, NULL, &tv);
    if (ret < 0) {
        watchset_error = true;
        return 0;
    } else {
        watchset_error = false;
        return ret;
    }
}

int watchset_waitread(int maxtime) {
    struct timeval tv;
    int ret;
    tv.tv_sec = maxtime; tv.tv_usec = 0;
    memcpy(&cset_in, &set_in, sizeof(fd_set));
    memcpy(&cset_err, &set_err, sizeof(fd_set));
    ret = select(FD_SETSIZE - 1, &cset_in, NULL, &cset_err, &tv);
    tv.tv_sec = 0; tv.tv_usec = 0;
    select(FD_SETSIZE - 1, NULL, &cset_out, NULL, &tv);
    if (ret < 0) {
        watchset_error = true;
        return 0;
    } else {
        watchset_error = false;
        return ret;
    }
}	

int watchset_waitforever() {
    int ret;
    memcpy(&cset_in, &set_in, sizeof(fd_set));
    memcpy(&cset_out, &set_out, sizeof(fd_set));
    memcpy(&cset_err, &set_err, sizeof(fd_set));
    ret = select(FD_SETSIZE - 1, &cset_in, &cset_out, &cset_err, NULL);
    if (ret < 0) {
        watchset_error = true;
        return 0;
    } else {
        watchset_error = false;
        return ret;
    }
}

int watchset_check() {
    return watchset_wait(0);
}

bool watchset_isdata(int sockid) {
    return !watchset_error && FD_ISSET(sockid, &cset_in);
}

bool watchset_iserror(int sockid) {
    return !watchset_error && FD_ISSET(sockid, &cset_err);
}

bool watchset_canwrite(int sockid) {
	return !watchset_error && FD_ISSET(sockid, &cset_out);
}

/* implementation of well-known services */
int getprotocolbyname(char *name) {
    struct protoent *protocol;
    if ( (protocol = getprotobyname(name)) == NULL)
        return -1;
    else
        return protocol->p_proto;
}

int getserviceportbyname(char *name) {
    struct servent *service;
    if ( (service = getservbyname(name, NULL)) == NULL)
        return -1;
    else
        return ntohs(service->s_port);
}

/* address functions */

bool tcp_is_open(int sockid) {
   char buff[1];
   if (watchset_isdata(sockid)) {
     if (recv(sockid, buff, 1, MSG_PEEK) <= 0) return false; else return true;
   } else {
     return true;
   }
}

int gettcpaddr(int sockid, _addr *add) {
    socklen_t len = sizeof(sockaddr_in);
    return (getpeername(sockid, (sockaddr*)add, &len));
}

int getaddress(_addr *res, char *ip, int port) {
#ifdef HAVE_IPV6
    char *ptr = strchr(ip, ':');

    if (ptr) {
        /* ipv6 */
        memset(res, 0, sizeof(sockaddr_in6));
#ifdef HAVE_SIN6_LEN
        ((sockaddr_in6 *)res)->sin6_len = sizeof(sockaddr_in6);
#endif
        ((sockaddr_in6 *)res)->sin6_family = AF_INET6;
        ((sockaddr_in6 *)res)->sin6_port = htons(port);
        if (txt_to_ipv6((char *)&((sockaddr_in6 *)res)->sin6_addr, ip) != 0) return -1;
        return 0;
    }
#endif
    /* ipv4 */
    memset(res, 0, sizeof(sockaddr_in));
#ifdef HAVE_SIN_LEN
    ((sockaddr_in *)res)->sin_len = sizeof(sockaddr_in);
#endif
    ((sockaddr_in *)res)->sin_family = AF_INET;
    ((sockaddr_in *)res)->sin_port = htons(port);
    if (txt_to_ip((char *)&((sockaddr_in *)res)->sin_addr, ip) != 0) return -1;
    return 0;
}

int getaddress_ip4(_addr *res, char *ipv4_data, int port) {
    memset(res, 0, sizeof(_addr));
#ifdef HAVE_SIN_LEN
    ((sockaddr_in *)res)->sin_len = sizeof(_addr);
#endif
    ((sockaddr_in *)res)->sin_family = AF_INET;
    ((sockaddr_in *)res)->sin_port = htons(port);
    memcpy(&((sockaddr_in *)res)->sin_addr.s_addr, ipv4_data, 4);
    return 0;
}

#ifdef HAVE_IPV6

int getaddress_ip6(_addr *res, char *ipv6_data, int port = DNS_PORT) {
    memset(res, 0, sizeof(_addr));
#ifdef HAVE_SIN6_LEN
    ((sockaddr_in6 *)res)->sin6_len = sizeof(_addr);
#endif
    ((sockaddr_in6 *)res)->sin6_family = AF_INET;
    ((sockaddr_in6 *)res)->sin6_port = htons(port);
    memcpy(&((sockaddr_in6 *)res)->sin6_addr, ipv6_data, 16);
    return 0;
}
#endif



bool is_ipv4(_addr *addr) {
    return (addr->s_family == AF_INET);
}

#ifdef HAVE_IPV6
bool is_ipv6(_addr *addr) {
    return (addr->s_family == AF_INET6);
}
#endif


char *addr_to_string(_addr *addr) {
    char *caddr, msg[64];

    if (addr->s_family == AF_INET) {
        /* IPv4 */
        caddr = (char *)&((sockaddr_in *)addr)->sin_addr;
        sprintf(msg, "%d.%d.%d.%d#%d", caddr[0], caddr[1], caddr[2], caddr[3],
                ntohs(((sockaddr_in *)addr)->sin_port) & 32767);
        return strdup(msg);
    }
#ifdef HAVE_IPV6
    if (addr->s_family == AF_INET6) {
        /* IPv6 */
        caddr = (char *)&((sockaddr_in6 *)addr)->sin6_addr;
        sprintf(msg, "%x%x:%x%x:%x%x:%x%x:%x%x:%x%x:%x%x:%x%x#%d",
                caddr[0], caddr[1], caddr[2], caddr[3], caddr[4], caddr[5], caddr[6],
                caddr[7], caddr[8], caddr[9], caddr[10], caddr[11], caddr[12],
                caddr[13], caddr[14], caddr[15],
                ntohs(((sockaddr_in6 *)addr)->sin6_port) & 32767);
        return strdup(msg);
     }
#endif

     sprintf(msg, "<unknown socket family %d>", addr->s_family);
     return strdup(msg);
}

bool compare_hostport(_addr *addr1, _addr *addr2) {
    if (compare_host(addr1, addr2)) {
        if (addr1->s_family == AF_INET)
            return ((sockaddr_in *)addr1)->sin_port == ((sockaddr_in *)addr2)->sin_port;
#ifdef HAVE_IPV6
        else if (addr1->s_family == AF_INET6)
            return ((sockaddr_in6 *)addr1)->sin6_port == ((sockaddr_in6 *)addr2)->sin6_port;
#endif
    }
    return false;
}

bool compare_host(_addr *addr1, _addr *addr2) {
    if (addr1->s_family != addr2->s_family) return false;

    if (addr1->s_family == AF_INET)
        return (memcmp(&((sockaddr_in *)addr1)->sin_addr, &((sockaddr_in *)addr2)->sin_addr, 4) == 0);
#ifdef HAVE_IPV6
    else if (addr1->s_family == AF_INET6)
        return (memcmp(&((sockaddr_in6 *)addr1)->sin6_addr, &((sockaddr_in6 *)addr2)->sin6_addr, sizeof(in6_addr)) == 0);
#endif

    return false;
}

int answeraddress(_addr *res, _addr *src, int port) {
    memcpy(res, src, sizeof(_addr));

    if (port > 0) {
        if (src->s_family == AF_INET)
            ((sockaddr_in *)res)->sin_port = ntohs(port);
#ifdef HAVE_IPV6
        else if (src->s_family == AF_INET6)
            ((sockaddr_in6 *)res)->sin6_port = ntohs(port);
#endif
        else return -1;
    }

    return 0;
}


bool address_lookup(_addr *res, char *name, int port) {
    struct hostent *ent;

    if (strchr(name, ':')) {
        /* try ipv6 */
        if (getaddress(res, name, port) == 0) return true;
    }

    ent = gethostbyname(name);
    if (!ent) return false;

    memset(res, 0, sizeof(_addr));
#ifdef HAVE_SIN_LEN
    ((sockaddr_in *)res)->sin_len = sizeof(_addr);
#endif
    ((sockaddr_in *)res)->sin_family = ent->h_addrtype;
    ((sockaddr_in *)res)->sin_port = htons(port);
    memcpy(&((sockaddr_in *)res)->sin_addr, ent->h_addr, ent->h_length);
    return true;
}
