/*
    Posadis - A DNS Server
    Authoritive Zones & Handling
    Copyright (C) 2001  Meilof Veeningen <meilof@users.sourceforge.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <include/unistring.h>

#include "authoritive.h"
#include "cachelookup.h"
#include "classes.h"
#include "log.h"
#include "masterfile.h"
#include "posadis.h"
#include "configuration.h"

/* zone groups we implement */

AuthGroup::AuthGroup() { type = ZONE_AUTH; }
PrimaryGroup::PrimaryGroup() { type = ZONE_PRIMARY; }
SecondaryGroup::SecondaryGroup() { type = ZONE_SECONDARY; }

/* the Authoritive Zone */

AuthoritiveZone::AuthoritiveZone(_domain _dom) : Zone(_dom) {
    nottr = NULL;
}

AuthoritiveZone::~AuthoritiveZone() {
}

void AuthoritiveZone::start_notification() {
    int x, c;
    Vector <_addr *> *notlist = new Vector<_addr *>();

    if (nottr) nottr->stop = true;

    if (auth_config.notify_to_slaves) {
        c = multi_get_count(auth_config.slaves);
        for (x = 0; x < c; x++)
            addaddressbyident(notlist, multi_getptr(auth_config.slaves, x));

    }
    c = multi_get_count(auth_config.also_notify);
    for (x = 0; x < c; x++)
        addaddressbyident(notlist, multi_getptr(auth_config.also_notify, x));

    if (notlist->count == 0) {
        delete notlist;
    } else {
        nottr = new NotifyServerThread(this, notlist);
        threads.add(nottr);
    }
}

_thread_answer AuthoritiveZone::do_lookup_thread(AnsweringThread*& thr, stdquery_info *info, _domain QNAME) {
    /* authoritively answers the query */
    DnsMessage *q = info->query, *answer = info->answer;

    CacheItem *item;
    rr_set *set;
    int ret, x;
    int QTYPE;
    char tmp[256], tmp2[256];
    complete_rr *rr;
	
    if (answer->answer.count == 0) answer->AA = true;

    /* determine what we're trying to answer */
    QTYPE = q->question[0]->QTYPE;

    ret = RelativeLookup(item, this, QNAME, lookup_flag_cname);
    if (!(ret & LOOKUP_FOUND)) {
        /* NXDOMAIN: at first, we should check for a DNAME */
        set = item->get_rrset(DNS_TYPE_DNAME);
        if (set && set->items.count) {
            domto(tmp, QNAME, ret);
            if (domcat(tmp, set->items[0]->RDATA)) {
                /* yo: we'de dealing with a fine DNAME: restart the query */
                rr = new complete_rr();
                rr->NAME = domdup(QNAME);
                rr->TYPE = DNS_TYPE_CNAME;
                rr->TTL = set->ttl;
                rr->RDLENGTH = domlen(tmp);
                rr->RDATA = domdup(tmp);
                answer->answer.add(rr);
                ret = domfrom(tmp2, QNAME, ret);
                if (ret < 0) {
                    answer->RCODE = RCODE_SERVFAIL;
                    return TA_SUCCESS;
                } else {
                    if (!config.hide_dname)
                        answer->add_rrset_qtype(tmp2, item, DNS_TYPE_DNAME, true, S_ANSWER);
                    return global_do_lookup_thread(thr, info, domdup(tmp));
                }
            }
        }
        /* sorry, we did everything we could */
        answer->RCODE = RCODE_NXDOMAIN;
        /* add the SOA record to the authority section*/
        answer->add_rrset_qtype(domain, citem_root, DNS_TYPE_SOA, true, S_AUTHORITY);
        return TA_SUCCESS;
    }

    /* if we got here, the domain name was found */
    set = item->get_rrset(DNS_TYPE_CNAME);
    if (QTYPE != DNS_TYPE_CNAME && set && set->items.count) {
        /* we need to follow the CNAME chain */
        answer->add_rrset_qtype(QNAME, item, DNS_TYPE_CNAME, true, S_ANSWER);
        if (reached_max_cname_level(answer)) {
            answer->RCODE = RCODE_SERVERFAILURE;
        } else {
            return global_do_lookup_thread(thr, info, domdup(set->items[0]->RDATA));
        }
    } else {
        /* the domain ain't no cname, so if the requested rrset exists
           we'll add it */
        x = answer->answer.count;
        answer->add_rrset_qtype(QNAME, item, QTYPE, true, S_ANSWER);
        if (x == answer->answer.count)
            /* apparently, we're dealing with a NODATA condition here */
            answer->add_rrset_qtype(domain, citem_root, DNS_TYPE_SOA, true, S_AUTHORITY);
        else
            /* we did find some data, so we add NS records */          
            answer->add_rrset_qtype(domain, citem_root, DNS_TYPE_NS, true, S_AUTHORITY);
    }
    return TA_SUCCESS;

}


/* after getting data, we should pass through glue to the proper authorities */
char cd_msg_tmp[256];

/* return number of subzones we verified */
int get_rid_of_glue(Zone *zn, CacheItem *citem, _domain dom) {
    int x, y = 0;
    rr_set *set;
    char tmp[256];
    Zone *dzone;

    if (citem != zn->citem_root) {
        set = citem->get_rrset(DNS_TYPE_NS);
        if (set) {
            /* delegation here */
            LookupZone(dzone, dom);
            if (dzone == zn) {
                /* we don't have that subzone */
                domtostring(cd_msg_tmp, dom);
                domtostring(tmp, zn->domain);
                log_print(LOG_LEVEL_ERROR, "Received subzone %s for zone %s we don't know about!", cd_msg_tmp, tmp);
            } else {
                /* allright, we know that zone */
                dzone->offer_glue(citem);
                y = 1;
            }
            return y;
        }
    }

    for (x = citem->subitems.count - 1; x >= 0 ; x--) {
        domlabelfromname(tmp, citem->subitems[x]->name);
        domcat(tmp, dom);
        y += get_rid_of_glue(zn, citem->subitems[x], tmp);
    }

    return y;
}


/* PrimaryZone: take information from a local master file */
PrimaryZone::PrimaryZone(_domain _dom) : AuthoritiveZone(_dom) { type = ZONE_PRIMARY; }
PrimaryZone::~PrimaryZone() { }

void PrimaryZone::start_updating() {
    int ret;
    char temp[256];

#ifdef _EXTDEBUG_
	domtostring(temp, domain);
	log_print(LOG_LEVEL_INFO, "Starting PRIMARY update process for %s", temp);
#endif

    next_update = curtime + primary_config.zone_ttl;

    /* at first, remove the zone if it exists */
    if (!is_being_updated) { /* prevent a zone from being updated by multiple threads */
    	is_being_updated = true;
    	zone_is_loaded = false;
    	delete citem_root;
    	citem_root = new CacheItem("");

        if (primary_config.file == NULL) {
            domtostring(temp, domain);
            log_print(LOG_LEVEL_ERROR, "No file given for the %s zone. The zone has been removed. Retrying in %d seconds...", temp, primary_config.zone_ttl);
	        is_being_updated = false;
			return;
        } else {
            /* now load the zone from the master file */
            ret = load_master_file(this, domain, primary_config.file);
            if (ret < 0) { /* we have a problem here... */
                log_print(LOG_LEVEL_ERROR, "Fatal error while reading %s. The zone has been removed. Retrying in %d seconds...", primary_config.file, primary_config.zone_ttl);
		    	delete citem_root;
    			citem_root = new CacheItem("");
		        is_being_updated = false;
                return;
            }
            ret = get_rid_of_glue(this, citem_root, domain);
            start_notification();

            if (ret != subzones.count) {
                domtostring(cd_msg_tmp, domain);
                log_print(LOG_LEVEL_ERROR, "Primary %s: %d subzones were not found in master file!", cd_msg_tmp, subzones.count - ret);
            }
            zone_is_loaded = true;
        }
        is_being_updated = false;
    }
}

/* secondary zone: get info from xfr */

SecondaryZone::SecondaryZone(_domain _dom) : AuthoritiveZone(_dom) {
	type = ZONE_SECONDARY;
	last_loaded = 0;
	soa_serial = 0;
	soa_retry = 0;
}

SecondaryZone::~SecondaryZone() { }

void SecondaryZone::start_updating() {
    int x, y;
    char tmpb[256];

#ifdef _EXTDEBUG_
	domtostring(tmpb, domain);
	log_print(LOG_LEVEL_INFO, "Starting SECONDARY update process for %s", tmpb);
#endif	
    
    next_update = curtime + secondary_config.initial_retry_interval; /* default value */

    /* this function is called when the refresh period has expired
       if this action is enqueued, a previous thread will likely have loaded
       a new zone version, making the refresh process unneccesary. */

    if (masters.count == 0) {
        /* intialize ns list */
        y = multi_get_count(secondary_config.masters);

        for (x = 0; x < y; x++)
            addaddressbyident(&masters, multi_getptr(secondary_config.masters, x));

        if (masters.count == 0) {
            domtostring(tmpb, domain);
            log_print(LOG_LEVEL_ERROR, "No valid masters defined for %s secondary!", tmpb);
        }
    }

    /* check whether we may call ourselves authoritive */
   	if (!last_loaded || last_loaded + soa_expire < curtime)
   		zone_is_loaded = false;
   	else
   		zone_is_loaded = true;

    /* calculate next update time */
    if (soa_retry)
        next_update = curtime + soa_retry;
    else
        next_update = curtime + secondary_config.initial_retry_interval;

    /* we have some nameservers, so start an update */
    update_start(new XfrClientThread(this));

}

void SecondaryZone::update_start(MainThread *tr) {
    pending_updates.add(tr);
    if (pending_updates.count == 1) {
        /* only process: might as well start now */
        threads.add(tr);
#ifdef _DEBUG_
        log_print(LOG_LEVEL_INFO, "Starting directly...");
#endif
    }
#ifdef _DEBUG_
    else log_print(LOG_LEVEL_INFO, "Enqueuing process...");
#endif
}

void SecondaryZone::update_ready() {
     /* switch over to next thread */
     if (pending_updates.count > 1) {
#ifdef _DEBUG_
         log_print(LOG_LEVEL_INFO, "Starting process...");
#endif
         /* start next thread */
         threads.add(pending_updates[0]);
     }
}

#define absint(x) (x < 0 ? -x : x)

/* rfc1982 serial arithmetic */
inline int soa_firstsmaller(unsigned int i1, unsigned int i2) {
	return (i1 != i2 &&
		   ((i1 < i2 && i2 - i1 < 2147483647) ||
	        (i1 > i2 && i1 - i2 > 2147483647)));
}

/* xfr client thread: get xfr info */

XfrClientThread::XfrClientThread(SecondaryZone *zn, bool _is_forced, Vector<_addr *> *_nameservers) :
	MainThread() {
	
	q = new DnsMessage(); 	

    if (_nameservers == NULL)
 	  nameservers = &zn->masters;
    else
      nameservers = _nameservers;

 	zone = zn; 	
 	child = NULL;
 	current = -1;

    is_forced = _is_forced;
    state = XFR_START;
}

XfrClientThread::~XfrClientThread() {
	if (nameservers != &zone->masters) delete nameservers;
	delete q;
	if (child) delete child;
    zone->update_ready();
    zone->pending_updates.removeitem(this, false);
}

int XfrClientThread::execute() {
	int child_ret;
	char dname[256];
	question_item *quest;
	int x;
	char *ptr;
	
	txt_to_dname(dname, zone->domain);

    if (state == XFR_START) {
        if (!is_forced && nameservers == &zone->masters &&
            zone->zone_is_loaded && curtime < (zone->last_loaded + zone->soa_refresh)) {
            /* there's no need to update anymore */
            return TRET_CLOSED;
        }
        if (zone->zone_is_loaded)
            state = XFR_SOA_INIT; // at first, we'll see if something is updated
        else
            state = XFR_AXFR_INIT; // we'll need the complete zone
    }


	if (child && !(child_timeout && child_timeout <= curtime)) {
		child_ret = child->execute();
		if (child_ret == TRET_OKBUSY) return TRET_OKBUSY;
	} else {
		child_ret = TRET_ERROR;
	}
	
	child_timeout = 0;
	
	if (state == XFR_SOA_INIT) {
		/* initialize a new SOA request to see if comething changed */
		if (++current >= nameservers->count) {
			/* we've had all of our servers, and none was able to answer! */
			log_print(LOG_LEVEL_WARNING, "Failed to reach nameservers for update of %s zone!", dname);
			return TRET_ERROR;
		}
		
		/* send a soa request */
		q->reinit();
		q->ID = new_request_id();
		quest = new question_item();
		quest->QNAME = domdup(zone->domain);
		quest->QTYPE = DNS_TYPE_SOA;
		q->question.add(quest);
		
		conn = initialize_tcp_connection(*(*nameservers)[current]);
		if (!conn) {
      char *tmpstr = addr_to_string((*nameservers)[current]);
      log_print(LOG_LEVEL_ERROR, "Failed to connect for SOA query to master %s", tmpstr);
      free(tmpstr);
      return TRET_ERROR;
    }
		child = new MessageOutThreadTCP(conn, q);
		twait = new ThreadWait(this, q->ID, *(*nameservers)[current], conn->sockid);
		answer_waiting_threads.add(twait);
		
		state = XFR_SOA_SEND;
		return TRET_OKBUSY;
	}
	
	if (state == XFR_SOA_SEND) {
		/* either ready sending, or error */
		delete child; child = NULL;
		
		if (child_ret == TRET_ERROR) {
			/* sending failed: switch to next */
      char *tmpstr = addr_to_string((*nameservers)[current]);
      log_print(LOG_LEVEL_ERROR, "Failed to send SOA query from master %s", tmpstr);
      free(tmpstr);
			answer_waiting_threads.removeitem(twait);
			state = XFR_SOA_INIT;
			return TRET_OKBUSY;
		}
		/* now wait for answer */
		wakeup_time = curtime + config.patience + 1;
		trstate = THREAD_SLEEP;
		state = XFR_SOA_RECV;
		return TRET_OKBUSY;
	}
	
	if (state == XFR_SOA_RECV) {
		if (twait->received) {
			/* received soa answer */
			if (twait->msg->RCODE != RCODE_NOERROR ||
				!twait->msg->answer.count ||
				twait->msg->answer[0]->TYPE != DNS_TYPE_SOA) {
				/* oops, we don't like the answer: switch to next */
				state = XFR_SOA_INIT;
				answer_waiting_threads.removeitem(twait);
				return TRET_OKBUSY;
			}
			
			ptr = twait->msg->answer[0]->RDATA;
			ptr += domlen(ptr); /* primary master */
			ptr += domlen(ptr); /* mailbox of responsible person */
			x = ptr[0]  * 16777216 + ptr[1] * 65536 + ptr[2] * 256 + ptr[3];
			
			if (soa_firstsmaller(zone->soa_serial, x)) {
				/* new version */
				current--;
				state = XFR_IXFR_INIT;
				answer_waiting_threads.removeitem(twait);
				return TRET_OKBUSY;
			} else {
				/* no new version: we're done */
				zone->last_loaded = curtime;
				zone->next_update = curtime + zone->soa_refresh;
				zone->zone_is_loaded = true;
				zone->is_being_updated = false;
				answer_waiting_threads.removeitem(twait);
                zone->start_notification();
				return TRET_CLOSED;
			}
		} else {
			/* no answer from this server: try next */
			state = XFR_SOA_INIT;
      char *tmpstr = addr_to_string((*nameservers)[current]);
      log_print(LOG_LEVEL_ERROR, "Failed to receive SOA answer from master %s", tmpstr);
      free(tmpstr);
			answer_waiting_threads.removeitem(twait);
      return TRET_OKBUSY;
		}
	}
	
	if (state == XFR_IXFR_INIT) {
		/* initialize a IXFR request to the current nameserver */
		if (++current >= nameservers->count) {
			/* no nameserver would respond! */
			log_print(LOG_LEVEL_WARNING, "Couldn't contact zone transfer server for %s", dname);
			return TRET_ERROR;
		}
				
		q->reinit();
		quest = new question_item();
		quest->QNAME = domdup(zone->domain);
		quest->QTYPE = QTYPE_IXFR;
		q->question.add(quest);
		q->add_rrset(zone->domain, zone->citem_root->get_rrset(DNS_TYPE_SOA),
		             false, S_AUTHORITY);
		
		conn = initialize_tcp_connection(*(*nameservers)[current]);
		if (!conn) {
      char *tmpstr = addr_to_string((*nameservers)[current]);
      log_print(LOG_LEVEL_ERROR, "Failed to connect for IXFR query to master %s", tmpstr);
      free(tmpstr);
      return TRET_ERROR;
    }
		child = new MessageOutThreadTCP(conn, q);
		twait = new ThreadWait(this, q->ID, *(*nameservers)[current], conn->sockid);
		answer_waiting_threads.add(twait);
		
		state = XFR_IXFR_SEND;
		return TRET_OKBUSY;
	}
	
	if (state == XFR_IXFR_SEND) {
		/* either ready sending, or error */
		delete child; child = NULL;
		
		if (child_ret == TRET_ERROR) {
			/* sending failed, so quit */
      char *tmpstr = addr_to_string((*nameservers)[current]);
      log_print(LOG_LEVEL_ERROR, "Failed to send IXFR query to master %s", tmpstr);
      free(tmpstr);
			state = XFR_SOA_INIT;
			answer_waiting_threads.removeitem(twait);
            return TRET_OKBUSY;
		}
		/* now wait for answer */
		wakeup_time = curtime + config.patience + 1;
		trstate = THREAD_SLEEP;
		state = XFR_IXFR_RECV;
		return TRET_OKBUSY;
	}
	
	if (state == XFR_IXFR_RECV) {
		if (twait->received) {
			/* received ixfr answer */
			if (twait->msg->RCODE == RCODE_NOTIMPLEMENTED) {
				/* ixfr not supported: try axfr */
				current--;
				state = XFR_AXFR_INIT;
				return TRET_OKBUSY;
			}
			x = interpret_xfr_message();
			if (x < 0)
				return TRET_ERROR;
			else
				return TRET_CLOSED;
		} else {
			/* no answer from this server */
      char *tmpstr = addr_to_string((*nameservers)[current]);
      log_print(LOG_LEVEL_ERROR, "Failed to receive IXFR answer from master %s", tmpstr);
      free(tmpstr);
			state = XFR_SOA_INIT;
			answer_waiting_threads.removeitem(twait);
            return TRET_OKBUSY;
		}
	}
	
	if (state == XFR_AXFR_INIT) {
		/* initialize a IXFR request to the current nameserver */
		if (++current >= nameservers->count) {
			/* no nameserver would respond! */
			log_print(LOG_LEVEL_WARNING, "Couldn't contact zone transfer server for %s", dname);
			return TRET_ERROR;
		}
				
		/* send a soa request */
		q->reinit();
		q->ID = new_request_id();
		quest = new question_item();
		quest->QNAME = domdup(zone->domain);
		quest->QTYPE = QTYPE_AXFR;
		q->question.add(quest);
		
		conn = initialize_tcp_connection(*(*nameservers)[current]);
		if (!conn) {
      char *tmpstr = addr_to_string((*nameservers)[current]);
      log_print(LOG_LEVEL_ERROR, "Failed to connect for AXFR query to master %s", tmpstr);
      free(tmpstr);
      return TRET_ERROR;
    }
		child = new MessageOutThreadTCP(conn, q);
		twait = new ThreadWait(this, q->ID, *(*nameservers)[current], conn->sockid);
		answer_waiting_threads.add(twait);
		
		state = XFR_AXFR_SEND;
		return TRET_OKBUSY;
	}
	
	if (state == XFR_AXFR_SEND) {
		/* either ready sending, or error */
		delete child; child = NULL;
		
		if (child_ret == TRET_ERROR) {
			/* sending failed, so quit */
      char *tmpstr = addr_to_string((*nameservers)[current]);
      log_print(LOG_LEVEL_ERROR, "Failed to send AXFR query to master %s", tmpstr);
      free(tmpstr);
			state = XFR_SOA_INIT;
			answer_waiting_threads.removeitem(twait);
            return TRET_OKBUSY;
		}
		/* now wait for answer */
		wakeup_time = curtime + config.patience + 1;
		trstate = THREAD_SLEEP;
		state = XFR_AXFR_RECV;
		return TRET_OKBUSY;
	}
	
	if (state == XFR_AXFR_RECV) {
		if (twait->received) {
			/* received axfr answer */
			if (twait->msg->RCODE != RCODE_NOERROR) {
				state = XFR_SOA_INIT;
				answer_waiting_threads.removeitem(twait);
            		return TRET_OKBUSY;
			}
							
			x = interpret_xfr_message();
			if (x < 0)
				return TRET_ERROR;
			else
				return TRET_CLOSED;
		} else {
			/* no answer from this server */
      char *tmpstr = addr_to_string((*nameservers)[current]);
      log_print(LOG_LEVEL_ERROR, "Failed to receive AXFR answer from master %s", tmpstr);
      free(tmpstr);
			state = XFR_SOA_INIT;
			answer_waiting_threads.removeitem(twait);
	            return TRET_OKBUSY;
		}
	}
	
	return TRET_ERROR;
}

int XfrClientThread::interpret_xfr_message() {
	DnsMessage *ans = twait->msg;
	CacheItem *citem_old = NULL, *citem;
	char *ptr;
	int x, serial;
				
   	if (!ans->is_answer_on(q)) {
   		/* not answering our query... */
   		goto xfr_err;
   	}
   	
   	if (ans->RCODE != RCODE_NOERROR ||
   		ans->authority.count || ans->additional.count ||
   		ans->answer.count < 2 ||
   		ans->answer[0]->TYPE != DNS_TYPE_SOA ||
   		ans->answer[ans->answer.count - 1]->TYPE != DNS_TYPE_SOA) {
   		/* we don't like it! */
   		goto xfr_err;
   	}

   	citem_old = zone->citem_root->clone(); /* in case it fails */
				
   	/* check whether the version is actually newer */
   	ptr = ans->answer[0]->RDATA;
   	ptr += domlen(ptr); /* primary master */
   	ptr += domlen(ptr); /* mailbox of responsible person */
   	serial = ptr[0] * 16777216 + ptr[1] * 65536 + ptr[2] * 256 + ptr[3];
				
   	if (zone->zone_is_loaded && !soa_firstsmaller(zone->soa_serial, serial)) goto xfr_err;
   	
    if (ans->answer[1]->TYPE != DNS_TYPE_SOA) {
   		/* we have an AXFR answer */
   		delete zone->citem_root;
   		zone->citem_root = new CacheItem("");
   	} else {
        /* just delete soa record */
        zone->citem_root->remove_rrset(DNS_TYPE_SOA);
    }

    /* add soa record */	
   	if (!zone->citem_root->add_msg_rr(ans->answer[0], true)) goto xfr_err;
				
   	x = 0; /* current answer item */
   	
   	while (++x < ans->answer.count - 1) {
   		if (ans->answer[x]->TYPE == DNS_TYPE_SOA) {
   			/* list of RRs we need to delete */
   			while (ans->answer[++x]->TYPE != DNS_TYPE_SOA) {
   				if (RelativeLookup(citem, zone, ans->answer[x]->NAME) != LOOKUP_FOUND)
   					goto xfr_err;
   				if (!citem->remove_msg_rr(ans->answer[x]))
   					goto xfr_err;
   			}
   		} else {
   			/* list of RRs we need to add */
   			if (CreateDomain(citem, zone, ans->answer[x]->NAME) < 0)
   				goto xfr_err;
						
   			if (!citem->add_msg_rr(ans->answer[x], true)) goto xfr_err;
   		}
   	}

    /* check for newly-crafted delegation list */
    x = get_rid_of_glue(zone, zone->citem_root, zone->domain);
    if (x != zone->subzones.count) {
        domtostring(cd_msg_tmp, zone->domain);
        log_print(LOG_LEVEL_ERROR, "Secondary %s: %d subzones were not passed through in XFR!", cd_msg_tmp, zone->subzones.count - x);
    }

				
   	/* apparently, the zone transfer succeeded */	
				
   	ptr = ans->answer[0]->RDATA;
   	ptr += domlen(ptr); /* primary master */
   	ptr += domlen(ptr); /* mailbox of responsible person */
   	zone->soa_serial = ptr[0]  * 16777216 + ptr[1] * 65536 + ptr[2] * 256 + ptr[3];
   	zone->soa_refresh = ptr[4]  * 16777216 + ptr[5] * 65536 + ptr[6] * 256 + ptr[7];
   	zone->soa_retry = ptr[8]  * 16777216 + ptr[9] * 65536 + ptr[10] * 256 + ptr[11];
   	zone->soa_expire = ptr[12]  * 16777216 + ptr[13] * 65536 + ptr[14] * 256 + ptr[15];
   	zone->soa_minimum = ptr[16]  * 16777216 + ptr[17] * 65536 + ptr[18] * 256 + ptr[19];
   	x = curtime + zone->soa_refresh;
   	if (zone->next_update > x) zone->next_update = x;
   	zone->last_loaded = curtime;
   	zone->zone_is_loaded = true;
   	zone->is_being_updated = false;
				
   	answer_waiting_threads.removeitem(twait);
				
   	delete citem_old;
   	return 0;				
				

xfr_err:
   	delete zone->citem_root;
   	zone->citem_root = citem_old;
   	
   	return -1;
}

/* notify client */

_thread_answer SecondaryZone::do_notify_thread(AnsweringThread*& thr, DnsMessage *q, _addr *requestor, DnsMessage *answer, MainThread *tr) {
    int x, t, z, g_found = 0, found = 0;
    Vector<_addr *> *vec;
    CacheItem *item;
    rr_set *set;
    _addr addr;
    answeraddress(&addr, requestor, DNS_PORT);

    /* check whether the sender is one of our listed nameservers */
    for (x = 0; x < masters.count; x++) {
        if (compare_host(requestor, masters[x])) {
            /* let's see on the unsecure hint */
            if (q->OPCODE == OPCODE_NOTIFY) {
                g_found = 1;
                for (t = 0; t < q->answer.count; t++) {
                    found = 0;
                    if (domisparent(domain, q->answer[x]->NAME) &&
                        RelativeLookup(item, this, q->answer[x]->NAME) & LOOKUP_FOUND &&
                        (set = item->get_rrset(q->answer[x]->TYPE)))
                        /* this rr set should contain the data */
                        for (z = 0; z < set->items.count; z++)
                            if (set->items[x]->RDLENGTH == q->answer[x]->RDLENGTH &&
                                memcmp(set->items[x]->RDATA, q->answer[x]->RDATA, q->answer[x]->RDLENGTH))
                                /* yo! */
                                found = 1;
                    if (found == 0) {
                        g_found = 0;
                        break;
                    }
                }
            } else g_found = 0;
            if (g_found == 0) {
                /* data differs: zone transfer is nessecary */
                vec = new Vector<_addr *>();
                vec->add((_addr *)memdup(&addr, sizeof(_addr)));
                update_start(new XfrClientThread(this, true, vec));
            }
            return TA_SUCCESS;
        }
    }

    answer->RCODE = RCODE_REFUSED;
    return TA_SUCCESS;
}

/* notify server */

NotifyServerThread::NotifyServerThread(AuthoritiveZone *_zn, Vector<_addr *> *_recpts) {
    current = -1;
    attempt = 0;
    zn = _zn;
    recpt = _recpts;
    state = NOT_INIT;
    child = NULL;
    q = new DnsMessage();
    stop = 0;
}

NotifyServerThread::~NotifyServerThread() {
    delete recpt;
    if (child) delete child;
    delete q;
    zn->nottr = NULL;
}

int NotifyServerThread::execute() {
    int ret = 0;

    if (child) {
        ret = child->execute();
        if (ret == TRET_OKBUSY) return TRET_OKBUSY;
    }

    if (state == NOT_INIT) {
        /* notify a new one */
        if (++current >= recpt->count) return TRET_CLOSED;
        state = NOT_START;
        attempt = 0;
    }

    if (state == NOT_START) {
        if (stop) return TRET_CLOSED;
        if (++attempt >= zn->auth_config.notify_retry) {
            state = NOT_INIT;
            return TRET_OKBUSY;
        }

        if (!zn->zone_is_loaded || !zn->citem_root->get_rrset(DNS_TYPE_SOA))
            return TRET_ERROR;
        q->reinit();
        q->AA = true;
        q->ID = new_request_id();
        q->OPCODE = OPCODE_NOTIFY;
        q->question.add(new question_item(zn->domain, DNS_TYPE_SOA));
        q->add_rrset(zn->domain, zn->citem_root->get_rrset(DNS_TYPE_SOA));

        child = new MessageOutThreadUDP(-1, *(*recpt)[current], q);
        twait = new ThreadWait(this, q->ID, *(*recpt)[current], ((MessageOutThread*)child)->sockid);
        answer_waiting_threads.add(twait);

        state = NOT_SEND;
        return TRET_OKBUSY;
    }

    if (state == NOT_SEND) {
        delete child; child = NULL;
        if (ret == TRET_ERROR) {
            /* go to next */
            state = NOT_START;
        } else {
            wakeup_time = curtime + config.patience +
                          zn->auth_config.notify_interval / zn->auth_config.notify_retry;
            trstate = THREAD_SLEEP;
            state = NOT_RECV;
        }
        return TRET_OKBUSY;
    }

    if (state == NOT_RECV) {
        trstate = THREAD_RUNNING;
        if (!twait->received)
            /* go to next */
            state = NOT_START;
        else
            state = NOT_INIT;
        /* basically, we don't care about the answer: it only shows the server is active */
        answer_waiting_threads.removeitem(twait);
        return TRET_OKBUSY;
    }

    return TRET_ERROR;
}

/* xfr server */

_thread_answer AuthoritiveZone::do_axfr_thread(AnsweringThread*& thr, xfr_info *info) {
	int ret;
  if (!zone_is_loaded) {
    info->answer->RCODE = RCODE_SERVFAIL;
    return TA_SUCCESS;
  }
	ret = exec_axfr(info, this);
	if (ret >= 0) return TA_SUCCESS;
	return TA_ERROR;
}

_thread_answer AuthoritiveZone::do_ixfr_thread(AnsweringThread*& thr, xfr_info *info) {
	int ret;
  if (!zone_is_loaded) {
    info->answer->RCODE = RCODE_SERVFAIL;
    return TA_SUCCESS;
  }
	ret = exec_ixfr(info, this);
	if (ret >= 0) return TA_SUCCESS;
	return TA_ERROR;
}

int add_axfr_rrs(DnsMessage *msg, CacheItem *item, _domain dom) {
	char dname[255];
    int t, ret;

    /* add rrs */
    for (t = 0; t < item->rr_sets.count; t++)
		if (item->rr_sets[t]->type != DNS_TYPE_SOA)
	    	msg->add_rrset(dom, item->rr_sets[t], false);
    	
    /* add subitems */
    for (t = 0; t < item->subitems.count; t++) {
	    if (domlen(dom) + strlen(item->subitems[t]->name) > 255 ||
	    	strlen(item->subitems[t]->name) > 64) return -1;

        domlabelfromname(dname, item->subitems[t]->name);
        domcat(dname, dom);
	
        ret = add_axfr_rrs(msg, item->subitems[t], dname);
        if (ret != 0) return ret;
	}
	
	return 0;
}

int exec_axfr(xfr_info *info, AuthoritiveZone *zn) {
	int ret;
    DnsMessage *q = info->query, *a = info->answer;
	
    a->AA = true;

    /* the first step we take is to see if it's the zone top */
    if (!domcmp(zn->domain, q->question[0]->QNAME)) {
        a->RCODE = RCODE_NXDOMAIN;
        return 0;
    }

    /* next, check whether the querier is permitted to do xfr (TODO) */
    if (zn->auth_config.xfr_to != XFR_TO_ANY &&
        !(zn->auth_config.xfr_to == XFR_TO_SLAVES &&
         is_in_identlist(zn->auth_config.slaves, info->querier)) &&
        !is_in_identlist(zn->auth_config.also_xfr_to, info->querier)) {
        a->RCODE = RCODE_REFUSED;
        return 0;
    }

    /* now add the RRs */
    /* begin with the SOA */
    if (a->add_rrset_qtype(q->question[0]->QNAME, zn->citem_root,
    	DNS_TYPE_SOA, false) < 0) {
        a->RCODE = RCODE_SERVERFAILURE;
        return 0;
    }

    /* and then recursively add other RR's */
    ret = add_axfr_rrs(a, zn->citem_root, q->question[0]->QNAME);
    if (ret < 0) return ret;
    if (ret > 0) {
        a->RCODE = RCODE_SERVERFAILURE;
        return 0;
    }

    /* end with the SOA */
    if (a->add_rrset_qtype(q->question[0]->QNAME, zn->citem_root,
    	DNS_TYPE_SOA, false) < 0) {
        a->RCODE = RCODE_SERVERFAILURE;
        return 0;
    }

    return 0;
}


int exec_ixfr(xfr_info *info, AuthoritiveZone *zn) {
	return exec_axfr(info, zn);
}
