/*
    Posadis - A DNS Server
    Authoritive Zones & Handling
    Copyright (C) 2001  Meilof Veeningen <meilof@users.sourceforge.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef __AUTHORITIVE_H
#define __AUTHORITIVE_H

#include "zones.h"

/* the authoritive groups */
class AuthGroup : public ZoneGroup { public: AuthGroup(); auth_configuration_data auth_config; };
class PrimaryGroup : public AuthGroup { public: PrimaryGroup(); primary_configuration_data primary_config; };
class SecondaryGroup : public AuthGroup { public: SecondaryGroup(); secondary_configuration_data secondary_config; };

class NotifyServerThread;

/*!
  The AuthoritiveZone class is a direct descendant of the Zone class, and it
  contains some default behavior for zones that always return authoritive
  answers. These 'authoritive zones' are descendants of the AuthoritiveZone
  class (e.g. PrimasterZone, SlaveZone).

  The AuthoritiveZone class currently doesn't override any of the Zone's
  class behavior.
*/

class AuthoritiveZone : public Zone {
  protected:
	AuthoritiveZone(_domain);                           // constructor to be called by descendants
    virtual ~AuthoritiveZone();


    //! not an official Zone function; will be called by do_*recursive_thread
    _thread_answer do_stdquery_thread(AnsweringThread*& thr, DnsMessage *q, DnsMessage *a, MainThread *tr);

    virtual _thread_answer do_lookup_thread(AnsweringThread*& thr, stdquery_info *info, _domain QNAME);

    virtual _thread_answer do_axfr_thread(AnsweringThread*& thr, xfr_info *info);
    virtual _thread_answer do_ixfr_thread(AnsweringThread*& thr, xfr_info *info);
 public:
    auth_configuration_data auth_config;				// some settings
    void start_notification();
    NotifyServerThread *nottr;

};

/*!
  The PrimaryZone is the zone class for a primary. A primary for a domain
  reads it information from local DNS master files, and it is usually
  willing to answer XFR (zone transfer) requests.

  This class subclasses AuthoritiveZone, the base class for all zones which
  always answer authoritively.
*/

class PrimaryZone : public AuthoritiveZone {
  public:
    PrimaryZone(_domain domain);                       // public constructor
    virtual ~PrimaryZone();

    virtual void start_updating();                      // overrided to load local dns master files
    primary_configuration_data primary_config;          // some settings
};

typedef Vector<_addr *> serverlist;

/*!
  The Secondary is the zone class for a slave. A slave retrieves its
  information using XFR. It might also offer XFR itself, in this case, it is
  called a 'slave master' or 'secondary master'.

  This class subclasses AuthoritiveZone, the base class for all zones which
  always answer authoritively.
*/

class SecondaryZone : public AuthoritiveZone {
  public:
    SecondaryZone(_domain domain);                      // public contructor
    virtual ~SecondaryZone();

    /* updating functions */
    virtual _thread_answer do_notify_thread(AnsweringThread*& thr, DnsMessage *q, _addr *requestor, DnsMessage *a, MainThread *tr);

    secondary_configuration_data secondary_config;      // some settings
    long last_loaded;									// the last time the zone was succesfully loaded
    int soa_serial, soa_refresh, soa_retry, soa_expire, soa_minimum; // various SOA fields  	
    serverlist masters;						            // ns list

    virtual void start_updating();                      // overrided to load from xfr
    virtual void update_ready();                        // if an update is ready, it calls this function
    virtual void update_start(MainThread *tr);          // starts or enqueues an update op
    Vector<MainThread *> pending_updates;               // pending updates
};  	

/*!
  The XfrClientThread tries to initialize a zone by sending XFR requests to a
  predefined list of nameservers.
*/

class XfrClientThread : public MainThread {
  public:
  	XfrClientThread(SecondaryZone *zn, bool forced = false, Vector<_addr *> *list = NULL); // constructor
  	virtual ~XfrClientThread();							// destructor

  	virtual int execute();								// overrided

    bool is_forced;                                     // whether we should still update
                                                        // with an empty server list and an up-to-date zone  	
	DnsMessage *q;										// current message
  	SecondaryZone *zone;								// our zone
  	int current;										// current nameserver index
    Vector <_addr *> *nameservers;                      // pointer to our masters
  	
  	enum { XFR_START,
  	       XFR_SOA_INIT, XFR_SOA_SEND, XFR_SOA_RECV,
  	       XFR_AXFR_INIT, XFR_AXFR_SEND, XFR_AXFR_RECV,
  	       XFR_IXFR_INIT, XFR_IXFR_SEND, XFR_IXFR_RECV
  	     } state; // our state
  	
  	ThreadWait *twait;									// wait data
  	Thread *child;										// child thread
  	int child_timeout;									// end time for child (or 0)
  	TcpConnection *conn;
  	
  	int interpret_xfr_message();						// handles the xfr answer
};

/* notify sending */

class NotifyServerThread : public MainThread {
  public:
    NotifyServerThread(AuthoritiveZone *zn, Vector<_addr *>*_recpts);
    virtual ~NotifyServerThread();

    virtual int execute();
    int stop;
  protected:
    AuthoritiveZone *zn;
    Vector <_addr *> *recpt;
    int current, attempt;
    enum { NOT_INIT, NOT_START, NOT_SEND, NOT_RECV } state;
    Thread *child;
    DnsMessage *q;
    ThreadWait *twait;
};

/* functions for serving xfr requests */

int exec_axfr(xfr_info *info, AuthoritiveZone *zn);
int exec_ixfr(xfr_info *info, AuthoritiveZone *zn);

#endif /* __AUTHORITIVE_H */
