/*
    Posadis - A DNS Server
    Cache Lookup Functions
    Copyright (C) 2001  Meilof Veeningen <meilof@users.sourceforge.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "dns.h"
#include "include/unistring.h"

/* default implementations for zones. this can be overrided, for example, to
   enable disk caching */

int Zone::create_domain(CacheItem*& res, _domain dom) {
	char node[DOMLABEL_RMAX];
	int nlabels = domlabelcount(dom) - domlabelcount(domain);
	int curlabel, x;
	bool found;
	CacheItem *newitem;

	res = citem_root;
	if (!domisparent(domain, dom)) return LOOKUP_NOTFOUND;
	if (res == NULL) return LOOKUP_NOTFOUND;
		
	/* try 'n find each label in the domain name */
	
	for (curlabel = nlabels - 1; curlabel >= 0; curlabel--) {
		domlabel(node, dom, curlabel);
		
		found = false;
		for (x = 0; x < res->subitems.count; x++) {
			if (strcmpi(res->subitems[x]->name, node) == 0) {
				/* we found the label */
				found = true;
				res = res->subitems[x];
				
				/* we now have the complete domain name */
				if (curlabel == 0) return LOOKUP_FOUND;
			}
		}
		
		if (!found) {
			/* the domain name does not appear to exist */
			newitem = new CacheItem(node);
			res->subitems.add(newitem);
			res = newitem;
			if (curlabel == 0) return LOOKUP_FOUND | LOOKUP_CREATED;
		} else
             if (res->get_rrset(DNS_TYPE_DNAME)) return LOOKUP_NOTFOUND;
	}

	return LOOKUP_FOUND;
}


int Zone::do_internal_lookup(CacheItem*& res, _domain dom, int flags) {
	CacheItem *catchall = NULL;
	char node[DOMLABEL_RMAX];
    char catchall_label[DOMLABEL_RMAX] = "";
	int nlabels = domlabelcount(dom) - domlabelcount(domain);
	int curlabel, x;
	bool found;
	
	res = citem_root;
	if (!domisparent(domain, dom)) return LOOKUP_NOTFOUND;
	if (res == NULL) return LOOKUP_NOTFOUND;

    /* get the catchall label */
    domlabel(catchall_label, "\1*", 0);
			
	/* try 'n find each label in the domain name */
	
	for (curlabel = nlabels - 1; curlabel >= 0; curlabel--) {
		domlabel(node, dom, curlabel);
		
		found = false;
		for (x = 0; x < res->subitems.count; x++) {
			if (strcmpi(res->subitems[x]->name, node) == 0) {
				/* we found the label */
				found = true;
				res = res->subitems[x];
				
				/* we now have the complete domain name */
				if (curlabel == 0) return LOOKUP_FOUND;
			} else if (strcmp(res->subitems[x]->name, catchall_label) == 0)
				/* catch-all */
				catchall = res->subitems[x];
		}
		
		/* the domain name does not appear to exist */
		if (!found) {
			if (flags && lookup_flag_cname && catchall) {
                res = catchall;
                return LOOKUP_FOUND;
            } else return curlabel + 1 + LOOKUP_NOTFOUND;
		}
	}
	return LOOKUP_FOUND;
}


int LookupZone(Zone*& res, _domain dom) {
	int t;
	bool found;
	
	if (rootzone == NULL) return LOOKUP_NOTFOUND;
	res = rootzone;
	if (dom[0] == '\0') return LOOKUP_FOUND;
	
	while (res) {
		found = false;
		for (t = 0; t < res->subzones.count; t++) {
			if (domisparent(res->subzones[t]->domain, dom)) {
				res = res->subzones[t];
				if (domlen(res->domain) == domlen(dom)) {
					/* we've found the zone! */
					return LOOKUP_FOUND;
				} else {
					found = true;
					break;
				}
			}
		}	
		if (!found) return LOOKUP_NOTFOUND;
	}
	return LOOKUP_NOTFOUND;
}

int RelativeLookup(CacheItem*& res, Zone *zone, _domain dom, int flags) {
	return zone->do_internal_lookup(res, dom, flags);
}

int CreateDomain(CacheItem*& res, Zone *zone, _domain dom) {
	return zone->create_domain(res, dom);
}

int LookupDomain(CacheItem*& resitem, Zone*& reszone, _domain dom, int flags) {
	LookupZone(reszone, dom);
 	return RelativeLookup(resitem, reszone, dom, flags);
}

/* this functions, which are not meant merely for internal purposes and not
   for the dns resolution process, also follow CNAME, * and DNAME records */

int RelativeLookupEx(CacheItem*& res, Zone *zone, _domain dom, int reclevel) {
     int ret;
     rr_set *dname_rr;
     char ourdom[256];
     Zone *znres;

     if (reclevel >= config.max_cname) return LOOKUP_NOTFOUND;
     ret = RelativeLookup(res, zone, dom);

     if (! (ret & LOOKUP_FOUND)) {
         /* might have a wildcard or DNAME */
         dname_rr = res->get_rrset(DNS_TYPE_DNAME);
         if (dname_rr && dname_rr->items.count) {
             domto(ourdom, dom, ret);
             if (domcat(ourdom, dname_rr->items[0]->RDATA)) {
                 return LookupDomainEx(res, znres, ourdom, reclevel + 1);
             }
         }
         /* sorry, we've done everything we can */
     } else {
         /* might be a CNAME */
         dname_rr = res->get_rrset(DNS_TYPE_CNAME);
         if (dname_rr && dname_rr->items.count)
             return LookupDomainEx(res, znres, dname_rr->items[0]->RDATA, reclevel + 1);
     }

     return (int)ret;
}

int LookupDomainEx(CacheItem*& resitem, Zone*& reszone, _domain dom, int reclevel) {
     if (reclevel >= config.max_cname) return LOOKUP_NOTFOUND;
     LookupZone(reszone, dom);
     return RelativeLookupEx(resitem, reszone, dom);
}
