/*
    Posadis - A DNS Server
    DNS foundation classes
    Copyright (C) 2001  Meilof Veeningen <meilof@users.sourceforge.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "include/unistring.h"
#include "dns.h"

inline u_short read_2oct_value(char *buff) {
	return buff[0] * 256 + buff[1];
}

inline u_int read_4oct_value(char *buff) {
	return read_2oct_value(buff) * 65536 + read_2oct_value(&buff[2]);
}

question_item::question_item() {
	QNAME = NULL;
	QCLASS = CLASS_IN;
}

question_item::question_item(_domain dom, u_short qtype) {
	QNAME = domdup(dom);
	QTYPE = qtype;
	QCLASS = CLASS_IN;
}

question_item::~question_item() {
    free(QNAME);
}

complete_rr::complete_rr() {
	NAME = NULL;
	RDATA = NULL;
	RDLENGTH = 0;
	CLASS = CLASS_IN;
}

complete_rr::complete_rr(_domain _NAME, u_short _TYPE, u_short _CLASS, u_int _TTL, u_short _RDLENGTH, char *_RDATA) {
	NAME = _NAME;
	TYPE = _TYPE;
	CLASS = _CLASS;
	TTL = _TTL;
	RDLENGTH = _RDLENGTH;
	RDATA = _RDATA;
}

complete_rr::~complete_rr() {
    free(NAME);
    if (RDLENGTH != 0) free(RDATA);
}

/* implementation of the DnsMessage class */

DnsMessage::DnsMessage() {
    init();
}

#ifndef _COMPILE_AS_LIBRARY

DnsMessage::DnsMessage(DnsMessage *q) {
	DnsMessage();
	answer_init(q);
}

#endif

DnsMessage::~DnsMessage() {
}

void DnsMessage::init() {
    ID = 0;
    QR = false;
    OPCODE = OPCODE_QUERY;
    AA = false;
    TC = false;
    RD = false;
    RA = false;
    Z = 0;
    RCODE = 0;
}

void DnsMessage::clean() {
 	question.destroy();
 	answer.destroy();
 	authority.destroy();
 	additional.destroy();
}

void DnsMessage::reinit() {
	clean();
	init();
}

//! Reads the RDATA from a DNS message, if nessecary uncompressing dnames
/*! This function reads the RDATA for a RRset from a DNS Message.

    If nessecary (e.g. the rrset's needs_decompression has been set), this
    function will expand the domain names found in the message portion as
    appropriate.

    The fist argument is the buffer in which the decompressed data is put.
    The second argument is the RRtype (DNS Message value) to determine
    whether decompression is neccesary and which properties are available,
    the *msg is the DNS message, and len its length, pos is the current
    position (updated by this function), and rdlen is the value of the
    RDLENGTH field of the RR whose data is being read.

    This function returns the length of the data in the 'res' buffer, or < 0
    if the function failed.
*/
int read_rdata(char*& res, u_int rrtype,
               char *msg, int len, int &pos, int rdlen) {
	AgBuffer *buff;
	char *curprop, *tmp;
	rr_type *rrinfo;
	int begin_pos = pos, tmpint;

	rrinfo = rrtype_getinfo(rrtype);

  if (rrinfo && rrinfo->needs_decompression) {
		/* we may need to do domain name decompression! */
		buff = new AgBuffer();
		curprop = rrinfo->properties;
		while (*curprop) {
			if (pos >= len) { delete buff; return -1; }
			switch (*curprop) {
        case 'm':
				case 'd': /* ah, interesting... this is a domain name! */
					tmp = domdup(&msg[pos], rdlen - (pos-begin_pos), msg, len);
					if (tmp == NULL) { delete buff; return -1; }
					buff->cat(tmp, domlen(tmp));
					pos += domlen(&msg[pos]); // we need the compressed size!
					free(tmp);
					break;
				case 'i': /* 4-byte ip number */
				case 'l': /* 4-byte numeric */
                    if (pos + 4 > len) { delete buff; return -1; }
                    buff->cat(&msg[pos], 4);
                    pos += 4;
                    break;
    			case '6': /* 16-byte ipv6 number */
    				if (pos + 16 > len) { delete buff; return -1; }
    				buff->cat(&msg[pos], 16);
    				pos += 16;
    				break;
                case 's': /* 2-byte numeric */
                    if (pos + 2 > len) { delete buff; return -1; }
                    buff->cat(&msg[pos], 2);
                    pos += 2;
                    break;
                case 'c': /* character string */
                    tmpint = msg[pos];
                    if (pos + tmpint + 1 > len) { delete buff; return -1; }
                    buff->cat(&msg[pos], tmpint + 1);
                    pos += tmpint + 1;
                    break;
                case '7': /* A6-type */
                    if (pos + ((135 - msg[pos]) / 8) + 1 > len) { delete buff; return -1; }
                    tmpint = 0;
                    if (msg[pos] != 0) {
                        /* dname follows, which may NOT be compressed */
                        tmpint = pos + ((135 - msg[pos]) / 8) + 1;
                        tmpint = domlen(msg + tmpint, len - tmpint);
                        if (tmpint <= 0) { delete buff; return -1; }
                    }
                    tmpint += ((135 - msg[pos]) / 8) + 1;
                    buff->cat(&msg[pos], tmpint);
                    pos += tmpint;
                    break;    
                case 'h': /* character strings */
                case 'f': /* foo: nothing particular */
                case 'w': /* well-known services */
                    /* since these must always be the last item and they don't
                       require compression, we don't really check anymore */
                    buff->cat(&msg[pos], rdlen - (pos-begin_pos));
                    pos = begin_pos + rdlen; break;
                default:
                    delete buff; return -1;
            }
            curprop++;
        }
        if (begin_pos + rdlen != pos) { delete buff; return -1; }
        res = buff->buff;
        tmpint = buff->bufflen;
        buff->unreg();
        delete buff;
        return tmpint;
    } else {
    	res = (char *) malloc(rdlen);
    	if (!res) return -1;
    	memcpy(res, &msg[pos], rdlen);
    	pos += rdlen;
    	return rdlen;
    }
}

//! Reads a particular dns section (answer, auth, additional) from raw data
/*! This function reads a DNS section containing RR sets (answer, authoritive
    additional) from a raw DNS message and adds it to the specified DnsMessage
    section object. There is usually no need for functions outside DnsMessage
    to call this function; use DnsMessage::loadFromData instead.

    The first argument is the Vector in which the RRs will be put. The second
    one is the number of RRsets to read from the message (e.g. adcount or
    similar). Data is the raw DNS message, and pos is the current position in
    the message (updated by the function), and len is the length of the raw
    data.

    This function returns the position in the DNS buffer (pos), or a value
    smaller than zero if an error occured.
*/

int DnsMessage::readsection(Vector<complete_rr*>& section, int nitems,
							char *data, int &pos, int len) {
	int t, x, prevpos;
	int rdlen;
	complete_rr *newrr;

	for (t = 0; t < nitems; t++) {
		if (pos >= len) return -1;
		prevpos = pos; /* the location of the domain name */
		x = domlen(&data[pos], len - pos);
		if (x <= 0) return -1;
		pos += x;
		if (pos + 10 > len) return -1;
		rdlen = read_2oct_value(&data[pos + 8]);
		if (pos + 10 + rdlen > len) return -1;
   		/* the data fits, so add it */
   		newrr = new complete_rr();
   		newrr->NAME = domdup(&data[prevpos], len - prevpos, data, len);
   		if (!newrr->NAME) { delete newrr; return -1; }
   		newrr->TYPE = read_2oct_value(&data[pos]);
   		newrr->CLASS = read_2oct_value(&data[pos + 2]);
   		newrr->TTL = read_4oct_value(&data[pos + 4]);
		pos += 10; /* now points to the beginning of the RDATA */
		rdlen = read_rdata(newrr->RDATA, newrr->TYPE, data, len, pos, rdlen);
		if (rdlen < 0) { newrr->RDATA = NULL; delete newrr; return -1; }
		newrr->RDLENGTH = rdlen;
   		section.add(newrr);
	}
	return pos;
}

//! Loads the given DNS message to the object
/*! This function will at first reset the current contents of the DNS message
    by calling reinit(). After that, the given raw DNS message will be read
    and the results will be put in the class.

    The first argument is the raw DNS message and the second is its length.

    This function returns 0 if the action succeeds, or a value smaller than
    zero if an error occured.
*/


int DnsMessage::loadFromData(char *data, int len) {
    int pos = 12, prevpos;
    int x; u_short t;
    u_short qdc, adc, nsc, arc;
    question_item *newq;

    reinit();

    if (len < 12) /* not even a complete header */
        return -1;

    ID = data[0]*256 + data[1];
    QR = (data[2] & 128);
    OPCODE = (ntohs(data[2]) & 120) / 16;
    AA = (data[2] & 4); /* also a reserved bit for dns update */
    TC = (data[2] & 2); /* also a reserved bit for dns update */
    RD = (data[2] & 1); /* also a reserved bit for dns update */
    RA = (data[3] & 128); /* also a reserved bit for dns update */
    Z = (data[3] & 120) / 16;
    RCODE = (data[3] & 15);

    qdc = data[4]*256 + data[5];
    adc = data[6]*256 + data[7];
    nsc = data[8]*256 + data[9];
    arc = data[10]*256 + data[11];

    /* now check the question sector */
    for (t = 0; t < qdc; t++) {
        if (pos >= len) return -1;
        prevpos = pos; /* the position of the domain name */
        pos += domlen(&data[pos], len - pos);
        if (pos+4 > len) return -1;
        newq = new question_item();
        newq->QNAME = domdup(&data[prevpos], len - prevpos, data, len);
        if (!newq->QNAME) { delete newq; return -1; }
        newq->QTYPE = read_2oct_value(&data[pos]);
        newq->QCLASS = read_2oct_value(&data[pos + 2]);
        question.add(newq);
        pos += 4;
    }

    /* question sector */
    x = readsection(answer, adc, data, pos, len); if (x < 0) return x;
    x = readsection(authority, nsc, data, pos, len); if (x < 0) return x;
    x = readsection(additional, arc, data, pos, len); if (x < 0) return x;

    if (pos != len) return -1;

    return 0;
}

//! Retrieves a section
_dm_section& DnsMessage::getsection(_secttype type) {
    switch (type) {
        case S_ANSWER: return answer;
        case S_AUTHORITY: return authority;
        case S_ADDITIONAL: return additional;
    }
    return answer;
}

#ifndef _COMPILE_AS_LIBRARY

//! Prepares the DnsMessage to be an answer to the given query
/*! This function prepares the DNS Message object to be an answer to the given
    question. It will copy the questions ID, RD and OPCODE, and it will set QR
    to true. If copy_request has been set to true, the entries in the question
    section of the question will be copied, which is required for standard
    queries but not nessecary for most other operations.

    The first argument is the question message, and the second argument
    specifies whether to copy the questions from the message.

    This function will always return 0 to indicate the operation succeeded.
*/

int DnsMessage::answer_init(DnsMessage *qmsg, bool copy_request) {
    int t;
    question_item *newitem;
    ID = qmsg->ID;
    QR = true;
    OPCODE = qmsg->OPCODE;
    RD = qmsg->RD;
    if (RD) RA = config.do_recursion; else RA = false;
    if (copy_request) {
        for (t = 0; t < qmsg->question.count; t++) {
            newitem = new question_item();
            newitem->QNAME = domdup(qmsg->question[t]->QNAME);
            newitem->QCLASS = qmsg->question[t]->QCLASS;
            newitem->QTYPE = qmsg->question[t]->QTYPE;
            question.add(newitem);
        }
    }
    return 0;
}

int DnsMessage::is_answer_on(DnsMessage *qw) {
	if (!(ID == qw->ID &&
	      QR == true &&
	      OPCODE == qw->OPCODE)) return false;
	
	if (question.count == 1 && qw->question.count == 1)
		if (!(domcmp(question[0]->QNAME, qw->question[0]->QNAME) &&
		      question[0]->QTYPE == qw->question[0]->QTYPE &&
		      question[0]->QCLASS == qw->question[0]->QCLASS))
			return false;
	
	return true;
}

bool DnsMessage::rrset_exists(_domain dname, u_short rrtype, _secttype section_code) {
    int t;
    _dm_section& section = getsection(section_code);

	for (t = section.count - 1; t >= 0; t--)
        if (section[t]->TYPE == rrtype &&
            domcmp(section[t]->NAME, dname))
            return true;

    return false;
}


//! Adds the given rrset to the dns message
/*! This function adds the given RR set to the DNS message. Before doing that,
    it will check whether any RRs with the same domain name and rr type
    already exist. If it encounters such duplicates, they will be removed.

    The first argument is the domain name to connect the RRset with, the
    second is the RRset, and the last, optional, argument is the section to
    add the RRset to (answer by default).

    This function returns 0 to indicate success.
*/
int DnsMessage::add_rrset(_domain dname, rr_set* rrset,
                           bool duplicheck, _secttype section_code) {
    complete_rr *newrr;
    int t;
    u_int ttl;
	int itemix;
	_dm_section& section = getsection(section_code);

    /* at first, check for duplicates */
    if (duplicheck) {
        if (rrset_exists(dname, rrset->type, section_code))
            return 0;
        if (section_code == S_ADDITIONAL) {
            if (rrset_exists(dname, rrset->type, S_ANSWER)) return 0;
            if (rrset_exists(dname, rrset->type, S_AUTHORITY)) return 0;
        }
    }

   	if (rrset->ttl != -1)
   		ttl = rrset->ttl;
   	else
    	ttl = rrset->expire - curtime;

	/* each time the rr is requested, we send another item first */
    if (rrset->last >= rrset->items.count) rrset->last = 0;

    /* allright, now add the RRs */
    for (t = rrset->last; t < rrset->last + rrset->items.count; t++) {
        if (t >= rrset->items.count)
        	itemix = t - rrset->items.count;
        else
        	itemix = t;

    	newrr = new complete_rr();
    	newrr->NAME = domdup(dname);
    	newrr->TYPE = rrset->type;
    	newrr->CLASS = CLASS_IN;
		newrr->TTL = ttl;
		newrr->RDLENGTH = rrset->items[itemix]->RDLENGTH;
		newrr->RDATA = (char *)memdup(rrset->items[itemix]->RDATA, newrr->RDLENGTH);
		section.add(newrr);
	}

	rrset->last++;

	return 0;
}

//! Adds all rrsets matched by the qtype to the dns message
/*! This function adds all RRs of the given CacheItem to the DNS Message that
    match the given qtype. It will call add_rrset for each found rr set, so
    duplicate checking will be performed.

    This function asks for the domain name to connect the RRset(s) with, the
    CacheItem describing the domain, the question type and the section to add
    the rrset to (answer by default), respectively.

    This function returns 0 if the operation succeeds, or a value smaller than
    zero when an error has occured (for example, the qtype is not understood).
    This function will not complain if the qtype is supported but no matching
    rrsets are found (the 'NODATA' case).
*/

int DnsMessage::add_rrset_qtype(_domain dname, CacheItem *item, int qtype,
								bool dupchk, _secttype section_code) {
	rr_set *rrset;
	int ret;

 	/* special RR type */
 	if (qtype == QTYPE_ALL) {
 		/* all RR sets */
 		for (int x = 0; x < item->rr_sets.count; x++) {
 			ret = add_rrset(dname, item->rr_sets[x], dupchk, section_code);
 			if (ret != 0) return ret;
 		}
 		return 0;
 	} else if (qtype == QTYPE_MAILB) {
 		rrset = item->get_rrset(DNS_TYPE_MD);
 		if (rrset) {
 			ret = add_rrset(dname, rrset, dupchk, section_code);
 			if (ret < 0) return ret;
 		}
 		rrset = item->get_rrset(DNS_TYPE_MF);
 		if (rrset) {
 			ret = add_rrset(dname, rrset, dupchk, section_code);
 			if (ret < 0) return ret;
 		}
 		return 0;
 	} else {
 		rrset = item->get_rrset(qtype);
 		if (rrset)
 			return add_rrset(dname, rrset, dupchk, section_code);
 		else
 			return 0;
  }
}

//! Performs additional section processing for each rr in the message
/*! This function will perform additional section processing for each RR in
    the answer and authoritive sections of the DNS message. For each RR which
    has a type for which section_proc is set, the domain names in the RDATA
    will be tracked, and, if any addresses for that domain are available,
    those will be added to the additional section.

    Returns 0 to indicate success.
*/

int DnsMessage::do_additional_section_processing() {
	int ret;

	ret = do_additional_section_processing_for(S_ANSWER);
	if (ret != 0) return ret;

	ret = do_additional_section_processing_for(S_AUTHORITY);
	if (ret != 0) return ret;

	return 0;
}

//! Performs additional section processing for each rr in the section
/*! This function is called by do_additional_section_processing() for the
    two sections it operates on. For general info, see the description of
    that function.
*/

int DnsMessage::do_additional_section_processing_for(_secttype section_code) {
	int t, z, u;
	int ret;
	rr_type *rrtype;
	char *ptr, *curprop;
	Zone *zn; CacheItem *ci;
	rr_set *rrset;
	_dm_section& section = getsection(section_code);

	for (t = 0; t < section.count; t++) {
		rrtype = rrtype_getinfo(section[t]->TYPE);
		if (rrtype && rrtype->section_proc == SECTION_PROC_A) {
			/* we want additional section processing! */
			curprop = rrtype->properties;
			ptr = section[t]->RDATA;

			while (*curprop) {
				switch (*curprop) {
					case 'd': /* a domain name we need to process */
						ret = LookupDomain(ci, zn, ptr);
						if (ret & LOOKUP_FOUND) {
							/* the item has even been found! */
							rrset = ci->get_rrset(DNS_TYPE_CNAME);
							if (rrset) {
								add_rrset(ptr, rrset, true, S_ADDITIONAL);
							} else  {
								add_rrset_qtype(ptr, ci, DNS_TYPE_A,
												true, S_ADDITIONAL);

                add_rrset_qtype(ptr, ci, DNS_TYPE_A6,
												true, S_ADDITIONAL);

                add_rrset_qtype(ptr, ci, DNS_TYPE_AAAA,
												true, S_ADDITIONAL);
							}
						}
						ptr += domlen(ptr);
						break;
                    case 'm':           ptr += domlen(ptr); break;
					case 'i': case 'l': ptr += 4; break;
					case '6':			ptr += 16; break;
          case '7': /* ipv6 */
            z = 16;
            u = 129;
            while (ptr) {          
              ret = ptr[0];
              if (ret == 0 || --z <= 0 || ret > u) break;
              u = ret;
              ptr += ((135 - ret) / 8) + 1;
              ret = LookupDomain(ci, zn, ptr);
              if (ret & LOOKUP_FOUND) {
                rrset = ci->get_rrset(DNS_TYPE_A6);
                if (rrset && rrset->items.count) {
                  add_rrset(ptr, rrset, true, S_ADDITIONAL);
                  ptr = rrset->items[0]->RDATA;
                } else ptr = NULL;
              } else ptr = NULL;
              if (zn->citem_root) add_rrset_qtype(zn->domain, zn->citem_root, DNS_TYPE_NS, true, S_ADDITIONAL);
            }
            break;
					case 's':			ptr += 2; break;
					case 'c':			ptr += *ptr + 1; break;
					default:			ptr = NULL; break;
				}
				curprop++;
			}
		}
	}
	return 0;
}

#endif /* !_COMPILE_AS_LIBRARY */

//! Compiles the DNS message class to a sendable buffer
/*! This function compiles the DNS message. It will put the result in an
    AgBuffer, which can be used to send to clients or servers.

    This function asks for the AgBuffer to put the result in, and the maximum
    length of the DNS message (or -1 if no limit). If the DNS message exceeds
    the max length, we will at first try to cut the additional section. If
    it still doesn't fit, the TC bit will be set and the overhead will be cut
    off.

    This function returns 0 on success or a smaller value on error.
*/

#define overflow() ( maxlen != -1 && buff->bufflen > maxlen )

int DnsMessage::compile(AgBuffer *buff, int maxlen) {
    int t, ret, prevpos;
    /* at first, do the header */
    buff->catshort(ID);
    if (QR) buff->setbit(2, 0);
    buff->set4bit(2, 1, OPCODE);
    if (AA) buff->setbit(2, 5);
    if (RD) buff->setbit(2, 7);
    if (RA) buff->setbit(3, 0);
    buff->set4bit(3, 4, RCODE);
    buff->catshort(0);
    buff->catshort(0);
    buff->catshort(0);
    buff->catshort(0);

    /* question sector */
    for (t = 0; t < question.count; t++) {
        prevpos = buff->bufflen;
        ret = write_compressed_domain(question[t]->QNAME, buff);
        if (ret < 0) { return -1; }
        buff->catshort(question[t]->QTYPE);
        buff->catshort(question[t]->QCLASS);
        if (overflow()) {
            buff->cutto(prevpos);
            buff->setbit(2, 6);
            break;
        }
    }

    buff->modshort(4, t);
    if (overflow()) return 0;

    /* other sections */
    for (t = 0; t < answer.count; t++) {
        prevpos = buff->bufflen;
        ret = write_rr(answer[t], buff);
        if (ret < 0) return ret;
        if (overflow()) {
            buff->cutto(prevpos);
            buff->setbit(2, 6); /* TC bit */
            break;
        }
    }

    buff->modshort(6, t);
    if (overflow()) return 0;

    for (t = 0; t < authority.count; t++) {
        prevpos = buff->bufflen;
        ret = write_rr(authority[t], buff);
        if (ret < 0) return ret;
        if (overflow()) {
            buff->cutto(prevpos);
            buff->setbit(2, 6); /* TC bit */
            break;
        }
    }

    buff->modshort(8, t);
    if (overflow()) return 0;

    for (t = 0; t < additional.count; t++) {
        prevpos = buff->bufflen;
        ret = write_rr(additional[t], buff);
        if (ret < 0) return ret;
        if (overflow()) {
            buff->cutto(prevpos);
            break;
        }
    }

    buff->modshort(10, t);

    return 0;
}

//! Writes one particular DNS section to a DNS message
/*! This function writes one of the DNS sections answer, additional or
    authority to a DNS message buffer. It will compress domain names and all
    that, but it won't check whether the length of the message exceeds the
    maximum, a task which should be performed by the caller (usually the
    DnsMessage::compile function).

    This function returns 0 on success or a smaller value on error.
*/

int DnsMessage::write_rr(complete_rr *rr, AgBuffer *buff) {
	int ret, rdl_pos;
	rr_type *rrinfo;
	char *curprop, *rdata;

	if (!rr->NAME) return -1;
	ret = write_compressed_domain(rr->NAME, buff);
	if (ret < 0) return -1;
	buff->catshort(rr->TYPE);
	buff->catshort(rr->CLASS);
	buff->catint(rr->TTL);

	rrinfo = rrtype_getinfo(rr->TYPE);
	if (rrinfo && rrinfo->compress_dname) {
		/* we need to compress domain names in the RR */
		rdl_pos = buff->bufflen;
		buff->catshort(0); /* placeholder for the RDLENGTH */
		curprop = rrinfo->properties;
		rdata = rr->RDATA;
		while (*curprop) {
			switch(*curprop) {
                case 'm':
				case 'd': /* domain name: needs to be compressed */
					ret = write_compressed_domain(rdata, buff);
					if (ret < 0) return -1;
					rdata += domlen(rdata);
					break;
				case 'i': case 'l':
					buff->cat(rdata, 4); rdata += 4; break;
				case '6':
					buff->cat(rdata, 16); rdata += 16; break;
                case '7':
                    ret = rdata[0];
                    if (ret != 0) { ret = ((135 - ret) / 8) + 1; ret += domlen(rdata + ret); }
                    else ret = ((135 - ret) / 8) + 1;
                    buff->cat(rdata, ret);
                    rdata += ret;
                    break;
				case 's':
					buff->cat(rdata, 2); rdata += 2; break;
				case 'c':
					buff->cat(rdata, *rdata + 1); rdata += *rdata + 1; break;
				case 'h': case 'f': case 'w':
					/* these are always the last entries in a RDATA */
					buff->cat(rdata, rdata - rr->RDATA); break;
			}
			curprop++;
		}
		/* now set the right rd length */
		buff->modshort(rdl_pos, buff->bufflen - rdl_pos - 2);
	} else {
		/* no compression nessecary, so just do a straightforward copy */
		buff->catshort(rr->RDLENGTH);
		buff->cat(rr->RDATA, rr->RDLENGTH);
	}
    return 0;
}

//! Writes a domain name to an AgBuffer, compressing it if possible.
/*! This function writes a domain name to an AgBuffer, performing DNS domain
	name compression if possible.

    Returns 0 on success or a value smaller than zero on error.
*/

int write_compressed_domain(_domain domain, AgBuffer *buff) {
	char domportion[255];
	char *cbuff;
	int t, x, ret, nlabels, max, pos;

	nlabels = domlabelcount(domain);
	for (t = 0; t < nlabels - 1; t++) {
		ret = domfrom(domportion, domain, t);
		if (ret < 0) return -1;

		/* look for the domain in the dns message */
		max = buff->bufflen - domlen(domportion);
		cbuff = buff->buff;
		for (x = 0; x < max; x++) {
			if (cbuff[x] == domportion[0]) {
				/* worth running domcmp() */
				if (advdomcmp(domportion, x, cbuff, buff->bufflen)) {
					/* write the portion we couldn't compress */
					ret = domto(domportion, domain, t);
					buff->cat(domportion, domlen(domportion) - 1);
					/* write a reference to the position */
					pos = buff->bufflen;
					buff->catshort(x);
					buff->setbit(pos, 0); /* 'compressed domain' magic */
					buff->setbit(pos, 1);
					return 0;
				}
			}
		}
	}
	/* apparently, the compression failed. thus, write the complete name */
	buff->cat(domain, domlen(domain));
	return 0;
}

#ifndef _COMPILE_AS_LIBRARY

_domain get_current_qname(DnsMessage *a) {
    int t;
    _domain dom = a->question[0]->QNAME;
    for (t = 0; t < a->answer.count; t++)
        if (a->answer[t]->TYPE == DNS_TYPE_CNAME)
            dom = a->answer[t]->RDATA;

    return dom;
}

bool reached_max_cname_level(DnsMessage *a) {
    int ret = 0, x;
    for (x = 0; x < a->answer.count; x++)
        if (a->answer[x]->TYPE == DNS_TYPE_CNAME) ret++;
    if (ret >= config.max_cname)
        return true;
    else
        return false;
}

#endif
