/*
    Posadis - A DNS Server
    Configuration implementation
    Copyright (C) 2001  Meilof Veeningen <meilof@users.sourceforge.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "dns.h"
#include "getconfigdir.h"
#include "include/unistring.h"
#include "include/unifile.h"

/*!
  This source file contains the Posadis configuration engine. More information
  on how to add a configuration setting can be found in the configuration.h
  header -- this text handles how the data is stored internally.

  Each configuration item (e.g. main, primary, hint) has its own conf_settings
  structure which contains a list of all the possible options. An item also
  has a class, derived from "configuration_data", which can be seen as an
  instance of the configuration item. The class contains properties for the
  settings from the conf_settings struct, in exactly the same order, with only
  the pointer and int datatypes. Because pointers and ints have the same size
  (a prerequisite for this system to work), the configuration engine can simply
  determine the result address for a setting by looking which index the setting
  has in the conf_settings array.

  Settings can be eiter multiple or not (determined in the conf_settings
  struct). The way values are stored depends on this.

  If a setting isn't 'multiple', only one value can be given in the
  configuration file -- each new value will overwrite the previous one. For the
  DT_BOOL, DT_INT and DT_ENUM datatypes, the value is written directly
  to the int in the configuration_data-derived class.

  DT_STRING means a pointer to the string.

  DT_IDENT is, unlike in previous versions, a pointer to a
  DT_IDENT is also a  char*. If the first char is 0, the identifier is an ip
  number and the next 4 byts contain the ip number. If it's 1, the identifier
  is a domain name and the next bytes contain the domain name. After the
  identifier, two bytes follow containing the port (DNS_PORT by default).

  DT_IDENTPAIR is a pair of a domain
  name and an address. It's a char*, at first, it always contains the domain
  name, followed by the type of address which is given (currently 0=none and
  1=IPv4 are supported), and after that the actual address (if the type of
  address is not 0=none).

  If a setting is multiple, it will always be a char*. The first two bytes of
  the data it points to are the number of items (the second byte being the most
  significant). After that, the data comes that would normally either have been
  stored directly into the result address (a numeric value or a pointer).

  Even after this small explanation, many questions will probably still remain
  unanswered. I'm sorry about that, but you can always mail me if things are
  not clear.
*/

/* internal subroutines */
void print_help(bool is_err = false); /* print command-line help */
void print_version(); /* print command-line version */

/* ----- configuration variables ----------------------------------------------------------------- */

void clear_single_value(_conf_datatype datatype, intptr_t *val) {
    intptr_t tmpval;
    switch (datatype) {
        case DT_STRING:
        case DT_IDENT:
        case DT_IDENTPAIR:
            memcpy(&tmpval, val, sizeof(intptr_t));
            free((char *)tmpval);
            break;
        default:
            break;
    }
}

void clear_setting(setting *setting, intptr_t *val) {
    if (setting->multiple) { /* multiple settings */
        multi_free( (multi_val *) val, setting->datatype);
    } else { /* only one */
        clear_single_value(setting->datatype, val);
    }
}

void clone_single_value(_conf_datatype datatype, intptr_t *src, intptr_t *res) {
    char **srcptr = (char **) src;
    char **resptr = (char **) res;

    int len;

    switch (datatype) {
        case DT_BOOL:
        case DT_INT:
        case DT_ENUM:
            *res = *src;
            break;
        case DT_STRING:
            if (*srcptr) *resptr = strdup(*srcptr); else *resptr = NULL;
            break;
        case DT_IDENT:
            if (*srcptr) {
                if ( (*srcptr)[0]) /* an ip number */
                    *resptr = (char *) memdup(*srcptr, 7);
                else /* a domain name */
                    *resptr = (char *) memdup(*srcptr, domlen(*srcptr + 1) + 3);
            } else *resptr = NULL;
            break;
        case DT_IDENTPAIR:
            if (*srcptr) {
            	len = domlen(*srcptr);
            	if ( (*srcptr)[len] == 0 ) { /* no address */
            		*resptr = (char *) memdup(*srcptr, len + 1);
            	} else { /* ipv4 number */
            		*resptr = (char *) memdup(*srcptr, len + 5);
            	}
            } else {
                *resptr = NULL;
            }
    }

}

void clone_setting(setting *setting, intptr_t *src, intptr_t *res) {
    int len;
    int x;

    char** srcptr = (char **) src;
    char** resptr = (char **) res;

    if (setting->multiple) { /* multiple settings */
        if (*srcptr == NULL) {
            /* no value as of yet */
            len = 0;
            *resptr = NULL;
        } else {
            len = (*srcptr)[0] + (*srcptr)[1]*256;

            *resptr = (char *)malloc(2 + len * sizeof(intptr_t)); // allocate value
            memcpy(*resptr, *srcptr, 2); // copy two lenght bytes

            for (x = 0; x < len; x++) {
                /* call clone_single_value for each single setting */
                clone_single_value(setting->datatype,
					(intptr_t *) (*srcptr + 2 + x*sizeof(intptr_t)),
					(intptr_t *) (*resptr + 2 + x*sizeof(intptr_t)));
            }
        }
    } else { /* only one */
        clone_single_value(setting->datatype, src, res);
    }

}

/* functions for multi_vals */
int multi_get_count(multi_val val) {
    if (val == NULL)
        return 0;
    else
        return val[0] + val[1] * 256;
}

void multi_add(multi_val* val, intptr_t value) {
    int len = multi_get_count(*val);
    *val = (multi_val) realloc(*val, ++len * sizeof(intptr_t) + 2);
    memcpy(*val + (len * sizeof(intptr_t)) - 2, &value, sizeof(intptr_t));
    (*val)[0] = len; (*val)[1] = len / 256;
}

void multi_free(multi_val* val, _conf_datatype datatype) {
    int x, len;
    if (*val == NULL) {
        /* no value as of yet */
        return;
    } else {
        len = (*val)[0] + (*val)[1]*256;

        for (x = 0; x < len; x++) {
            /* call clone_single_value for each single setting */
            clear_single_value(datatype,
                (intptr_t *) (*val + 2 + x*sizeof(intptr_t)));
        }
    }
    free(*val);
    *val = NULL;
}

void multi_remove(multi_val* val, int ix, _conf_datatype datatype) {
    int count = multi_get_count(*val);
    clear_single_value(datatype, (intptr_t *) (*val + 2 + ix * sizeof(intptr_t)));
    *val = (multi_val) realloc( (char*) *val, --count * sizeof(intptr_t) + 2);
    (*val)[0] = count; (*val)[1] = count / 256;
}

char* multi_getptr(multi_val val, int ix) {
    char *res;
    memcpy(&res, &val[2 + ix * sizeof(intptr_t)], sizeof(intptr_t));
    return res;
}

intptr_t multi_getval(multi_val val, int ix) {
    return (intptr_t) multi_getptr(val, ix);
}

multi_val multi_clone(multi_val *setting, _conf_datatype type) {
    _setting sett = { "", type, true, NULL };
    multi_val val;

    clone_setting(&sett, (intptr_t *)setting, (intptr_t *)&val);
    return val;
}

/* functions */
int addaddressbyident(Vector<_addr *> *resvec, char *ptr) {
    resvec->add((_addr *)memdup(ptr, sizeof(_addr)));
    return 0;
}

bool is_in_identlist(multi_val identlist, _addr *addr) {
    int nitems = multi_get_count(identlist), x;

    for (x = 0; x < nitems; x++)
        if (compare_host(addr, (_addr *)multi_getptr(identlist, x))) return true;

    return false;
}


/* implementation of the base configuration class */

configuration_data::configuration_data(conf_settings *_info, intptr_t *_begin) {
    info = _info;
    begin = _begin;
    overrided = new BitList(info->n_settings);
}


configuration_data::~configuration_data() {
    int t;
    intptr_t *val;

    val = (intptr_t *)begin;

    for (t = 0; t < info->n_settings; t++) {
        clear_setting(&info->settings[t], val);

        val++;
    }

    delete overrided;
}

void configuration_data::clone(configuration_data *item) {
    int t;
    char *res, *src;

    res = (char *)begin;
    src = (char *)item->begin;

    for (t = 0; t < info->n_settings; t++) {
        clone_setting(&info->settings[t], (intptr_t *)src, (intptr_t *)res);

        res += sizeof(intptr_t);
        src += sizeof(intptr_t);
    }
}

void configuration_data::inherit(configuration_data *item) {
    int t;
    char *res, *src;

    res = (char *)begin;
    src = (char *)item->begin;

    for (t = 0; t < info->n_settings; t++) {
        if (item->overrided->isset(t))
            clone_setting(&info->settings[t], (intptr_t *)src, (intptr_t *)res);

        res += sizeof(intptr_t);
        src += sizeof(intptr_t);
    }
}

void configuration_data::override_c(char *optname) {
    override_p(getpropaddr(optname));
}

void configuration_data::override_p(void *setting) {
    overrided->set( ((intptr_t)setting-(intptr_t)this->begin) / sizeof(intptr_t) );
}

void configuration_data::unoverride_c(char *optname) {
    unoverride_p(getpropaddr(optname));
}

void configuration_data::unoverride_p(void *setting) {
    overrided->set( ((intptr_t)setting-(intptr_t)this->begin) / sizeof(intptr_t) );
}

bool configuration_data::isoverrided_c(char *optname) {
    return isoverrided_p(getpropaddr(optname));
}

bool configuration_data::isoverrided_p(void *setting) {
    return overrided->isset( ((intptr_t)setting-(intptr_t)this->begin) / sizeof(intptr_t) );
}

void *configuration_data::getpropaddr(char *optname) {
    void *addr = begin;
    for (int t = 0; t < info->n_settings; t++) {
        if (strcmpi(info->settings[t].command, optname) == 0) return addr;
        addr = (char *)addr + sizeof(char*);
    }
    return NULL;
}

setting *configuration_data::getcommandinfo(char *optname, conf_settings *cinfo) {
    int t;
    if (!cinfo) cinfo = info;

    for (t = 0; t < cinfo->n_settings; t++)
        if (strcmpi(cinfo->settings[t].command, optname) == 0) return &cinfo->settings[t];

    return NULL;
}



/* ----- main configuration: affects the complete posadis process -------------------------------- */
main_configuration_data::main_configuration_data() :
    configuration_data(&main_config, (intptr_t *)&this->udp_listen) {
    _addr addr;

    /* set hard-coded defaults */
    udp_listen = 0;
    getaddress(&addr, "0.0.0.0", DNS_PORT);
    multi_add(&udp_listen, (intptr_t) memdup(&addr, sizeof(addr)));
    
    tcp_listen = 0;
    multi_add(&tcp_listen, (intptr_t) memdup(&addr, sizeof(addr)));

#ifdef _WIN32
    ipv6_client = 0;
#else
    ipv6_client = 1;
#endif
    max_cname = 10;
    do_recursion = 0;
    max_threads = 500;
    datadir = NULL;
    tcp_in_keepalive = 120;
    tcp_out_keepalive = 30;
    patience = 5;
    tcp_patience = 60;
    do_query_logging = 0;
    hide_dname = 0;
#ifndef _WIN32
    user = NULL;
    group = NULL;
#endif
}

setting main_settings[] = {
    { "udp_listen", DT_IDENT, true, NULL },
    { "tcp_listen", DT_IDENT, true, NULL },
    { "ipv6_client", DT_BOOL, false, NULL },
    { "max_cname", DT_INT, false, NULL },
    { "do_recursion", DT_BOOL, false, NULL },
    { "max_threads", DT_INT, false, NULL },
    { "datadir", DT_STRING, false, NULL },
    { "tcp_in_keepalive", DT_INT, false, NULL },
    { "tcp_out_keepalive", DT_INT, false, NULL },
	  { "patience", DT_INT, false, NULL },
    { "tcp_patience", DT_INT, false, NULL },
    { "do_query_logging", DT_BOOL, false, NULL },
    { "hide_dname", DT_BOOL, false, NULL },
#ifndef _WIN32
    { "runuser", DT_STRING, false, NULL },
    { "rungroup", DT_STRING, false, NULL },
#endif
};

conf_settings main_config = {
    /* settings */ main_settings,
    /* n_settings */ sizeof(main_settings) / sizeof(setting),
};


main_configuration_data config;

/* ----- zone configuration: affects all zones --------------------------------------------------- */
zone_configuration_data::zone_configuration_data(bool is_std) :
    configuration_data(&zone_config, (intptr_t *)&this->auth_servers) {

    if (is_std) { /* set hard-coded defaults */
        auth_servers = NULL;
    } else { /* from std_zone_config */
        clone(&std_zone_config);
    }
}

setting zone_settings[] = {
    { "auth_servers", DT_IDENTPAIR, true, NULL },
};

conf_settings zone_config = {
    /* settings */ zone_settings,
    /* n_settings */ sizeof(zone_settings) / sizeof(setting),
};
zone_configuration_data std_zone_config(true);

/* ----- master configuration: settings for xfr providing ---------------------------------------- */
enumeration a_xfr_to_enum[] = {
	{ XFR_TO_NONE, "none" },
	{ XFR_TO_SLAVES, "slaves" },
	{ XFR_TO_ANY, "any" },
	{ 0, "" },
};

auth_configuration_data::auth_configuration_data(bool is_std) :
	configuration_data(&auth_config, &this->xfr_to) {
	
	if (is_std) {
        xfr_to = XFR_TO_ANY;
        also_xfr_to = NULL;
        slaves = NULL;

        notify_to_slaves = 1;
        also_notify = NULL;
        notify_interval = 30;
        notify_retry = 3;
	} else {
		clone(&std_auth_config);
	}
}

setting auth_settings[] = {
	{ "xfr_to", DT_ENUM, false, a_xfr_to_enum },
    { "also_xfr_to", DT_IDENT, true, NULL },

    { "slaves", DT_IDENT, true, NULL },

    { "notify_to_slaves", DT_BOOL, false, NULL },
    { "also_notify", DT_IDENT, true, NULL },
    { "notify_interval", DT_INT, false, NULL },
    { "notify_retry", DT_INT, false, NULL },
};
conf_settings auth_config = {
    /* settings */ auth_settings,
    /* n_settings */ sizeof(auth_settings) / sizeof(setting),
};

auth_configuration_data std_auth_config(true);

/* ----- primary configuration: master file readers ---------------------------------------------- */

primary_configuration_data::primary_configuration_data(bool is_std) :
    configuration_data(&primary_config, (intptr_t*)&this->file) {

    if (is_std) { /* set hard-coded defaults */
    	file = NULL;
        zone_ttl = 3600;
    } else {
        clone(&std_primary_config);
    }
}

setting primary_settings[] = {
	{ "file", DT_STRING, false, NULL },
    { "zone_ttl", DT_INT, false, NULL },
};

conf_settings primary_config = {
    /* settings */ primary_settings,
    /* n_settings */ sizeof(primary_settings) / sizeof(setting),
};

primary_configuration_data std_primary_config(true);

/* ----- secondary configuration: ixfr recipients ------------------------------------------------ */
secondary_configuration_data::secondary_configuration_data(bool is_std) :
    configuration_data(&secondary_config, (intptr_t *)&this->masters) {

    if (is_std) { /* set hard-coded defaults */
    	masters = NULL;
    	initial_retry_interval = 3600; /* one hour */
    } else {
        clone(&std_secondary_config);
    }
}

setting secondary_settings[] = {
	{ "masters", DT_IDENT, true, NULL },
	{ "initial_retry_interval", DT_INT, false, NULL },
};

conf_settings secondary_config = {
    /* settings */ secondary_settings,
    /* n_settings */ sizeof(secondary_settings) / sizeof(setting),
};

secondary_configuration_data std_secondary_config(true);

/* ----- hint configuration: the hints that are given -------------------------------------------- */

hint_configuration_data::hint_configuration_data(bool is_std) :
	configuration_data(&hint_config, (intptr_t *)&this->nameserver) {

	if (is_std) { /* set hard-coded defaults */
            nameserver = NULL;
            ns_ttl = 7200; // 2 hours
            source = NULL;
            pre_expire = 60; /* minute before expire */
	} else {
		clone(&std_hint_config);
	}
}

setting hint_settings[] = {
	{ "nameserver", DT_IDENTPAIR, true, NULL },
	{ "ns_ttl", DT_INT, false, NULL },
    { "source", DT_IDENT, true, NULL },
    { "pre_expire", DT_INT, false, NULL },
};

conf_settings hint_config = {
	/* settings */ hint_settings,
	/* n_settings */ sizeof(hint_settings) / sizeof(setting),
};

hint_configuration_data std_hint_config(true);

/* ----- tunnel configuration: the nameserver to tunnel to -------------------------------------------- */

/* tunnel::recursive_answer_fail enumeration */
enumeration t_recfail_enum[] = {
	{ T_RF_NODATA, "nodata" },
	{ T_RF_NXDOMAIN, "nxdomain" },
    { T_RF_HINT, "hint" },
	{ T_RF_NONE, "none" },
	{ 0, "" },
};

tunnel_configuration_data::tunnel_configuration_data(bool is_std) :
		configuration_data(&tunnel_config, (intptr_t *)&this->tunnelserver) {

		if (is_std) { /* set hard-coded defaults */
			tunnelserver = NULL;
			recursive_answer_fail = T_RF_HINT;
			retry = 3;
            no_aa = 0;
		} else {
			clone(&std_tunnel_config);
		}
}

setting tunnel_settings[] = {
		{ "tunnelserver", DT_IDENT, true, NULL },
		{ "recursive_answer_fail", DT_ENUM, false, t_recfail_enum },
		{ "retry", DT_INT, false, NULL },
        { "no_aa", DT_INT, false, NULL },
};

conf_settings tunnel_config = {
		/* settings */ tunnel_settings,
		/* n_settings */ sizeof(tunnel_settings) / sizeof(setting),
};

tunnel_configuration_data std_tunnel_config(true);


/* global config settings */
char conf_configdir[PATH_MAX] = "";

#ifndef _WIN32

#ifdef _DEBUG_
bool do_fork = false;
#else
bool do_fork = true;
#endif

#else
bool auto_minimize = true;
#endif

/* loads the command-line arguments and determines conf directory */
#define ARGUMENT_MATCHES(str) (stricmp(argv[x], str) == 0)
int process_cmdline_args(int argc, char **argv) {
    int x;
    char tmp[PATH_MAX];
    /* evaluate command-line arguments */
    for (x = 1; x < argc; x++) {
        if (strlen(argv[x]) >= 1024) argv[x][1024] = '\0'; /* cut of items that are too long */
        if (ARGUMENT_MATCHES("--help") || ARGUMENT_MATCHES("-h")) {
            print_help();
            return 1;
        }
        if (ARGUMENT_MATCHES("--version") || ARGUMENT_MATCHES("-v")) {
            print_version();
            return 1;
        }
        if (ARGUMENT_MATCHES("--conf-dir") || ARGUMENT_MATCHES("-c")) {
            if (x == argc - 1) {
                log_print(LOG_LEVEL_PANIC, "Parameter <dir> for --conf-dir option required!");
                print_help();
                return 1;
            } else {
                if (strlen(argv[++x]) + 1 >= PATH_MAX) {
                    log_print(LOG_LEVEL_PANIC, "Conf directory too long!");
                    return -1;
                }
                strcpy(conf_configdir, argv[x]);
                if (conf_configdir[strlen(conf_configdir)-1] != pathdelim[0]);
                    strcat(conf_configdir, pathdelim);
            }
            continue;
        }
        if (ARGUMENT_MATCHES("--data-dir") || ARGUMENT_MATCHES("-d")) {
            if (x == argc - 1) {
                log_print(LOG_LEVEL_PANIC, "Parameter <dir> for --data-dir option required!");
                print_help();
                return 1;
            } else {
                if (strlen(argv[++x]) + 1 >= PATH_MAX) {
                    log_print(LOG_LEVEL_PANIC, "Data directory too long!");
                    return -1;
                }
                strcpy(tmp, argv[x]);
                if (tmp[strlen(tmp) - 1] != pathdelim[0]);
                    strcat(tmp, pathdelim);

                if (is_posadis_dir("data", tmp)) {
                    config.datadir = strdup(tmp);
		    config.override_p(config.datadir);
                } else {
                    log_print(LOG_LEVEL_PANIC, "Data dir on command line is incorrect!");
                    return -1;
                }
            }
            continue;
        }
#ifndef _WIN32
        if (ARGUMENT_MATCHES("--fork") || ARGUMENT_MATCHES("-f")) {
        	do_fork = true;
        	continue;
        }
        if (ARGUMENT_MATCHES("--foreground") || ARGUMENT_MATCHES("-g")) {
            do_fork = false;
            continue;
        }
#else
	  if (ARGUMENT_MATCHES("--show") || ARGUMENT_MATCHES("-s")) {
		auto_minimize = false;
		continue;
	  }
#endif
        log_print(LOG_LEVEL_PANIC, "Unrecognized option: %s", argv[x]);
        print_help();
        return 1;
    }

    return 0;
}

void print_help(bool is_err) {
    if (!is_err) log_print(LOG_LEVEL_INFO, _POSADIS_VERINFO_ "\n");
    log_print(LOG_LEVEL_INFO, "Usage: posadis [options]\n"
           "Options:\n"
           " -c, --conf-dir <dir>    Use <dir> dir for configuration files\n"
           " -d, --data-dir <dir>    Use <dir> dir for internal storage\n"
#ifndef _WIN32
           " -f, --fork              Perform a fork and run as a daemon"
#ifndef _DEBUG_
           " [default]"
#endif
           "\n"
           " -g, --foreground        Run Posadis in the foreground (don't fork)"
#ifdef _DEBUG_
           " [default]"
#endif
           "\n"
#else
           " -s, --show              Don't auto-minimize the log window\n"
#endif
           " -h, --help              Print this help message\n"
           " -v, --version           Prints version information\n\n"
           "For more information on Posadis, consult the Posadis Reference or\n"
           "visit http://posadis.sourceforge.net.\n\n"
           "This is also the place to be if you find any bugs.");
}

void print_version() {
    log_print(LOG_LEVEL_INFO, "Posadis - A DNS server\n"
           "Platform:    " PLATFORM "\n"
           "Version:     " VERSION "\n"
           "Compiled on: " __DATE__ "\n"
#ifdef _DEBUG_
	       "Debug build: yes\n"
#endif
#ifdef _EXTDEBUG_
		   "Extra's:     Excessive debugging!\n"
#endif
		   "\n"
           "For more information on Posadis, consult the Posadis Reference or\n"
           "visit http://posadis.sourceforge.net.\n\n"
           "This is also the place to be if you find any bugs.");
}

int determine_datadir() {
  char tmp[PATH_MAX + 1];
  if (config.datadir == NULL || !strlen(config.datadir)) {
      if (get_config_dir("data", tmp) != 0) {
        log_print(LOG_LEVEL_PANIC, "Could not find a valid data dir!");
	return -1;
      }
      config.datadir = strdup(tmp);
      config.override_p(config.datadir);
  }
  
  if (config.datadir[strlen(config.datadir) - 1] != pathdelim[0]) {
    char *tptr = (char *)malloc(strlen(config.datadir) + 2);
    strcpy(tptr, config.datadir);
    strcat(tptr, pathdelim);
    free(config.datadir);
    config.datadir = tptr;
  }

  return 0;
}

/* ----- configuration file implementation ----- */


#define line_read_error(errcode)  {											\
	cptr = file.geterrmsg(); 												\
	log_print(LOG_LEVEL_ERROR, "%s:%d: %s", file.name, file.linenum, cptr);	\
	free(cptr);																\
	return -1;																\
	}


/* reads the arguments of a configuration statement */
int handle_config_item(posadis_infile& file, setting *command, configuration_data *ent) {
    char item[256], yadb[260], *val = item, *tmp;
    intptr_t *res = (intptr_t *)ent->getpropaddr(command->command);
    char **resptr = (char **)res;
    intptr_t intres = 0; /* the new value */
    bool another_value = false; /* another value of the same type follows */
    enumeration *curenum;
    int ret, x;
    char *cptr;

	ret = file.read_entry(item, sizeof(item));
	if (ret < 0) line_read_error(ret);
	
	if (command->multiple) {
	    if (item[0] == '+') {
	        /* the value needs to be added */
	        if (item[1] == '\0') {
	            /* only a plus! */
            	ret = file.read_entry(item, sizeof(item));
            	if (ret < 0) line_read_error(ret);
            } else {
                val++; /* now points to second char */
            }
        } else {
            if (!ent->isoverrided_p(res) && *resptr) {
                clear_setting(command, (intptr_t *)res);
                *resptr = NULL;
            }
        }
        if (item[0] == '-' && item[1] == 0) {
            /* delete the value list */
            if (!ent->isoverrided_p(res) && *resptr) {
                clear_setting(command, (intptr_t *)res);
                *resptr = NULL;
            }
            return 0;
        }
        if (item[strlen(item) - 1] == ',') {
            /* another value for the setting follows */
            item[strlen(item) - 1] = '\0';
            another_value = true;
        }
    }

    switch (command->datatype) {
        case DT_BOOL: /* boolean: can be 'true' or 'false' */
            if (stricmp(val, "true") == 0) intres = 1;
            else if (stricmp(val, "false") == 0) intres = 0;
            else if (stricmp(val, "yes") == 0) intres = 1;
            else if (stricmp(val, "no") == 0) intres = 0;
            else if (stricmp(val, "1") == 0) intres = 1;
            else if (stricmp(val, "0") == 0) intres = 0;
            else { log_print(LOG_LEVEL_ERROR, "%s:%d %s is an invalid boolean value!", file.name, file.linenum, val); return -1; }
            break;
        case DT_INT: /* numeric value */
            intres = txt_to_long(val);
            if (intres == -2) {
                log_print(LOG_LEVEL_ERROR, "%s:%d: %s is not a correct numeric value!", file.name, file.linenum, val);
                return -1;
            }
            break;
        case DT_ENUM: /* enumeration */
            curenum = command->enumerate;
            while (curenum->name[0]) {
                if (stricmp(val, curenum->name) == 0) { intres = curenum->value; break; }
                curenum++;
            }
            if (!curenum->name[0]) {
                log_print(LOG_LEVEL_ERROR, "%s:%d: %s is not a valid value for this option!", file.name, file.linenum, val);
                return -1;
            }
            break;
        case DT_STRING: /* character string */
            intres = (intptr_t)strdup(val);
            if (intres == 0) {
                log_print(LOG_LEVEL_ERROR, "%s:%d: Could not read string %s!", file.name, file.linenum, val);
                return -1;
            }
            break;
        case DT_IDENT: /* identifier: perhaps an ip number plus perhaps a port */
            tmp = strchr(val, '#');
            if (tmp != NULL) {
        		/* both have been given */
                *tmp++ = '\0';
                x = txt_to_long(tmp);
                tmp = val;
            } else {
                /* only one of 'em: got to find out which */
                x = txt_to_long(val);
                if (x < 0) {
                    /* doesn't quite seem like a port to me! */
                    tmp = val;
                    x = DNS_PORT;
                } else {
                    /* ah, no address */
                    getaddress((_addr *)yadb, "0.0.0.0", x);
                }
            }

            /* see if we have a port */
            if (x < 0 || x > 32768) {
                log_print(LOG_LEVEL_ERROR, "%s:%d: invalid port!", file.name, file.linenum);
                return -1;
            }

            /* if tmp is assigned, it points to an address */
            if (tmp && getaddress((_addr *)yadb, tmp, x) != 0) {
                log_print(LOG_LEVEL_ERROR, "%s:%d: %s is not a valid address!", file.name, file.linenum, tmp);
                return -1;
            }

            intres = (intptr_t)memdup(yadb, sizeof(_addr));
            break;
        case DT_IDENTPAIR: /* identifier pair: domain name and address */
            /* in config files, the syntax used is domain-name[@address] */

        	if (txt_to_ip(yadb, val) == 0) {
        		log_print(LOG_LEVEL_WARNING, "%s:%d: Domain name expected, "
        			"but this looks like an IP number: %s",
        			file.name, file.linenum, val);
        	}

            tmp = strchr(val, '@');
            if (tmp == NULL) { /* apparently, no address: only a domain name given */
                if (txt_to_dname(yadb, val) >= 0) {                     /* the domain name */
                    yadb[domlen(yadb)] = '\0';                          /* address type: 0=none */
                    intres = (intptr_t)memdup(yadb, domlen(yadb) + 1);
                } else {
                    log_print(LOG_LEVEL_ERROR, "%s:%d: %s is an invalid identifier pair!", file.name, file.linenum, val);
                    return -1;
                }
            } else {
                *tmp++ = '\0';
                if (txt_to_dname(yadb, val) >= 0 &&                     /* domain name */
                    txt_to_ip(&yadb[domlen(yadb) + 1], tmp) >= 0) {     /* ip number */
                    yadb[domlen(yadb)] = '\1';                          /* address type: 1=ipv4 */
                    intres = (intptr_t)memdup(yadb, domlen(yadb) + 5);
                } else {
                    *(tmp - 1) = '@';
                    log_print(LOG_LEVEL_ERROR, "%s:%d: %s is not a valid identifier pair!", file.name, file.linenum, val);
                    return -1;
                }
            }
            break;
    }

    if (command->multiple) {
        multi_add(resptr, intres);
    } else {
        clear_single_value(command->datatype, (intptr_t*)res);
        *res = intres;
    }

    ent->override_p(res);
    if (another_value) return handle_config_item(file, command, ent);
    return 0;
}

configuration_data_name std_config_structs[] = {
	{ &config, "" },
	{ &std_zone_config, "zone" },
	
	{ &std_auth_config, "auth" },
	{ &std_primary_config, "primary" },
	{ &std_secondary_config, "secondary" },
	
	{ &std_hint_config, "hint" },
	{ &std_tunnel_config, "tunnel" },
};

int n_std_config_structs = sizeof(std_config_structs) / sizeof(configuration_data_name);

int load_config_file(char *f_name) {
	char fname[MAXFILE];
    char item[64], *cpt;
    posadis_infile file;
    setting *ptr;
    char *cptr;
    int ret, x;
    bool already_found;

    strcpy(fname, conf_configdir); strcat(fname, f_name);
    if (!file.open(fname)) {
        log_print(LOG_LEVEL_PANIC, "Could not open %s!", f_name);
        return -30;
    }

    while (true) {
        file.start_line();
		if (file.is_eof()) break;
  		ret = file.read_entry(item, sizeof(item));
  		if (ret < 0) line_read_error(ret);

		if (item[0] == '\0') {
			/* empty line */
			continue;
		}

  		if (strcmpi(item, "zone") == 0) {
  		    ret = read_zone_line(file);
  		    if (ret < 0) return ret;  			
		} else if (strcmpi(item, "group") == 0) {
		    ret = read_group_line(file);
		    if (ret < 0) return ret;
		} else {
            if ( (cpt=strchr(item, '.')) != NULL) {
            	/* item contains a dot: maybe it's a zone.zone_ttl like setting */
    		  	*cpt++ = '\0'; /* setting name */
    		  	for (x = 0; x < n_std_config_structs; x++) {
    		  		if (strcmp(item, std_config_structs[x].name) == 0) {
    		  			ptr = std_config_structs[x].addr->getcommandinfo(cpt);
    		            if (ptr) {
        	        	    ret = handle_config_item(file, ptr, std_config_structs[x].addr);
            		        if (ret < 0) return ret;
    		            }
    		  			break;
    		  		}
    		  	}
    		  	if (x != n_std_config_structs) continue;
    		}

  		  	already_found = false; /* if set to true, the name is ambiguous */
            for (x = 0; x < n_std_config_structs; x++) {
                ptr = std_config_structs[x].addr->getcommandinfo(item);
                if (ptr) {
                	if (already_found) {
                		log_print(LOG_LEVEL_WARNING, "%s:%d: Command %s "
                			"is ambiguous!", file.name, file.linenum, item);
                		break;
                	}
                	already_found = true;
                    ret = handle_config_item(file, ptr, std_config_structs[x].addr);
       		        if (ret < 0) return ret;
                    break;
                }
            }
  		
            if (x == n_std_config_structs) {
                log_print(LOG_LEVEL_ERROR, "%s:%d: Unknown command %s!", file.name, file.linenum, item);
                return -30;
            }
        }
    }

    file.close();
    return 0;
}
