/*
    Posadis - A DNS Server
    Configuration data
    Copyright (C) 2001  Meilof Veeningen <meilof@users.sourceforge.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef __CONFIGURATION_H
#define __CONFIGURATION_H

#include <include/unifile.h>
#include "bitlist.h"
#include "lexfn.h"
#include "syssocket.h"
#include "vector.h"

#ifdef __BORLANDC__
/* borland c++ doesn't define intptr_t */
typedef int intptr_t;
#else
# ifdef _WIN32
#  include <stdint.h>
# else
#  include <inttypes.h>
#endif
#endif

typedef char* multi_val;

int process_cmdline_args(int argc, char **argv);     // loads the command-line arguments
int determine_datadir();

/* global settings */
extern char conf_configdir[PATH_MAX];       // the directory with dns config files
#ifndef _WIN32
extern bool do_fork;						// whether to behave like a true server by fork()ing
#else
extern bool auto_minimize; /* whether to minimize the win32 log window if no errors */
#endif

/* the structures which provide a universal means of interpreting configuration settings */


/* to add a configuration parameter, one has to take the following steps:

 (*) decide what kind of setting it is: general/zone/master/slave/primary
 (*) locate the appropriate _configuration_data class in configuration.h and add a variable which
     will hold your setting: an intptr_t for DT_BOOL, DT_INT and DT_ENUM, a char* for the others
 (*) locate the constructor of the _configuration_data class you just found, the former should reside
     in configuration.cpp. among the other hard-coded defaults, specify the default for your setting.
 (*) locate the appropriate _settings[] array in configuration.cpp, and add your setting to it.
     IMPORTANT: the sequence of settings in the _settings array and the _configuration_data class
                *must* be the same!
                additionally, if your new setting will be the first in the settings list, change the
                call from the *_configuration_data constructor to the configuration_data constructor
                to give the address of your setting as the second argument.

 afterwards, your setting should work! if you did something wrong, you will find out soon enough
 because posadis will probably segfault immediately. good luck...
*/

enum _conf_datatype { DT_BOOL, DT_INT, DT_ENUM, DT_STRING, DT_IDENT, DT_IDENTPAIR };

struct enumeration {
    int value;
    char name[18];
};

typedef struct _setting {
    char command[32];           /* the name of the configuration command (case insensitive) */
    _conf_datatype datatype;    /* the datatype of the item */
    bool multiple;              /* whether multiple values can be given to the setting */
    enumeration *enumerate;     /* if (datatype == DT_ENUM), this specifies the possible values */
} setting;

struct conf_settings {
    setting *settings;              /* a pointer to the settings */
    int n_settings;                 /* the number of settings */
};


void clear_single_value(_conf_datatype datatype, intptr_t *val); // internal function called by configuration_datadestructor
void clear_setting(setting *setting, intptr_t *val); // frees all allocated data for a setting

void clone_single_value(_conf_datatype datatype, intptr_t *src, intptr_t *res); // internal function called by configuration_data: clone_setting
void clone_setting(setting *setting, intptr_t *src, intptr_t *res); // internal function called by configuration_data: clone and inherit

/* functions for multi-vals */
int multi_get_count(multi_val val);
void multi_add(multi_val* val, intptr_t value);
void multi_remove(multi_val* val, int ix, _conf_datatype datatype);
char* multi_getptr(multi_val val, int ix);
intptr_t multi_getval(multi_val val, int ix);

multi_val multi_clone(multi_val *setting, _conf_datatype type);
void multi_free(multi_val* val, _conf_datatype datatype);

/* functions for setttings */
int addaddressbyident(Vector<_addr *> *resvec, char *ptr);
bool is_in_identlist(multi_val indentlist, _addr *addr);

class configuration_data {
  protected:
    configuration_data(conf_settings *info, intptr_t *_begin);
    ~configuration_data();

  public:

    void clone(configuration_data *item);   // loads configuration from that specific item (overrided bit off course not set)
    void inherit(configuration_data *item); // inherit only settings that are overrided in the source class

    conf_settings *info;                    // the commands of the setting class
    intptr_t *begin;                        // begin of the data section

    BitList *overrided;                     // specifies which settings are overrided
    void override_c(char *optname);         // overrides an option by name
    void override_p(void *setting);         // overrides an option by the addess of the storage
    void unoverride_c(char *optname);       // un-overrides an option by name
    void unoverride_p(void *setting);       // un-overrides an option by the address of the storage
    bool isoverrided_c(char *optname);      // checks whether the option is overrided by name
    bool isoverrided_p(void *setting);      // checks whether the option is overrided by storage address

    void *getpropaddr(char *optname);       // retrieves the address of an option
    setting *getcommandinfo(char *optname, conf_settings *cinfo = NULL); // retrieves the information on an option

};

/* global configuration */

class main_configuration_data : public configuration_data {
  public:
    main_configuration_data();              // constructor; initializes to hard-coded defaults

    multi_val udp_listen;                   // udp devices to listen to
    multi_val tcp_listen;                   // tcp devices to listen to
    intptr_t ipv6_client;                   // whether to act as an IPv6 client
    intptr_t max_cname;                     // the maximal recursion level of cnames
    intptr_t do_recursion;                  // controls whether we support recursion
    intptr_t max_threads;                   // the maximum number of simultaneous threads
    char *datadir;							// the directory for log files and xfr history
    intptr_t tcp_in_keepalive;				// the number of seconds an incoming tcp connection is kept for reuse
    intptr_t tcp_out_keepalive;				// the number of seconds in outgoing tcp connection is kept for reusr
    intptr_t patience;						// the number of seconds to wait for an answer to a query
    intptr_t tcp_patience;                  // tcp patience value
    intptr_t do_query_logging;              // whether to log all incoming queries
    intptr_t hide_dname;                    // whether to hide DNAMEs
#ifndef _WIN32
    char *user;                             // specify the user to switch to
    char *group;                            // specify the group to switch to
#endif
};

extern conf_settings main_config;
extern main_configuration_data config;

/* zone configuration */

class zone_configuration_data : public configuration_data {
  public:
    zone_configuration_data(bool is_std = false); // set is_std true to use hard_coded defaults, false to use the defaults from std_zone_config

    multi_val auth_servers;                 // authoritive nameserver list for axfr
};

extern conf_settings zone_config;
extern zone_configuration_data std_zone_config;

/* authoritive configuration (axfr providing) */

#define XFR_TO_NONE 0
#define XFR_TO_SLAVES 1
#define XFR_TO_ANY 2

#define NOT_TO_NONE 0
#define NOT_TO_SLAVES 1
#define NOT_TO_ANY 2

class auth_configuration_data : public configuration_data {
  public:
 	auth_configuration_data(bool is_std = false);

    intptr_t xfr_to;                        // the type of client {none, ns, all} we allow xfr to by default
    multi_val also_xfr_to;                  // additional clients to allow xfr

    multi_val slaves;                       // slave nameservers

    intptr_t notify_to_slaves;              // whether to notify slaves
    multi_val also_notify;                  // additional peers
    intptr_t notify_interval;               // interval between messages
    intptr_t notify_retry;                  // number of times to retry
};

extern conf_settings auth_config;
extern auth_configuration_data std_auth_config;

/* primary configuration (master files) */

class primary_configuration_data : public configuration_data {
  public:
    primary_configuration_data(bool is_std = false); // set is_std to true to use hard-coded defaults, false to use defaults from std_primary_config
    char *file;                             // the location of the master file
    intptr_t zone_ttl;                      // the time to live before the zone is re-loaded from master file
};

extern conf_settings primary_config;
extern primary_configuration_data std_primary_config;

/* secondary configuration (ixfr from) */

struct secondary_configuration_data : public configuration_data {
  public:
    secondary_configuration_data(bool is_std = false); // set is_std to true to use hard-coded defaults, false to use defaults from std_slave_config

    multi_val masters;						// static ns list
    intptr_t initial_retry_interval;		// initial retry interval
};

extern conf_settings secondary_config;
extern secondary_configuration_data std_secondary_config;


/* hint configuration (what hints to give...) */

struct hint_configuration_data : public configuration_data {
  public:
    hint_configuration_data(bool is_std = false); // set is_std to true to use hard-coded defaults
	multi_val nameserver;                   // the nameservers we send back as a hint (can work complementary to 'file')
    intptr_t ns_ttl;                        // the ttl to send to clients
    multi_val source;                       // the source to update from	
    intptr_t pre_expire;                    // time before expire to begin updating
};

extern conf_settings hint_config;
extern hint_configuration_data std_hint_config;

/* tunnel configuration data (to which nameserver to tunnel...) */


/* tunnel::recursive_answer_fail enumeration */
#define T_RF_NODATA 1
#define T_RF_NXDOMAIN 2
#define T_RF_HINT 3
#define T_RF_NONE 4

struct tunnel_configuration_data : public configuration_data {
  public:
  	tunnel_configuration_data(bool is_std = false); // set is_std to true to use hard-coded defaults
  	multi_val tunnelserver;                 // the nameserver(s) to tunnel to
  	intptr_t recursive_answer_fail;			// what to do if the nameserver doesn't respond
  	intptr_t retry;							// the number of times to retry
    intptr_t no_aa;                         // set to true to suppress the aa bit
};


extern conf_settings tunnel_config;
extern tunnel_configuration_data std_tunnel_config;



/* function which reads the arguments for a configuration setting, and stores them at the right place */
int handle_config_item(posadis_infile& file, setting *command, configuration_data *ent);
int load_config_file(char *file);


/* list of default setting classes */
typedef struct {
	configuration_data *addr;
	char name[10];
} configuration_data_name;

extern configuration_data_name std_config_structs[];
extern int n_std_config_structs;

#endif /* ifndef __CONFIGURATION_H */

