/*
    Posadis - A DNS Server
    Hint Zones
    Copyright (C) 2001  Meilof Veeningen <meilof@users.sourceforge.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <include/unistring.h>

#include "hint.h"
#include "classes.h"
#include "threads.h"
#include "log.h"
#include "posadis.h"

/* hint zone group */

HintGroup::HintGroup() { type = ZONE_HINT; }

/* hint zone class */

HintZone::HintZone(_domain _dom) : NotauthZone(_dom) {
	type = ZONE_HINT;
    datasrc = SRC_NONE;
}

HintZone::~HintZone() {
}

void HintZone::set_hardcoded() {
    int ret, tmp, x;
    CacheItem *newitem;
    Zone *zn;
    char *ptr, *domptr;
    char tmpb[256];

    if (datasrc == SRC_GLUE) return;
	
    delete citem_root;
	citem_root = new CacheItem("");

    if (hint_config.nameserver) {
        /* number of nameservers */
        ret = multi_get_count(hint_config.nameserver);

        for (x = 0; x < ret; x++) {
            ptr = multi_getptr(hint_config.nameserver, x);

            domptr = ptr;

            /* read the domain name of the nameserver */
            citem_root->write_rrset(DNS_TYPE_NS)->add_rr(false,
            					 hint_config.ns_ttl, domlen(ptr), ptr);
                					
            /* and (if set) the address associated with the domain name */
            ptr += domlen(ptr);
            tmp = *(ptr++); /* the address type: 0=none 1=ip4 */
            switch (tmp) {
                case 0: continue;
                case 1:
                	/* at first check whether it's in our zone:
                	   else, it's unethical to set the address */
                    LookupZone(zn, domptr);
                	if (zn != this) {
                		domtostring(tmpb, domptr);
                		log_print(LOG_LEVEL_WARNING, "The nameserver %s is outside the hint zone!", tmpb);
                		continue;
                	}
                	if (create_domain(newitem, domptr) < 0) {
                        domtostring(tmpb, domptr);
                        log_print(LOG_LEVEL_WARNING, "The address for the %s nameserver of the hint zone could not be set!", tmpb);
                    } else {
                        if (newitem->write_rrset(DNS_TYPE_A)->items.count == 0)
                            newitem->write_rrset(DNS_TYPE_A)->add_rr
                            	(false, hint_config.ns_ttl, 4, ptr);
                                	
                        ptr += 4;
                    }
                    break;
                default:
                    log_print(LOG_LEVEL_ERROR, "Internal inconsistency: unknown address type %d in hint 'nameserver' property!", *(ptr-1));
            }
        }

    }
}

void HintZone::start_updating() {
    int ret, x;
    char *ptr;
    char tmpb[256];
    Vector<_addr *> *adds;
	
#ifdef _EXTDEBUG_
	domtostring(tmpb, domain);
	log_print(LOG_LEVEL_INFO, "Starting HINT update process for %s", tmpb);
#endif
    
    next_update = curtime + 120; /* default value */

   	/* prevent a zone from being updated by multiple threads */
    if (!is_being_updated) {
        zone_is_loaded = false;
        set_hardcoded();

        if (hint_config.source) {
            adds = new Vector<_addr *>();
            ret = multi_get_count(hint_config.source);
            for (x = 0; x < ret; x++) {
                ptr = multi_getptr(hint_config.source, x);
                if (addaddressbyident(adds, ptr) < 0) {
                    domtostring(tmpb, domain);
                    log_print(LOG_LEVEL_ERROR, "Error in sources list for hint zone %s!", tmpb);
                    delete adds;
                    return;
                }
            }
            threads.add(new HintUpdateThread(this, adds));
        }

        if (citem_root->get_rrset(DNS_TYPE_NS)) zone_is_loaded = true;

        if (hint_config.ns_ttl > 0)
            next_update = curtime + (long)(hint_config.ns_ttl * 0.95);
        else
            next_update = 0;

    }
}

void HintZone::offer_glue(CacheItem *item) {
    if (datasrc <= SRC_GLUE) {
        /* hey, what the heck: just copy it! */
        delete citem_root;
        citem_root = item->clone();
        free(citem_root->name);
        citem_root->name = strdup("");
        datasrc = SRC_GLUE;

        if (!hint_config.source) next_update = 0;
    }
}


_thread_answer HintZone::do_lookup_thread(AnsweringThread*& thr, stdquery_info *info, _domain QNAME) {
    info->answer->RA = false;

    if (info->QTYPE == DNS_TYPE_NS && domcmp(QNAME, domain))
    	/* they want to know the ns rrs, so we'll give them */
        info->answer->add_rrset_qtype(domain, citem_root, DNS_TYPE_NS, true, S_ANSWER);
    else
	/* add the NS rrs to the additional section */    	
    	info->answer->add_rrset_qtype(domain, citem_root, DNS_TYPE_NS, true, S_AUTHORITY);

    return TA_SUCCESS;
}


/* the hint update thread */

HintUpdateThread::HintUpdateThread(HintZone *_zone, Vector<_addr *> *_servers) {
    child = NULL;
    servers = _servers;
    cur = -1;
    state = STATE_INIT;
    zone = _zone;
    q = new DnsMessage();
    zone->is_being_updated = true;
}

HintUpdateThread::~HintUpdateThread() {
    delete servers;
    if (child) delete child;
    if (q) delete q;
}

int HintUpdateThread::add_to_cache(_dm_section *sec) {
    int x;
    Zone *zn;
    CacheItem *citem;
    complete_rr *cur;
    rr_set *set;

    for (x = 0; x < sec->count; x++) {
        cur = (*sec)[x];
        LookupZone(zn, cur->NAME);
        if (zn != zone) continue; /* not interested */

        if (zone->create_domain(citem, cur->NAME) < 0) return -1;

        set = citem->write_rrset(cur->TYPE);
        if (!set) return -1;
        if (!set->add_rr(false, cur)) return -1;
    }

    return 0;

}

int HintUpdateThread::execute() {
    int ret = 0;
    rr_set *set;
    CacheItem *backup = NULL;
    char *tmpstr;

    if (child) {
        ret = child->execute();
        if (ret == TRET_OKBUSY) return TRET_OKBUSY;
    }

    if (state == STATE_INIT) {
        if (++cur >= servers->count) {
            /* still nothing */
            zone->set_hardcoded();
            zone->is_being_updated = false;
            return TRET_ERROR;
        }

        q->reinit();
        q->ID = new_request_id();
        q->question.add(new question_item(zone->domain, DNS_TYPE_NS));
		
        child = new MessageOutThreadUDP(-1, *(*servers)[cur], q);
        twait = new ThreadWait(this, q->ID, *(*servers)[cur], ((MessageOutThreadUDP *)child)->sockid);
        answer_waiting_threads.add(twait);

        state = STATE_SEND;
        return TRET_OKBUSY;
    }

    if (state == STATE_SEND) {
        /* finally, sent! */
        delete child; child = NULL;
        if (ret == TRET_ERROR) {

            tmpstr = addr_to_string((*servers)[cur]);
            log_print(LOG_LEVEL_ERROR, "Failed to connect to %s for hint update", tmpstr);
            free(tmpstr);
          
            answer_waiting_threads.removeitem(twait);
            state = STATE_INIT;
        } else {
            wakeup_time = curtime + config.patience + 1;
            trstate = THREAD_SLEEP;
            state = STATE_RECV;
        }
        return TRET_OKBUSY;
    }

    if (state == STATE_RECV) {
        trstate = THREAD_RUNNING;
        if (twait->received) {
			
            if (!twait->msg->is_answer_on(q) ||
                twait->msg->RCODE != RCODE_NOERROR)
                goto _state_rcv_err;

            backup = zone->citem_root;
            zone->citem_root = new CacheItem("");

            /* add it to our cache */
            if (add_to_cache(&twait->msg->answer) < 0 ||
                add_to_cache(&twait->msg->authority) < 0 ||
                add_to_cache(&twait->msg->additional) < 0)
                goto _state_rcv_err;

            set = zone->citem_root->get_rrset(DNS_TYPE_NS);

            if (set == NULL) goto _state_rcv_err;

            if (set->expire <= curtime + zone->hint_config.pre_expire + 5) goto _state_rcv_err;

            /* well, we're pretty satisfied */
            answer_waiting_threads.removeitem(twait);
            zone->zone_is_loaded = true;

            zone->next_update = set->expire - zone->hint_config.pre_expire;
            delete backup;
            zone->is_being_updated = false;
            return TRET_CLOSED;

        }

        /* no answer: try next */
        tmpstr = addr_to_string((*servers)[cur]);
        log_print(LOG_LEVEL_ERROR, "No answer from %s in hint update", tmpstr);
        free(tmpstr);
_state_rcv_err:
        delete zone->citem_root;
        zone->citem_root = backup;
        zone->zone_is_loaded = false;
        state = STATE_INIT;
        answer_waiting_threads.removeitem(twait);
        return TRET_OKBUSY;
    }

    zone->is_being_updated = false;
    return TRET_ERROR;
}
