/*
    Posadis - A DNS Server
    Loads and dumps DNS Master files
    Copyright (C) 2001  Meilof Veeningen <meilof@users.sourceforge.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "dns.h"
#include "include/unifile.h"
#include "include/unistring.h"

char pow2[8] = { 128, 64, 32, 16, 8, 4, 2, 1 };

#define line_read_error(errcode)  {											\
	cptr = file.geterrmsg(); 												\
	log_print(LOG_LEVEL_ERROR, "%s:%d: %s", file.name, file.linenum, cptr);	\
	free(cptr);																\
	return -1;																\
	}

int load_master_file(Zone *zone, char *name, char *fname, bool is_cache, bool recursive, int entryix, char *p_origin, int _ttl) {
    posadis_infile file; /* the input file object */
    char completefile[PATH_MAX]; /* the file name */
    char origin[255]; /* the current origin (changable by $origin directives) */
    int default_ttl, ttl;
    int ret, x, curprop, pos, y;
    char tmpb[256], tmp2[256], yadb[1024]; /* temporary buffers */
    char dname[255], prev_dname[255]; /* the current and previous dnames */
    rr_type *rrinfo;
    char *cptr;
    CacheItem *citem;
    rr_set *rrset;

    if (strlen(conf_configdir) + strlen(fname) + 1 > PATH_MAX) {
        log_print(LOG_LEVEL_ERROR, "File name for master file %s too long!", fname);
		return -2;
    }

    strcpy(completefile, conf_configdir);
    strcat(completefile, fname);

    if (!file.open(completefile)) {
        log_print(LOG_LEVEL_ERROR, "Could not open %s (file not found or recursive include)", file.name);
        return -1;
    }

    if (p_origin == NULL) {
        memcpy(origin, name, domlen(name));
    } else {
        memcpy(origin, p_origin, domlen(p_origin));
    }
    default_ttl = _ttl;

    while (true) {
        file.start_line();
		if (file.is_eof()) break;
  		ret = file.read_entry(tmpb, sizeof(tmpb));
  		if (ret < 0) line_read_error(ret);

  		/* check for empty line */
		if (tmpb[0] == '\0' && file.is_eol()) continue;

        if (tmpb[0] == '$') { // we might have a directive
            if (strcmpi(tmpb, "$include") == 0) {
                /* the include directive - includes another file
                   usage: $include file [origin-for-file] */

            	if (recursive) {
            		log_print(LOG_LEVEL_ERROR, "%s:%d: Cannot handle recursive include", file.name, file.linenum);
            		return -1;
            	}
            	
            	/* read the file name */
            	ret = file.read_entry(tmpb, sizeof(tmpb));
            	if (ret < 0) line_read_error(ret);            	
                if (strchr(tmpb, '\\') != NULL || strchr(tmpb, '/') != NULL) {
                    log_print(LOG_LEVEL_ERROR, "%s:%d: includes may not be in other directories!", file.name, file.linenum);
                    return -1;
                }

                strcpy(yadb, conf_configdir);
                strcat(yadb, tmpb);

                /* read the origin name if it is given */
                if (!file.is_eol()) {
	            	ret = file.read_entry(tmp2, sizeof(tmp2));
					if (ret < 0) line_read_error(ret);
					ret = txt_to_dname(dname, tmp2, origin);
					if (ret < 0) {
						log_print(LOG_LEVEL_ERROR, "%s:%d: Invalid domain name in $include: %s", file.name, file.linenum, tmpb);
						return -1;
					}
				} else {
                    strcpy(dname, origin);
                }

                /* read the child masterfile */
				ret = load_master_file(zone, name, yadb, is_cache, true, entryix, dname, default_ttl);
				if (ret < 0) return ret;
                continue;
            }

            if (strcmpi(tmpb, "$origin") == 0) {
                /* the origin directive - sets the origin for relative dnames
                   usage: $origin dname */

                /* read the domain name */
                ret = file.read_entry(tmpb, sizeof(tmpb));
                if (ret < 0) line_read_error(ret);

                ret = txt_to_dname(dname, tmpb, name);
   				if (ret < 0) {
   					log_print(LOG_LEVEL_ERROR, "%s:%d: Invalid domain name in $origin: %s", file.name, file.linenum, tmpb);
   					return -1;
   				}
   				
				strcpy(origin, dname);
                continue;
            }

            if (strcmpi(tmpb, "$ttl") == 0) {
            	/* the $ttl directive - sets the default ttl for lazy rrs
            	   usage: $ttl std_ttl */
            	
            	ret = file.read_entry(tmpb, sizeof(tmpb));
            	if (ret < 0) line_read_error(ret);
            	
                ret = txt_to_long(tmpb);
                if (ret < 0)
                    log_print(LOG_LEVEL_ERROR, "%s:%d: Incorrect numeric value given for the $ttl directive!", file.name, file.linenum);
                else
                    default_ttl = ret;
                continue;
            }
        }

        ttl = default_ttl;

        /* at first, check whether the domain name is specified */
        if (tmpb[0] == '\0') {
        	/* no dname specified -> use previous */
        	if (entryix == -1) { /* first item, so no previous available */
                log_print(LOG_LEVEL_ERROR, "%s:%d: First item should define a domain name!", file.name, file.linenum);
                return -1;
            }
            memcpy(dname, prev_dname, domlen(prev_dname));
        } else {
        	/* domain name specified */
        	ret = txt_to_dname(dname, tmpb, origin);
        	if (ret < 0) {
                log_print(LOG_LEVEL_ERROR, "%s:%d: The domain name %s is invalid!", file.name, file.linenum, tmpb);
        		return -1;
          	}
		}

        memcpy(prev_dname, dname, sizeof(dname));

        /* the order of the next items:
        	[ttl] [class] type rdata
        	[class] [ttl] type rdata */

        /* read the next code */
        ret = file.read_entry(tmpb, sizeof(tmpb));
        if (ret < 0) line_read_error(ret);

        /* at first, try to extract a numeric value out of it. if this
           succeeds, it's a TTL */
        ttl = txt_to_long(tmpb);
        if (ttl < 0) {
        	/* apparently no valid ttl */
        	ttl = default_ttl;
        } else {
        	/* valid ttl */
        	ret = file.read_entry(tmpb, sizeof(tmpb));
        	if (ret < 0) line_read_error(ret);
        }

        /* now we have a code: a class [IN|HS|CS|CS] or rr type */
		if (stricmp(tmpb, "CS") == 0 || stricmp(tmpb, "HS") == 0 ||
			stricmp(tmpb, "CH") == 0) {
			log_print(LOG_LEVEL_ERROR, "%s:%d: Posadis doesn't support the "
					  "%s class, only IN!", file.name, file.linenum, tmpb);
			return -1;
		}
		
		if (stricmp(tmpb, "IN") == 0) {
			/* it is the one supported class */
			ret = file.read_entry(tmpb, sizeof(tmpb));
			if (ret < 0) line_read_error(ret);
			
			/* after the class, we still might have a ttl */
			if (ttl == default_ttl) {
				ttl = txt_to_long(tmpb);
				if (ttl >= 0) {
					/* valid ttl! */
					ret = file.read_entry(tmpb, sizeof(tmpb));
					if (ret < 0) line_read_error(ret);
				} else {
					ttl = default_ttl;
				}
			}
		}
		
		/* now we have a rr type */
        if (!domisparent(name, dname)) {
            domtostring(tmpb, dname);
            log_print(LOG_LEVEL_ERROR, "%s:%d: %s is outside the zone!", file.name, file.linenum, tmpb);
            return -1;
        }

        entryix++;
        if (entryix == 0 && stricmp(tmpb, "SOA") != 0) {
            log_print(LOG_LEVEL_ERROR, "%s:%d: First line must define a SOA!", file.name, file.linenum);
            return -1;
        }

        if (entryix != 0 && stricmp(tmpb, "SOA") == 0) {
        	log_print(LOG_LEVEL_ERROR, "%s:%d: Only first RR in master file may be SOA!", file.name, file.linenum);
        	return -1;
        }

        rrinfo = rrtype_getinfo(tmpb);
        if (!rrinfo) {
            log_print(LOG_LEVEL_ERROR, "%s:%d: RR type %s not supported", file.name, file.linenum, tmpb);
            return -1;
        }	

        /* try to handle the rr data and compile it into yadb */
        curprop = -1; pos = 0;
        while (rrinfo->properties[++curprop] != '\0') {
        	/* load the data for the argument */
        	ret = file.read_entry(tmpb, sizeof(tmpb));
        	if (ret < 0) line_read_error(ret);
            /* see what type of argument we expect */
            switch (rrinfo->properties[curprop]) {
                case 'm':
                  ret = txt_to_email(tmp2, tmpb, origin);
                  if (ret < 0) {
                    log_print(LOG_LEVEL_ERROR, "%s:%d: %s is not a valid email address!", file.name, file.linenum, tmpb);
                    return -1;
                  }
                  if (pos + ret >= (signed)sizeof(yadb)) {
                    log_print(LOG_LEVEL_ERROR, "%s:%d: Buffer overflow", file.name, file.linenum);
                  }
                  memcpy(&yadb[pos], tmp2, ret);
                  pos += ret;
                  break;
                case 'd': /* a domain name */
                	ret = txt_to_dname(tmp2, tmpb, origin);
                	if (ret < 0) {
                        log_print(LOG_LEVEL_ERROR, "%s:%d: %s is an invalid domain name!", file.name, file.linenum, tmpb);
                        return -1;
                    }
                    if (pos + ret >= (signed)sizeof(yadb)) {
                    	log_print(LOG_LEVEL_ERROR, "%s:%d: Buffer overflow", file.name, file.linenum);
                    	return -1;
                    }
                    memcpy(&yadb[pos], tmp2, ret);
                    pos += ret;
                    break;
                case 'i': /* ip number */
                	ret = txt_to_ip(tmp2, tmpb);
                	if (ret < 0) {
                		log_print(LOG_LEVEL_ERROR, "%s:%d: %s is an invalid ip number!", file.name, file.linenum, tmpb);
                		return -1;
                	}
                	if (pos + 4 >= (signed)sizeof(yadb)) {
                    	log_print(LOG_LEVEL_ERROR, "%s:%d: Buffer overflow", file.name, file.linenum);
                    	return -1;
                 	}
                 	memcpy(&yadb[pos], tmp2, 4);
                 	pos += 4;
                    break;
                case '6': /* ipv6 */
                	ret = txt_to_ipv6(tmp2, tmpb);
                	if (ret < 0) {
                		log_print(LOG_LEVEL_ERROR, "%s:%d: %s is an invalid ipv6 number!", file.name, file.linenum, tmpb);
                		return -1;
                	}
                	if (pos + 16 >= (signed)sizeof(yadb)) {
                		log_print(LOG_LEVEL_ERROR, "%s:%d: Buffer overflow", file.name, file.linenum);
                		return -1;
                	}
                	memcpy(&yadb[pos], tmp2, 16);
                	pos += 16;
                	break;
                case '7': /* A6 RR */
                    /* read prefix length */
                    ret = txt_to_long(tmpb);
                    if (ret < 0 || ret > 128) {
                        log_print(LOG_LEVEL_ERROR, "%s:%d: Incorrect IPv6 prefix length %d!", file.name, file.linenum, tmpb);
                        return -1;
                    }
                    /* read ip addr */                        
                    if ((y = file.read_entry(tmpb, sizeof(tmpb))) < 0) line_read_error(y);
                    if (txt_to_ipv6(tmp2, tmpb) < 0) {
                        log_print(LOG_LEVEL_ERROR, "%s:%d: %s is an invalid ipv6 number!", file.name, file.linenum, tmpb);
                        return -1;
                    }
                    if (pos + ((135 - ret) / 8) + 1 >= (signed)sizeof(yadb)) {
                    	log_print(LOG_LEVEL_ERROR, "%s:%d: Buffer overflow", file.name, file.linenum);
                    	return -1;
                    }
                    yadb[pos++] = ret; 
                    memcpy(&yadb[pos], tmp2 + 16 - ((135 - ret) / 8), ((135 - ret) / 8));
                    pos += ((135 - ret) / 8);
                    /* read dname if nessecary */
                    if (ret != 0) {
                        if ((y = file.read_entry(tmpb, sizeof(tmpb))) < 0) line_read_error(y);
                        ret = txt_to_dname(tmp2, tmpb, origin);
                        if (ret < 0) {
                            log_print(LOG_LEVEL_ERROR, "%s:%d: %s is an invalid domain name!", file.name, file.linenum, tmpb);
                            return -1;
                        }
                        if (pos + ret >= (signed)sizeof(yadb)) {
                            log_print(LOG_LEVEL_ERROR, "%s:%d: Buffer overflow", file.name, file.linenum);
                            return -1;
                        }
                        memcpy(&yadb[pos], tmp2, ret);
                        pos += ret;     
                    }
                    break;
                case 's': /* short (16-bit) */
                	ret = txt_to_long(tmpb);
                	if (ret < 0 || ret >= 65536) {
                		log_print(LOG_LEVEL_ERROR, "%s:%d: %s is an invalid number!", file.name, file.linenum, tmpb);
                		return -1;
                	}
                	if (pos + 2 >= (signed)sizeof(yadb)) {
                    	log_print(LOG_LEVEL_ERROR, "%s:%d: Buffer overflow", file.name, file.linenum);
                    	return -1;
                 	}
                 	yadb[pos++] = ret / 256;
                 	yadb[pos++] = ret;
                    break;
                case 'l': /* long (32-bit) */
                	ret = txt_to_long(tmpb);
                	if (ret < 0 || ret > 2147483647) {
                		log_print(LOG_LEVEL_ERROR, "%s:%d: %s is an invalid number!", file.name, file.linenum, tmpb);
                		return -1;
                	}
                	if (pos + 4 >= (signed)sizeof(yadb)) {
                    	log_print(LOG_LEVEL_ERROR, "%s:%d: Buffer overflow", file.name, file.linenum);
                    	return -1;
                 	}
                    yadb[pos++] = ret / 16777216;
                    yadb[pos++] = ret / 65536;
                    yadb[pos++] = ret / 256;
                    yadb[pos++] = ret;
                    break;
                case 'c': /* character string */
                	if (strlen(tmpb) > 255 || pos + strlen(tmpb) + 1 >= sizeof(yadb)) {
                    	log_print(LOG_LEVEL_ERROR, "%s:%d: Buffer overflow", file.name, file.linenum);
                    	return -1;
                 	}
                    yadb[pos++] = (char)strlen(tmpb);
                    strcpy(&yadb[pos], tmpb);
                    pos += strlen(tmpb);
                    break;
                case 'h': /* character strings */
					ret = 0;
					while (ret >= 0) {
						if (strlen(tmpb) > 255 || pos + strlen(tmpb) + 1 >= sizeof(yadb)) {
        	            	log_print(LOG_LEVEL_ERROR, "%s:%d: Buffer overflow", file.name, file.linenum);
	                    	return -1;
    	             	}
    	             	yadb[pos++] = (char)strlen(tmpb);
    	             	strcpy(&yadb[pos], tmpb);
    	             	pos += strlen(tmpb);
    	             	/* read a new entry */
    	             	if (file.is_eol()) { /* end of line reached */
    	             		ret = -1;
    	             	} else {
				        	ret = file.read_entry(tmpb, sizeof(tmpb));
				        	if (ret < 0) line_read_error(ret);
				  		}
				  	}
                    break;
                case 'f': /* nothing particular */
                	log_print(LOG_LEVEL_ERROR, "%s:%d: NULL RDATA may not be set in a master file!", file.name, file.linenum);
                	return -1;
                case 'w': /* well-known service list (including protocol name) */
                	if ((ret = getprotocolbyname(tmpb)) < 0) {
                		log_print(LOG_LEVEL_ERROR, "%s:%d: Unknown protocol in WKS: %s", file.name, file.linenum, tmpb);
                		return -1;
                	}
                	yadb[pos++] = (char) ret;
                	
                	x = 0; /* the highest port number */
                	memset(tmp2, 0, sizeof(tmp2));
                	while (true) {
                		if (file.is_eol()) break;
                		ret = file.read_entry(tmpb, sizeof(tmpb));
                		if (ret < 0) line_read_error(ret);
                		/* at first, see if it's numeric */
                		ret = txt_to_long(tmpb);
                		if (ret < 0) {
                			/* ain't no number */
	                		if ((ret = getserviceportbyname(tmpb)) == -1) {
	                			/* no service name either! */
	                			log_print(LOG_LEVEL_WARNING, "%s:%d: Unknown service in WKS ignored: %s", file.name, file.linenum, tmpb);
	                			continue;
	                		}
	                	}
	                	if (ret >= 8 * (signed)sizeof(tmp2)) {
	                		log_print(LOG_LEVEL_WARNING, "%s:%d: Too high server port ignored: %d", file.name, file.linenum, ret);
	                		continue;
	                	}
	                	tmp2[ret / 8] |= pow2[ret % 8];
	                	if (ret > x) x = ret;
	                }
	                /* now write the result to the buffer */
	                if (pos + (x / 8) >= (signed)sizeof(yadb)) {
       	            	log_print(LOG_LEVEL_ERROR, "%s:%d: Buffer overflow", file.name, file.linenum);
                    	return -1;
   	             	}
   	             	memcpy(&yadb[pos], tmp2, x / 8 + 1);
   	             	pos += (x / 8 + 1);
   	             	break;
   	             default:
   	             	log_print(LOG_LEVEL_ERROR, "%s:%d: Internal error: Unknown datatype!", file.name, file.linenum);
   	             	return -1;
            }
        }

        if (rrinfo->type == DNS_TYPE_CNAME) {
        	/* canonical name domain */
        	ret = CreateDomain(citem, zone, dname);
        	if (!(ret & LOOKUP_CREATED)) {
        		domtostring(tmpb, dname);
        		log_print(LOG_LEVEL_ERROR, "%s:%d: Could not create CNAME domain %s (already exists?)", file.name, file.linenum, tmpb);
        		return -1;
        	}
        	
        	rrset = new rr_set(DNS_TYPE_CNAME);
        	rrset->add_rr(!is_cache, ttl, pos, yadb);
        	citem->rr_sets.add(rrset);
        	continue;
       	}

        if (rrinfo->type == DNS_TYPE_SOA) {
        	/* Start Of Authority RR */
            if (!domcmp(dname, name)) {
                domtostring(tmpb, name);
                log_print(LOG_LEVEL_ERROR, "%s:%d: This file SOA must be %s!", file.name, file.linenum, tmpb);
                return -1;
            }

            if (ttl == -1) {
                log_print(LOG_LEVEL_ERROR, "%s:%d: No TTL in either SOA record or $ttl directive!", file.name, file.linenum);
                return -1;
            }

            if (default_ttl == -1) default_ttl = ttl;

            rrset = new rr_set(DNS_TYPE_SOA);
            rrset->add_rr(!is_cache, ttl, pos, yadb);
            zone->citem_root->rr_sets.add(rrset);

            continue;
        }

        /* apparently a common rr */
        ret = CreateDomain(citem, zone, dname);
        if (!(ret & LOOKUP_FOUND) && !(ret & LOOKUP_CREATED)) {
        	domtostring(tmpb, dname);
        	log_print(LOG_LEVEL_ERROR, "%s:%d: Could not initialize domain %s", file.name, file.linenum, tmpb);
        	return -1;
        }

        if (rrinfo->type == DNS_TYPE_DNAME) {
            if (citem->subitems.count) {
                domtostring(tmpb, dname);
                log_print(LOG_LEVEL_ERROR, "%s:%d: DNAME domain %s may not contain subdomains!", file.name, file.linenum, tmpb);
                return -1;
            }
        }

        if (citem->get_rrset(DNS_TYPE_CNAME) && rrinfo->type != DNS_TYPE_DNAME) {
			domtostring(tmpb, dname);
			log_print(LOG_LEVEL_ERROR, "%s:%d: Domain %s is a CNAME; unable to add common RR!", file.name, file.linenum, tmpb);
			return -1;
        }
		
        rrset = citem->write_rrset(rrinfo->type);
        rrset->add_rr(!is_cache, ttl, pos, yadb);
    }
    file.close();
    return 0;
}

