/*
    Posadis - A DNS Server
    Main source file
    Copyright (C) 2001  Meilof Veeningen <meilof@users.sourceforge.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <include/unisock.h>
#include <include/unifile.h>
#include <include/unistring.h>

/* include the signal-handling functions */
#ifndef _WIN32
#include <unistd.h>
#endif
#include <stdlib.h>
#include <signal.h>

/* include sleep() functions and w32 dialog */
#include <stdlib.h>
#ifdef _WIN32
#include <dos.h>
#include "w32/w32dialog.h"
#endif

/* include setuid */
#ifndef _WIN32
#include <pwd.h>
#include <grp.h>
#include <sys/types.h>
#include <unistd.h>
#endif


#include "configuration.h"
#include "threads.h"
#include "getconfigdir.h"
#include "log.h"
#include "zones.h"

void cleanup(int);
void sigpipe(int);
void sighup(int);

long curtime; /* used all throughout the program */

int udpclientsock;
#ifdef HAVE_IPV6
int udp6clientsock = 0;
#endif

bool stop = false; /* set to true when exiting */
bool update_flag = false; /* set to true to upate */
#ifndef _WIN32
bool has_forked = false; /* when we have forked */
#endif

void update_zone(Zone *zone);

int quitpos(int retcode) {
    /* cleanup code */
    int t;
    
    for (t = 0; t < tcp_connections_in.count; t++)
        tcpclose(tcp_connections_in[t]->sockid);
    for (t = 0; t < tcp_connections_out.count; t++)
        tcpclose(tcp_connections_out[t]->sockid);
    for (t = 0; t < tcp_connections_srv.count; t++)
        tcpclose(tcp_connections_srv[t]->sockid);
        
    for (t = 0; t < udp_connections_in.count; t++)
        udpclose(udp_connections_in[t]->sockid);
    for (t = 0; t < udp_connections_out.count; t++)
        udpclose(udp_connections_out[t]->sockid);

    clean_threads();
    closesocklib();
    log_close();
    if (rootzone) delete rootzone;

#ifdef _WIN32
    if (retcode != 0) w32dlg_set_title("Posadis shutting down with error...");
#endif

    return retcode;
}

#ifdef _WIN32
/* for win32, we use the WinMain entry point function which calls this one */
int entrypoint(int argc, char *argv[]) {
#else
int main(int argc, char *argv[]) {
#endif
    int t, x, errcode;
    bool have_active_thread, is_data = false;
    char *ptr;

    signal(SIGTERM, cleanup); /* SIGTERM causes graceful close */
    signal(SIGINT, cleanup); /* CRTL-C causes clean-up code */
#ifndef _WIN32
    signal(SIGPIPE, sigpipe); /* catch SIGPIPE */
    signal(SIGHUP, sighup); /* catch SIGHUP (reload) */
#endif

    errcode = process_cmdline_args(argc, argv);
    if (errcode != 0) return quitpos(1);

    /* try 'n find the config dir */
    if (conf_configdir[0] == '\0' &&
        get_config_dir("conf", conf_configdir) != 0) {
        log_print(LOG_LEVEL_PANIC, "Could not find a valid configuration dir!");
        return quitpos(1);
    }

    if ((t = load_config_file("posadis.conf")) < 0) {
    	log_print(LOG_LEVEL_PANIC, "Loading %sposadis.conf failed!", conf_configdir);
    	return quitpos(10);
    }
    if ((t = load_config_file("posadis-domains.conf")) < 0) {
    	log_print(LOG_LEVEL_PANIC, "Loading %sposadis-domains.conf failed!", conf_configdir);
    	return quitpos(10);
    }

    if (determine_datadir() < 0) {
      log_print(LOG_LEVEL_PANIC, "No Posadis data directory could be determined!");
      return quitpos(10);
    }

    log_init();

    if ( errcode != 0) {
        if (errcode < 0) log_print(LOG_LEVEL_PANIC, "Error loading configuration: %d", errcode);
        return quitpos(1);
    }

    log_print(LOG_LEVEL_INFO, "Posadis initializing from %s...", conf_configdir);
    errcode = initsocklib();
    if (errcode < 0) {
        log_print(LOG_LEVEL_PANIC, "Could not initialize socket library: %d", errcode);
        return quitpos(2);
    }

    curtime = time(NULL);
    if ( (errcode=zones_init()) < 0) {
        log_print(LOG_LEVEL_PANIC, "Could not init zones!");
        return quitpos(3);
    }
    watchset_init();
    
    /* now listen to all tcp interfaces in tcp_listen */
    _addr *curaddr;

    x = multi_get_count(config.tcp_listen);
    for (t = 0; t < x; t++) {
        curaddr = (_addr *)multi_getptr(config.tcp_listen, t);
        errcode = tcpcreateserver(curaddr);
        if (errcode < 0) {
            ptr = addr_to_string(curaddr);
            log_print(LOG_LEVEL_PANIC, "Could not listen on tcp %s!", ptr);
            free(ptr);
            return quitpos(4);
        }
        tcp_connections_srv.add(new TcpConnection(errcode));
    }

    /* now listen to udp interfaces */
    x = multi_get_count(config.udp_listen);
    for (t = 0; t < x; t++) {
        curaddr = (_addr *)multi_getptr(config.udp_listen, t);
        errcode = udpcreateserver(curaddr);
        if (errcode < 0) {
            ptr = addr_to_string(curaddr);
            log_print(LOG_LEVEL_PANIC, "Could not listen on udp %s!", ptr);
            free(ptr);
            return quitpos(4);
        }
        udp_connections_in.add(new UdpConnection(errcode));
    }

#ifndef _WIN32
    /* yo! drop them root privileges! */
    struct passwd *pwinfo;
    uid_t theuid;

    if (config.user && config.user[0]) {
        pwinfo = getpwnam(config.user);
        if (pwinfo) {
            theuid = pwinfo->pw_uid;
            if (setuid(theuid) != 0) {
                log_print(LOG_LEVEL_PANIC, "Could not drop privileges to %s[%d]!", config.user, theuid);
                return quitpos(5);
            }
        } else {
            log_print(LOG_LEVEL_PANIC, "Can't seem to find user name %s!", config.user);
            return quitpos(5);
        }
    }

    struct group *grinfo;
    gid_t grid;

    if (config.group && config.group[0]) {
         grinfo = getgrnam(config.group);
         if (grinfo) {
             grid = grinfo->gr_gid;
             if (setgid(grid) != 0) {
                 log_print(LOG_LEVEL_PANIC, "Could not drop group privileges to %s[%d]!", config.group, grid);
                 return quitpos(5);
             }
         } else {
             log_print(LOG_LEVEL_PANIC, "Can't seem to find group name %s!", config.group);
             return quitpos(5);
         }
    }

#endif
    
    /* and the udp client(s) */
    _addr any;
    getaddress(&any, "0.0.0.0", 0);
    if ( (udpclientsock = udpcreateserver(&any)) < 0) {
    	log_print(LOG_LEVEL_PANIC, "Could not initialize UDP client socket: %d!", udpclientsock);
    	return quitpos(6);
    }

    udp_connections_out.add(new UdpConnection(udpclientsock));

#ifdef HAVE_IPV6
    if (config.ipv6_client) {
      getaddress(&any, "::", 0);
      if ((udp6clientsock = udpcreateserver(&any)) < 0) {
          log_print(LOG_LEVEL_PANIC, "Could not initialize UDP IPv6 client socket: %d!", udp6clientsock);
          return quitpos(6);
      }
      udp_connections_out.add(new UdpConnection(udp6clientsock));
    }
#endif
    
    if (!udp_connections_in.count && !tcp_connections_srv.count)
        log_print(LOG_LEVEL_WARNING, "Feelin' pretty useless: not listening to any ports!");

    /* now create some everlasting threads */
    poll_thread = new PollThread();
    threads.add((MainThread *)poll_thread);                     /* checks server connections for data */
    zones_thread = new ZonesThread();
    threads.add((MainThread *)zones_thread);                    /* checks which zones need updating */
    tcp_connections_thread = new TcpConnectionsThread();
    threads.add((MainThread *)tcp_connections_thread);			/* closes tcp connections in time */


#ifndef _WIN32
    /* if nessecary, we will now fork */
    if (do_fork) {
        log_print(LOG_LEVEL_INFO, "Posadis running in the background...");
    	t = fork();
    	if (t == -1) {
			log_print(LOG_LEVEL_ERROR, "Forking failed!");
		} else if (t == 0) {
			has_forked = true;
		} else {
			/* we're the parent process so just quit */
			return 0;
		}
	} else log_print(LOG_LEVEL_INFO, "Posadis running...");
#else
      log_print(LOG_LEVEL_INFO, "Posadis running...");
#endif

#ifdef _WIN32
    w32dlg_set_title("Posadis is running...");
    if (auto_minimize && w32dlg_get_nitems() == 2) w32dlg_minimize();
#else
    if (has_forked) no_stdout_log = true;
#endif

	
    while (!stop) {
        have_active_thread = false;
        // execute child threads
        for (t = 0; t < threads.count; t++) {
            curtime = time(NULL);
            switch (threads[t]->trstate) {
                case THREAD_WAITDATA: /* waiting for data */
                    if (is_data || (threads[t]->wakeup_time && curtime >= threads[t]->wakeup_time) ) {
                        if (threads[t]->execute() != TRET_OKBUSY) { threads.removeindex(t); t--; }
                        have_active_thread = true;
                    }
                    break;
                case THREAD_SLEEP: /* asleep */
                    if (curtime >= threads[t]->wakeup_time) {
                        threads[t]->trstate = THREAD_RUNNING;
                        if (threads[t]->execute() != TRET_OKBUSY) { threads.removeindex(t); t--; }
                        have_active_thread = true;
                    }
                    break;
                case THREAD_RUNNING:
                    if (threads[t]->execute() != TRET_OKBUSY) { threads.removeindex(t); t--; }
                    have_active_thread = true;
                    break;
            }
        }

        if (have_active_thread)
            is_data = watchset_check();
        else
            is_data = watchset_waitread(1);
		
		if (update_flag) {
			update_flag = false;
			log_print(LOG_LEVEL_INFO, "Reloading all zones...");
			update_zone(rootzone);
		}
    }
    /* garbage collection ------------------------------------------------------------------------ */
#ifdef _WIN32
    w32dlg_set_title("Posadis is closing down...");
#endif
    log_print(LOG_LEVEL_INFO, "Posadis exiting...");
    return quitpos(0);
}

void cleanup(int sigcode) {
    stop = true;
}

void sigpipe(int) {
	log_print(LOG_LEVEL_WARNING, "Caught the SIGFPE signal!");
}

void update_zone(Zone *zone) {
	int x;
	
	if (!zone) return;
	
    zone->start_updating();
	for (x = 0; x < zone->subzones.count; x++)
		update_zone(zone->subzones[x]);
}

void sighup(int) {
	update_flag = true;
}
