/*
    Posadis - A DNS Server
    Some basic threads
    Copyright (C) 2001  Meilof Veeningen <meilof@users.sourceforge.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "dns.h"
#include "include/unistring.h"

#include <stdlib.h>

/* basic thread functions */
Vector<MainThread*> threads;

int init_threads() {
    return 0;
}

int clean_threads() {
    threads.destroy();
    return 0;
}

/* the abstract Thread superclass */

Thread::Thread() { }
Thread::~Thread() { }
int Thread::execute() { return TRET_CLOSED; }

MainThread::MainThread() : Thread() { wakeup_time = 0; trstate = THREAD_RUNNING; }
MainThread::~MainThread() { }

/* MessageInThread: reads message */

MessageInThread::MessageInThread() {
	inbuff = NULL;
	inbufflen = 0;
}

MessageInThread::~MessageInThread() {
	free(inbuff);
}

MessageOutThread *MessageInThread::make_reply_thread(DnsMessage *a) {
	return NULL;
}


MessageInThreadTCP::MessageInThreadTCP(TcpConnection *_inconn) :
	MessageInThread() {
		
	inconn = _inconn;
	sockid = inconn->sockid;
	inconn->state = TCP_BUSY;
	msgleft = -2;
	msgread = 0;

    endtime = curtime + config.tcp_patience + 1;
}

MessageInThreadTCP::~MessageInThreadTCP() {
	inconn->starttime = curtime;
	inconn->state = TCP_RECEIVED;
}

int MessageInThreadTCP::execute() {
    int ret;

    if (curtime > endtime) return TRET_ERROR;
    if (sockid < 0) return TRET_ERROR;
    if (watchset_iserror(sockid)) return TRET_ERROR;
    if (!watchset_isdata(sockid)) return TRET_OKBUSY;

    if (msgleft > 0) { /* just reading data */
        ret = tcpread(sockid, &inbuff[msgread], msgleft);
        if (ret == 0) return TRET_ERROR;
        msgread += ret;
        msgleft -= ret;
    } else if (msgleft == -2) { /* we don't know the size yet */
    	/* set the person who we're receiving from */
    	gettcpaddr(sockid, &addr);

    	/* try to read the first two (length) chars */
        inbuff = (char *)malloc(2);
        inbufflen = tcpread(sockid, inbuff, 2);
        if (inbufflen == 0) { free(inbuff); return TRET_ERROR; }
        if (inbufflen == 2) {
            msgleft = inbuff[0]*256 + inbuff[1];
            free(inbuff);
            inbuff = (char *)malloc(msgleft);
            inbufflen = msgleft;
        } else {
            msgsize1 = inbuff[0];
            msgleft = -1;
            free(inbuff);
        }
        return TRET_OKBUSY;
    } else if (msgleft == -1) {
        inbuff = (char *)malloc(1);
        inbufflen = tcpread(sockid, inbuff, 1);
        if (inbufflen != 1) return TRET_ERROR;
        msgleft = msgsize1*256 + inbuff[0];
        free(inbuff);
        inbuff = (char *)malloc(msgleft);
    }
    if (msgleft == 0) return TRET_CLOSED;
    return TRET_OKBUSY;
}

MessageOutThread *MessageInThreadTCP::make_reply_thread(DnsMessage *a) {
	return new MessageOutThreadTCP(inconn, a);
}

MessageInThreadUDP::MessageInThreadUDP(int _sockid) :
	MessageInThread() {
	sockid = _sockid;
}

MessageInThreadUDP::~MessageInThreadUDP() { }

int MessageInThreadUDP::execute() {
	inbuff = (char *)malloc(UDP_MSG_SIZE);
	inbufflen = udpread(sockid, inbuff, &addr);
	if (inbufflen < 0) return TRET_ERROR;
	return TRET_CLOSED;
}

MessageOutThread *MessageInThreadUDP::make_reply_thread(DnsMessage *a) {
	return new MessageOutThreadUDP(sockid, addr, a);
}

void log_query(_addr *from, DnsMessage *q, DnsMessage *a) {
    char tmp[256];
    char *addr;

    addr = addr_to_string(from);
    domtostring(tmp, q->question[0]->QNAME);

    log_print(LOG_LEVEL_INFO, "Query: %s: Q=%d:%s:%d A=%d:%d",
              addr, q->OPCODE, tmp, q->question[0]->QTYPE, a->RCODE, a->answer.count);
    free(addr);
}


RequestInThread::RequestInThread(MessageInThread *tr) :
    MainThread() {
	question = new DnsMessage();
	answer = NULL;
	state = R_LOADING;
	child = NULL;
	intr = tr;
	outtr = NULL;
	trstate = THREAD_RUNNING;
	do_additional_proc = false;
	buff = NULL;
    queryinfo = NULL;
}

RequestInThread::~RequestInThread() {
	if (question) delete question;
	if (intr) delete intr;
	if (child) delete child;
	if (outtr) delete outtr;
    if (queryinfo) delete queryinfo;
	
	if (answer) delete answer;
	if (buff) free(buff);
}

int RequestInThread::execute() {
    int ret1, ret2;
    _thread_answer ans;
    AnsweringThread *tmp;
    stdquery_info *qinfo;

    if (state == R_LOADING) {
		if (wakeup_time == 0)
			wakeup_time = curtime + config.patience + 1;
        ret1 = intr->execute();
        if (ret1 == TRET_CLOSED) {

#ifdef _EXTDEBUG_
			char *tmp = addr_to_string(&intr->addr);
			log_print(LOG_LEVEL_INFO, "Ready receiving query from %s", tmp);
			free(tmp);
#endif
           	/* at first, see if we'll want to create another thread */
           	if (config.max_threads && threads.count >= config.max_threads) {
                log_print(LOG_LEVEL_INFO, "Had to reject query due to lack of threads");
           		return TRET_ERROR;
            }
            		
            ret2 = question->loadFromData(intr->inbuff, intr->inbufflen);
            if (ret2 >= 0) { /* we have enough data */
				if (question->TC) return TRET_ERROR;
#ifdef _EXTDEBUG_
				log_print(LOG_LEVEL_INFO, "Allright, query understood.", tmp);
#endif
                state = R_PROCINIT; return TRET_OKBUSY;
			} else {
	            /* apparently, the message is invalid */
                if (intr->inbufflen >= 2) {
                    answer = new DnsMessage();
                    answer->ID = intr->inbuff[0] * 256 + intr->inbuff[1];
                    answer->RCODE = RCODE_FORMERR;
                    answer->QR = true;
                    state = R_SENDING;
                    return TRET_OKBUSY;
                }
#ifdef _EXTDEBUG_
				log_print(LOG_LEVEL_INFO, "The message we got was malformed!", tmp);
#endif
    	        return TRET_CLOSED;
			}
        } else {
			if (curtime > wakeup_time) return TRET_ERROR;
        	return ret1;
        }
    }

    if (state == R_PROCINIT) {
        free(buff); /* we don't need that anymore */
        buff = NULL;

        answer = new DnsMessage();
        switch (question->OPCODE) {
        	case OPCODE_QUERY: /* standard query */
                if (question->question.count == 1 &&
                    question->question[0]->QTYPE == DNS_TYPE_SOA &&
                    question->answer.count == 1 &&
                    question->answer[0]->TYPE == DNS_TYPE_SOA) {
                    /* appears bind9 is trying to tell us something */
                    goto is_notify;
                }

       			if (question->question[0]->QTYPE == QTYPE_IXFR) {
       				/* incremental zone transfer */
       				ans = global_do_ixfr_thread(child, new xfr_info(question, answer, &intr->addr, this));
       			} else if (question->question[0]->QTYPE == QTYPE_AXFR) {
       				ans = global_do_axfr_thread(child, new xfr_info(question, answer, &intr->addr, this));
       			} else {
                    if (question->question.count != 1 ||
                        question->answer.count ||
                        question->authority.count ||
                        question->additional.count) {
               		    answer->answer_init(question);
               			answer->RCODE = RCODE_QUERYERROR;
               			state = R_SENDING;
               			return TRET_OKBUSY;
               		}

                    for (ret1 = 0; ret1 < current_std_queries.count; ret1++) {
                        qinfo = current_std_queries[ret1];
                        if (compare_host(qinfo->querier, &intr->addr) &&
                            domcmp(qinfo->query->question[0]->QNAME, question->question[0]->QNAME) &&
                            qinfo->QTYPE == question->question[0]->QTYPE)
                            /* same query */
                            return TRET_CLOSED;
                    }

    				answer->answer_init(question);
                    queryinfo = new stdquery_info(question, answer, &intr->addr, this, question->question[0]->QTYPE);
   					ans = global_do_lookup_thread(child, queryinfo, domdup(question->question[0]->QNAME));
   					do_additional_proc = true;
    			}
        		break;
is_notify:
            case OPCODE_NOTIFY: /* dns notify */
                ans = global_do_notify_thread(child, question, &intr->addr, answer, this);
                break;
        	default: /* non-supported OPCODE */
        		answer->answer_init(question);
        		answer->RCODE = RCODE_NOTIMPLEMENTED;
        		state = R_SENDING;
        		return TRET_OKBUSY;
        }
        switch (ans) {
        	case TA_ERROR: /* DNS query is invalid and should be ignored */
        		return TRET_ERROR;
        	case TA_SUCCESS: /* successfully answered */
        		state = R_SENDING;
        		return TRET_OKBUSY;
        	case TA_THREAD: /* thread created */
        		state = R_PROCESSING;
        		return TRET_OKBUSY;
        }
    }

    if (state == R_PROCESSING) {
        ret1 = child->execute();
        if (ret1 == TRET_ERROR)
        	return TRET_ERROR;
        if (ret1 == TRET_CLOSED) {
            /* check what the child said */
            if (child->result == TA_THREAD) {
               tmp = child->additional_thread;
               delete child;
               child = tmp;
            } else {
       	        state = R_SENDING;
            }
	   	}
        return TRET_OKBUSY;
    }

    if (state == R_SENDING) {
		if (!outtr) {
            /* log the query */
#ifndef _EXTDEBUG_
            if (config.do_query_logging && question && question->question.count)
#endif
                 log_query(&intr->addr, question, answer);

	    	/* at first, when nessecary, do additional section processing */
    		if (do_additional_proc) answer->do_additional_section_processing();
			outtr = intr->make_reply_thread(answer);
			if (!outtr) return TRET_ERROR;
		}
        return outtr->execute();
    }

    return TRET_OKBUSY;
}

AnswerInThread::AnswerInThread(MessageInThread *thr) :
    MainThread() {
    child = thr;
    state = A_LOADING;
}

AnswerInThread::~AnswerInThread() {
	if (child) delete child;
}

int AnswerInThread::execute() {
    int ret, t;
    ThreadWait *tw;
    char *tmp;

    if (state == A_LOADING) {
    	ret = child->execute();
    	if (ret != TRET_CLOSED) return ret;
    }

    /* if we get here, the data has been recieved. now check if some
       thread can use it */

    ret = child->inbuff[0] * 256 + child->inbuff[1]; /* the ID */

    for (t = 0; t < answer_waiting_threads.count; t++) {
        if (answer_waiting_threads[t]->ID == ret &&
            answer_waiting_threads[t]->sockid == child->sockid &&
            compare_hostport(&child->addr, &answer_waiting_threads[t]->from)) {
            /* we've found a match! */
            tw = answer_waiting_threads[t];
            if (tw->msg == NULL) tw->msg = new DnsMessage();
            if (tw->msg->loadFromData(child->inbuff, child->inbufflen) < 0) {
                delete tw->msg;
                tw->msg = NULL;
                answer_waiting_threads.removeitem(tw);
                return TRET_ERROR;
            }
            tw->received = true;
            tw->maintr->trstate = THREAD_RUNNING;
            return TRET_CLOSED;
        }
    }

    /* discard the whole message and quit */
    tmp = addr_to_string(&child->addr);
    log_print(LOG_LEVEL_WARNING, "Received unexpected answer [ID %d] from %s", ret, tmp);
    free(tmp);
    return TRET_ERROR;
}

MessageOutThread::MessageOutThread(char *_buff, int _bufflen) :
	Thread() {
	buff = _buff;
	bufflen = _bufflen;
}

MessageOutThread::MessageOutThread(AgBuffer *_buff) :
	Thread() {
	buff = _buff->buff;
	bufflen = _buff->bufflen;
	_buff->unreg();
	delete _buff;
}

MessageOutThread::~MessageOutThread() {
	free(buff);
}

int MessageOutThread::execute() { return TRET_ERROR; }

AgBuffer *compilemsg(DnsMessage *msg, int maxlen) {
	AgBuffer *ansbuff = new AgBuffer();
	int rval = msg->compile(ansbuff, maxlen);
	if (rval != 0) {
		free(ansbuff->buff);
		ansbuff->buff = NULL; ansbuff->bufflen = 0;
	}
	return ansbuff;
}

MessageOutThreadTCP::MessageOutThreadTCP(TcpConnection *_conn, DnsMessage *a) :
	MessageOutThread(compilemsg(a, -1)) {
	
	conn = _conn;
	sockid = conn->sockid;
	msgleft = 0;
    endtime = curtime + config.tcp_patience + 1;
}

MessageOutThreadTCP::MessageOutThreadTCP(TcpConnection *_conn, char *_buff, int _bufflen) :
	MessageOutThread(_buff, _bufflen) {
	
	conn = _conn;
	sockid = conn->sockid;
	msgleft = 0;
}

MessageOutThreadTCP::~MessageOutThreadTCP() {
	conn->state = TCP_SENT;
}

int MessageOutThreadTCP::execute() {
    char len[2];
    int ret;

    if (curtime > endtime) return TRET_ERROR;
    if (buff == NULL) return TRET_ERROR;

    /* check whether we may write */
    watchset_check();
    if (!watchset_canwrite(sockid)) return TRET_OKBUSY;

    if (msgleft == 0) {
        /* we've just begun */
        len[0] = bufflen / 256;
        len[1] = bufflen;

        ret = tcpsend(sockid, len, 2);

        if (ret <= 0) return TRET_ERROR;
        if (ret == 1) msgleft = -1;
        if (ret == 2) msgleft = bufflen;
    } else if (msgleft == -1) {
        /* only one size byte written: write the other one */
        ret = tcpsend(sockid, &len[1], 1);

        if (ret <= 0) return TRET_ERROR;
        if (ret == 1) msgleft = bufflen;
    } else if (msgleft > 0) {
    	if (watchset_iserror(sockid)) return TRET_ERROR;
        ret = tcpsend(sockid, &buff[bufflen-msgleft], msgleft);

        if (ret <= 0) return TRET_ERROR;
        msgleft -= ret;
    }

    if (msgleft == 0) return TRET_CLOSED; else return TRET_OKBUSY;
}

MessageOutThreadUDP::MessageOutThreadUDP(int _sockid, _addr _address, DnsMessage *a) :
	MessageOutThread(compilemsg(a, UDP_MSG_SIZE)) {
	sockid = _sockid;
	addr = _address;
    if (sockid == -1) {
         if (is_ipv4(&_address)) sockid = udpclientsock;
#ifdef HAVE_IPV6
         if (is_ipv6(&_address)) sockid = udp6clientsock;
#endif
    }
}

MessageOutThreadUDP::MessageOutThreadUDP(int _sockid, _addr _address, char *_buff, int _bufflen) :
	MessageOutThread(_buff, _bufflen) {
	sockid = _sockid;
	addr = _address;
    if (sockid == -1) {
         if (is_ipv4(&_address)) sockid = udpclientsock;
#ifdef HAVE_IPV6
         if (is_ipv6(&_address)) sockid = udp6clientsock;
#endif
    }
}

int MessageOutThreadUDP::execute() {
    if (sockid <= 0) return TRET_ERROR;
    if (!watchset_canwrite(sockid)) return TRET_OKBUSY;
    if (udpsend(sockid, &addr, buff, bufflen) != bufflen) return TRET_ERROR;
    return TRET_CLOSED;
}

AnsweringThread::AnsweringThread(stdquery_info *_info, _domain _QNAME) :
    Thread() {
    info = _info;
    QNAME = _QNAME;
    result = TA_SUCCESS;
    additional_thread = false;
}

AnsweringThread::~AnsweringThread() {
}

/* The PollThread checkes whether a connection is available on the net servers */
PollThread::PollThread() : MainThread() {
    trstate = THREAD_WAITDATA;
}

PollThread::~PollThread() {
}

int PollThread::execute() {
	int t, id;
	TcpConnection *conn;
	MessageInThread *msgtr;
	
	/* check udp server connections */
	for (t = udp_connections_in.count - 1; t >= 0; t--) {
	    if (watchset_isdata(udp_connections_in[t]->sockid)) {
#ifdef _EXTDEBUG_
			log_print(LOG_LEVEL_INFO, "New incoming UDP request on %d...", udp_connections_in[t]->sockid);
#endif
            threads.add(new RequestInThread(new MessageInThreadUDP(udp_connections_in[t]->sockid)));
        }
    }
    
    /* check udp client connections */
    
    for (t = udp_connections_out.count - 1; t >= 0; t--) {
        if (watchset_isdata(udp_connections_out[t]->sockid)) {
#ifdef _EXTDEBUG_
            log_print(LOG_LEVEL_INFO, "New incoming UDP answer on %d...", udp_connections_out[t]->sockid);
#endif
            threads.add(new AnswerInThread(new MessageInThreadUDP(udp_connections_out[t]->sockid)));
        }    
    }
    
    /* check tcp server connections */
    
    for (t = tcp_connections_srv.count - 1; t >= 0; t--) {
        if (watchset_isdata(tcp_connections_srv[t]->sockid)) {
            id = tcpaccept(tcp_connections_srv[t]->sockid);    	
           	if (id) {
#ifdef _EXTDEBUG_
			    log_print(LOG_LEVEL_INFO, "New incoming TCP request [%d]...", id);
#endif
        		conn = new TcpConnection(id);
        		gettcpaddr(id, &conn->addr);
        		conn->state = TCP_BUSY;
        		tcp_connections_in.add(conn);
        		msgtr = new MessageInThreadTCP(conn);
        		threads.add(new RequestInThread(msgtr));
            }
    	}
    }

    conn = NULL;

	/* check tcp in keepalive connections */
	for (t = tcp_connections_in.count - 1; t >= 0; t--) {
		if (tcp_connections_in[t]->state == TCP_SENT &&
		    watchset_isdata(tcp_connections_in[t]->sockid)) {
            if (tcp_is_open(tcp_connections_in[t]->sockid)) {
#ifdef _EXTDEBUG_
				log_print(LOG_LEVEL_INFO, "New incoming TCP keepalive message [%d]...", tcp_connections_in[t]->sockid);
#endif
                msgtr = new MessageInThreadTCP(tcp_connections_in[t]);
                threads.add(new RequestInThread(msgtr));
            } else {
                /* hung up */
#ifdef _EXTDEBUG_
				log_print(LOG_LEVEL_INFO, "TCP client hung up [%d]", tcp_connections_in[t]->sockid);
#endif
                tcp_connections_in.removeindex(t);
            }
		}
	}

	/* check tcp out connections */
	for (t = tcp_connections_out.count - 1; t >= 0; t--) {
		if (tcp_connections_out[t]->state == TCP_SENT &&
		    watchset_isdata(tcp_connections_out[t]->sockid)) {
#ifdef _EXTDEBUG_
			log_print(LOG_LEVEL_INFO, "Answer for outgoing TCP message received");
#endif
		    msgtr = new MessageInThreadTCP(tcp_connections_out[t]);
			threads.add(new AnswerInThread(msgtr));
		}
	}
	
    return TRET_OKBUSY;
}

PollThread *poll_thread; /* initialized by main, cleaned up by Vector::destroy */

/* tcp connection control */
TcpConnection::TcpConnection(int _sockid) {
	sockid = _sockid;
	watchset_add(sockid);
	starttime = curtime;
}

TcpConnection::~TcpConnection() {
	watchset_remove(sockid);
	tcpclose(sockid);
}

Vector<TcpConnection *> tcp_connections_in;
Vector<TcpConnection *> tcp_connections_out;
Vector<TcpConnection *> tcp_connections_srv;

/* udp connection vector */

UdpConnection::UdpConnection(int _sockid) {
    sockid = _sockid;
    watchset_add(sockid);
}

UdpConnection::~UdpConnection() {
    watchset_remove(sockid);
}

Vector<UdpConnection *> udp_connections_in;
Vector<UdpConnection *> udp_connections_out;


int TcpConnectionsThread::execute() {
	int t;
	int next_in = 0, x;
	
	/* check when the first one is to be closed */
	for (t = tcp_connections_in.count - 1; t >= 0; t--) {
		if (tcp_connections_in[t]->state == TCP_SENT) {
			x = tcp_connections_in[t]->starttime + config.tcp_in_keepalive;
			if (x <= curtime)
				tcp_connections_in.removeindex(t);
			else if (!next_in || (next_in + curtime) > x)
				next_in = x;
		}
	}
	
	for (t = tcp_connections_out.count - 1; t >= 0; t--) {
		if (tcp_connections_out[t]->state == TCP_RECEIVED) {
			x = tcp_connections_out[t]->starttime + config.tcp_out_keepalive;
			if (x <= curtime)
				tcp_connections_out.removeindex(t);
			else if (!next_in || (next_in + curtime) > x)
				next_in = x;
		}
	}
	
	trstate = THREAD_SLEEP;
	if (next_in == 0) {
		next_in = (config.tcp_in_keepalive<config.tcp_out_keepalive) ?
			config.tcp_in_keepalive : config.tcp_out_keepalive;
		next_in += curtime;
	}
			
	if (config.tcp_in_keepalive && next_in - curtime > config.tcp_in_keepalive)
		next_in = curtime + config.tcp_in_keepalive;
		
	if (config.tcp_out_keepalive && next_in - curtime > config.tcp_out_keepalive)
		next_in = curtime + config.tcp_out_keepalive;
	
	wakeup_time = next_in;
	
	return TRET_OKBUSY;
}


TcpConnectionsThread *tcp_connections_thread;

TcpConnection *initialize_tcp_connection(_addr addr) {
	int t, sockid;
	TcpConnection *newconn;
	
	for (t = 0; t < tcp_connections_out.count; t++) {
		if (compare_host(&addr, &tcp_connections_out[t]->addr) &&
			(tcp_connections_out[t]->state == TCP_RECEIVED))
			return tcp_connections_out[t];
	}
	
	sockid = tcpopen(&addr);
	if (sockid < 0) return NULL;
	newconn = new TcpConnection(sockid);
	newconn->addr = addr;
	newconn->state = TCP_BUSY;
	tcp_connections_out.add(newconn);
	return newconn;
}

int tcp_in_conn_release(TcpConnection *in) {
	if (config.tcp_in_keepalive == 0) {
		/* don't do any keepalive */
		tcp_connections_in.removeitem(in);
	} else {
		in->state = TCP_SENT;
		in->starttime = curtime;
	}
	return 0;
}

int tcp_out_conn_release(TcpConnection *out) {
	if (config.tcp_out_keepalive == 0) {
		/* don't do any keepalive */
		tcp_connections_out.removeitem(out);
	} else {
		out->state = TCP_RECEIVED;
		out->starttime = curtime;
	}
	return 0;
}

Vector<ThreadWait *> answer_waiting_threads;

ThreadWait::ThreadWait(MainThread *_maintr, int _ID, _addr _from, int _sockid) {
	maintr = _maintr;
	ID = _ID;
	from = _from;
	
	msg = NULL; received = false;
    sockid = _sockid;
}


ThreadWait::~ThreadWait() {
    if (msg) delete msg;
    release_id(ID);
}

ThreadWait *get_waitthread_info(Thread *tr) {
    int t;
    for (t = 0; t < answer_waiting_threads.count; t++)
        if (tr == answer_waiting_threads[t]->maintr) return answer_waiting_threads[t];

    return NULL;
}

ThreadWait *get_waitthread_info(int ID) {
    int t;
    for (t = 0; t < answer_waiting_threads.count; t++)
        if (ID == answer_waiting_threads[t]->ID) return answer_waiting_threads[t];

    return NULL;
}

void remove_from_waitthreads(Thread *tr) {
    int t;
    for (t = 0; t < answer_waiting_threads.count; t++)
        if (tr == answer_waiting_threads[t]->maintr) {
            answer_waiting_threads.removeindex(t, false);
        }
}

void remove_from_waitthreads(int ID) {
    int t;
    for (t = 0; t < answer_waiting_threads.count; t++)
        if (ID == answer_waiting_threads[t]->ID) answer_waiting_threads.removeindex(t, false);
}

void remove_from_waitthreads(ThreadWait *item) {
    answer_waiting_threads.removeitem(item);
}

#define randmask 16383
char ids[2048];
char bits[] = {128, 64, 32, 16, 8, 4, 2, 1};

#ifdef _WIN32
#define srandom srand
#define random rand
#else
char random_state[256];
#endif

u_short new_request_id() {
    static int done = 0;
    unsigned val, bit;

    if (done == 0) {
        srandom(curtime);
        memset(ids, 0, sizeof(ids));
#ifndef _WIN32
        initstate(curtime, random_state, sizeof(random_state));
        setstate(random_state);
#endif
    }

    bit = random() & 7;
    val = random() & randmask;
    do if (++val > sizeof(ids)) val = 0; while (val & bits[bit]);
    ids[val] |= bits[bit];
    return val * 8 + bit;
}


void release_id(u_short id) {
     ids[id / 8] &= ~bits[id & 7];
}
