/*
    Posadis - A DNS Server
    Some basic threads
    Copyright (C) 2001  Meilof Veeningen <meilof@users.sourceforge.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef __THREADS_H
#define __THREADS_H

enum thread_state {
    THREAD_RUNNING, /* the thread gets to run always */
    THREAD_WAITDATA, /* only if new data has arrived */
    THREAD_SLEEP /* only after wakeup_time */
};

/*!
 * The Thread class is the base class for many objects in Posadis that
 * perform a task. Independent base tasks that want to be part of the
 * main thread handling function need to be derived from the Thread
 * child MainThread.
 */

class Thread {                              // base class for all threads
  public:
    Thread();                               // constructor
    virtual ~Thread();                      // destructor
    virtual int execute();                  // the function executed when the thread may run, returning a TRET_ constant
};

class MainThread : public Thread {
  public:
    MainThread();
    virtual ~MainThread();

    thread_state trstate;                   // the internal thread state
    long wakeup_time;                       // if the thread has a THREAD_SLEEP state, this is the time the thread wakes up
};

class AnsweringThread;

#include "syssocket.h"
#include "classes.h"
#include "agb.h"
#include "vector.h"
#include "zones.h"

#define TRET_OKBUSY              0          // the thread is still busy
#define TRET_CLOSED              1          // error code telling the main process to destroy the thread
#define TRET_ERROR               2          // the thread got a fatal error and should be destroyed directly

// basic thread functions
class Thread;
extern Vector<MainThread *> threads;

int init_threads();
int clean_threads();

class AnsweringThread;
class MessageOutThread;
class TcpConnection;

/*!
  The MessageInThread is the superclass for all thread that handle incoming
  DNS messages. It contains the virtual execute function, overrided by its
  TCP and UDP descendants, which is meant to actually read the data.

  Note that the MessageInThread is created as a child thread of a
  RequestInThread or AnswerInThread, and never standalone.
*/

class MessageInThread : public Thread {
  public:
	MessageInThread();						// constructor for descendants
	virtual ~MessageInThread();				// .. and the desctructor

  	virtual MessageOutThread *make_reply_thread(DnsMessage *a);	// makes a new thread for replying

    char *inbuff;                           // the buffer which temporarily contains the query/answer
    										// note that the MessageInHandler
    int inbufflen;                          // the length of the query/answer buffer
    _addr addr;								// this will be set to the person sending the request
    int sockid;                             // the socket id
};

/*!
  MessageInThreadTCP, a subclass of MessageInThread, overrides execute
  and getmaxmsgsize in order to add TCP support.
*/

class MessageInThreadTCP : public MessageInThread {
  public:
    MessageInThreadTCP(TcpConnection *inconn); // constructor, taking the socket id of the server socket
    virtual ~MessageInThreadTCP();         // destructor

    virtual int execute();					// overrided
  	virtual MessageOutThread *make_reply_thread(DnsMessage *a);	// overrided

  private:
	TcpConnection *inconn;					// entry in tcp in listcd  	
	
    long msgleft;                           // the number of bytes left, or -2 or -1 if we don't know yet;
    long msgread;                           // the number of bytes read
    char msgsize1;                          // on the rare occasion that we do have the first length byte, but not the
                                            // second, the first one is stored in here, and msgsize = -1

    int endtime;
};

/*!
  MessageInHandlerUDP, a subclass of MessageInHandler, overrides load_data
  and getmaxmsgsize in order to add UDP supprt.
*/

class MessageInThreadUDP : public MessageInThread {
  public:
    MessageInThreadUDP(int _sockid); // constructor, taking the socket id of the udp socket
    virtual ~MessageInThreadUDP();         	// destructor

    virtual int execute();                	// overrided
  	virtual MessageOutThread *make_reply_thread(DnsMessage *a);	// makes a new thread for replying
};

/*!
  The RequestInThread is the basic thread that handles all incoming DNS queries.
  At first it read the message (using a MessageInHandler), and it interprets
  the query and passes it through to an appropriate OpThread.

  There are two flavors of the RequestInThread: RequestInThreadTCP and
  RequestInThreadUDP.
*/

enum RequestInThreadState {
    R_LOADING,                        /* loading data from the virtual load_data function */
    R_PROCINIT,                       /* ready loading; see which OpThread we should consult */
    R_PROCESSING,                     /* waiting for OpThread to finish */
    R_SENDING                         /* sending the answer of the OThread back */
};

class RequestInThread : public MainThread {
  public:
    RequestInThread(MessageInThread *tr);   // constructor
    virtual ~RequestInThread();             // destructor
    virtual int execute();                  // overrided function called by main
 protected:
 	char *buff; int bufflen;				// the incoming data
 	int maxbufflen;							// the max buff len

	RequestInThreadState state;             // the internal state

    DnsMessage *question, *answer;          // the dns question

    MessageInThread *intr;					// the thread for incoming requests
    AnsweringThread *child;                 // the Opcode Thread we ask for information
    MessageOutThread *outtr;				// the thread which sends the information back
    stdquery_info *queryinfo;               // query info

    bool do_additional_proc;				// whether to do additional section processing
};

/*!
  The AnswerInThread handles incoming DNS messages to the ports we send
  answers through. After recieving the message, it looks which thread is
  waiting for the message and passes it through.

  This thread is being overrided by AnswerInThreadUDP and AnswerInThreadTCP.
*/

enum AnswerInThreadState {
    A_NOTHING,                              // nothing particular
    A_LOADING,                              // loading data
};

class AnswerInThread : public MainThread {
  public:
    AnswerInThread(MessageInThread *thr);   // constructor
    virtual ~AnswerInThread();              // destructor

    virtual int execute();                  // overrided
  protected:
  	MessageInThread *child;					// child thread

 	AnswerInThreadState state;              // the internal state
};


/*!
  The MessageOutThread is a thread which is responsible for sending a DNS
  message.
*/

class MessageOutThread : public Thread {
  public:
  	MessageOutThread(char *buff, int bufflen); // constructor directly taking buffer
  	                                           // (which is free()'d by destructor)
  	MessageOutThread(AgBuffer *msg);		// alternate constructor
  	
  	virtual ~MessageOutThread();			// destructor

  	virtual int execute();					// overrided
  	
  	char *buff; int bufflen;				// the data we're sending
  	int sockid;								// the socket
};

/*!
  The MessageOutThreadTCP adds TCP support for sending DNS messages
*/

class MessageOutThreadTCP : public MessageOutThread {
  public:
  	MessageOutThreadTCP(TcpConnection *conn, DnsMessage *a); // constructor
  	MessageOutThreadTCP(TcpConnection *conn, char *buff, int bufflen); // alternate constructor
  	~MessageOutThreadTCP();					// destructor

  	virtual int execute();					// overrided
  protected:
  	TcpConnection *conn;					// the connection we're using
  	int msgleft;							// the number of bytes left
    int endtime;
};

/*!
  The MessageOutThreadUDP adds UDP support for sending DNS messages
*/

class MessageOutThreadUDP : public MessageOutThread {
  public:
  	MessageOutThreadUDP(int sock, _addr add, DnsMessage *a); // constructor
  	MessageOutThreadUDP(int sock, _addr a, char *buff, int bufflen); // alternate constructor

  	virtual int execute();					// overrided
  protected:
  	_addr addr;								// the address
};

/*!
 * Now for the Posadis Request Answering Architecture. This system is based
 * on threads called 'AnsweringThreads'. AnsweringThreads are being
 * initialized by a RequestInThread, or, in case of CNAME recursion of
 * a standard query, by another AnsweringThread. Each type of query
 * (recursive std query, non-recursive std query and ixfr/axfr, has its own
 * set of AnsweringThreads, one for each Zone type.
 */

class AnsweringThread : public Thread {
  public:
  	AnsweringThread(stdquery_info *_info, _domain QNAME);
  	virtual ~AnsweringThread();
  	
  	stdquery_info *info;
    _domain QNAME;

    _thread_answer result;
    AnsweringThread *additional_thread;
};

/*!
  The OpThreadNotSupported thread is a very simple subclass of OpThread. It
  handles all OPCODEs we don't support by sending back a message with
  RCODE_NOTSUPPORTED.
*/

class PollThread : public MainThread {          // class which checks the http, tcp and udp sockets
  public:
    PollThread();
    ~PollThread();
    virtual int execute();                  // overrided
};
extern PollThread *poll_thread;

/* active tcp connection controls */

enum _tcpconn_state { TCP_KEEP, TCP_BUSY, TCP_SENT, TCP_RECEIVED };

class TcpConnection {
  public:
	TcpConnection(int sockid);
	~TcpConnection();
	
	int sockid;
	_tcpconn_state state;
	_addr addr;
	int starttime;
};

extern Vector<TcpConnection *> tcp_connections_in;
extern Vector<TcpConnection *> tcp_connections_srv;
extern Vector<TcpConnection *> tcp_connections_out;

class UdpConnection {
  public:
    UdpConnection(int sockid);
    ~UdpConnection();
    
    int sockid;
};

extern Vector<UdpConnection *> udp_connections_in;
extern Vector<UdpConnection *> udp_connections_out;

class TcpConnectionsThread : public MainThread {
	virtual int execute();
};

extern TcpConnectionsThread *tcp_connections_thread;

/* looks whether an open tcp connection is available */
TcpConnection *initialize_tcp_connection(_addr addr);

int tcp_in_conn_release(TcpConnection *in);
int tcp_out_conn_release(TcpConnection *out);

/* definitions for sending requests */

enum _threadwait_flags {
    TWF_READMESSAGE = 1,                    // indicates the RequestThread should try to read the dns message
    TWF_ANSWERRECEIVED = 2,	    			// indicates the data for the request has been arrived
};

class ThreadWait { /* information on threads waiting for information */
  public:
	ThreadWait(MainThread *maintr, int ID, _addr from, int sockid);
	~ThreadWait();                          // does some destruction
	
	MainThread *maintr;						// the thread which contains status info
	int ID;									// the ID to wait for
	_addr from;								// the box to expect an answer from
    int sockid;                             // the socket id we want an answer from
	
	DnsMessage *msg;						// the answer object
	
	bool received;							// the answer has been received
};

extern Vector<ThreadWait *> answer_waiting_threads;

ThreadWait* get_waitthread_info(Thread *tr);
ThreadWait* get_waitthread_info(int ID);

void remove_from_waitthreads(Thread *tr);
void remove_from_waitthreads(int ID);
void remove_from_waitthreads(ThreadWait *item);

// to obtain a new question id, you should increase this
u_short new_request_id();
void release_id(u_short id);

#endif /* ifndef _THREADS_H */

