/*
    Posadis - A DNS Server
    Tunneling implementation
    Copyright (C) 2001  Meilof Veeningen <meilof@users.sourceforge.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <include/unistring.h>
#include <config.h>

#include "tunnel.h"
#include "zones.h"
#include "dnsconstants.h"
#include "cachelookup.h"
#include "rr.h"
#include "log.h"
#include "classes.h"
#include "posadis.h"
#include "syssocket.h"

/* The Tunnel zone group */

TunnelGroup::TunnelGroup() { type = ZONE_TUNNEL; }

/* The TunnelZone (nothing much frightning here...) */

TunnelZone::TunnelZone(_domain _dom) :
    HintZone(_dom) {
    type = ZONE_TUNNEL;
}

TunnelZone::~TunnelZone() {

}

void TunnelZone::start_updating() {
	int n_items, ourserver, x;
	
#ifdef _EXTDEBUG_
	char tmpb[256];
	domtostring(tmpb, domain);
	log_print(LOG_LEVEL_INFO, "Starting TUNNEL update process for %s", tmpb);
#endif
    
    next_update = curtime + 120; /* default value */
	
    if (!tunnelservers.count) {
       zone_is_loaded = false;
       tunnelservers.destroy();
	
        n_items = multi_get_count(tunnel_config.tunnelserver);
      	for (ourserver = 0; ourserver < n_items; ourserver++)
            addaddressbyident(&tunnelservers, multi_getptr(tunnel_config.tunnelserver, ourserver));

       	if (tunnelservers.count > 0) zone_is_loaded = true;
		
		if (multi_get_count(hint_config.source) == 0)
			for (x = 0; x < tunnelservers.count; x++)
                multi_add(&hint_config.source,
                    (intptr_t) memdup(tunnelservers[x], sizeof(_addr)));
    }

    HintZone::start_updating();
}

void TunnelZone::offer_glue(CacheItem *item) {
    CacheItem *tmp = citem_root, *tmp2;
    Zone *ZnTmp;
    rr_set *set, *set2;
    _addr tmpad;
    int x, ret;

    if (multi_get_count(tunnel_config.tunnelserver) == 0) {
        /* temporary hack */
        citem_root = item;

        set = citem_root->get_rrset(DNS_TYPE_NS);

        for (x = 0; x < set->items.count; x++) {
            ret = LookupDomain(tmp2, ZnTmp, set->items[x]->RDATA);
            if (ret & LOOKUP_FOUND) {
                set2 = tmp2->get_rrset(DNS_TYPE_A);
                if (set2) {
                    for (x = 0; x < set2->items.count; x++)
                        ret = getaddress_ip4(&tmpad, set2->items[x]->RDATA);
                        tunnelservers.add((_addr *)memdup(&tmpad, sizeof(tmpad)));
                }
#ifdef HAVE_IPV6
                set2 = tmp2->get_rrset(DNS_TYPE_AAAA);
                if (set2) {
                    for (x = 0; x < set2->items.count; x++)
                        ret = getaddress_ip6(&tmpad, set2->items[x]->RDATA);
                        tunnelservers.add((_addr *)memdup(&tmpad, sizeof(tmpad)));
                }
#endif
            }
        }

        citem_root = tmp;
    }

    HintZone::offer_glue(item);
}

_thread_answer TunnelZone::do_lookup_thread(AnsweringThread*& thr, stdquery_info *info, _domain QNAME) {
    if (info->query->RD == true) {
        if (tunnelservers.count == 0) {
            info->answer->RCODE = RCODE_SERVERFAILURE;
            return TA_SUCCESS;
        }
        thr = new TunnelThread(info, QNAME, this);
        return TA_THREAD;
    } else {
        return HintZone::do_lookup_thread(thr, info, QNAME);
    }
}

/* the tunnel thread */

TunnelThread::TunnelThread(stdquery_info *_info, _domain _QNAME, TunnelZone *_z) :
    AnsweringThread(_info, _QNAME) {
    zone = _z;
	attemptix = -1;
	state = T_NOTHING;
	child = NULL;
    info = _info;
}

TunnelThread::~TunnelThread() {
    if (child) delete child;
}

int TunnelThread::execute() {
 	_addr *addr;
 	int ret, x, ourserver = 0, id;
	char thisdomain[256];
    char *tmp = NULL, *tmpstr;

    DnsMessage *query = info->query, *answer = info->answer;

	if (state == T_SENDING) {
	    ret = child->execute();
	    switch (ret) {
	        case TRET_ERROR:
	            /* error sending data */
              tmpstr = addr_to_string(zone->tunnelservers[ourserver]);
              log_print(LOG_LEVEL_ERROR, "Failed to contact tunnel server %s", tmpstr);
              free(tmpstr);
              
	            delete child; child = NULL;
	            answer_waiting_threads.removeitem(twait);
	            break;
	        case TRET_CLOSED:
	            /* done sending */
	            delete child; child = NULL;
	            info->tr->trstate = THREAD_SLEEP;
	            info->tr->wakeup_time = endtime;
	            state = T_WAITING;
	            return TRET_OKBUSY;
	    }
	}

	if (state == T_WAITING) {
	    /* check whether an answer has arrived */
	    if (twait->received) {
	        /* we got an answer! */
	        answer->RCODE = twait->msg->RCODE;
	        for (x = twait->msg->answer.count - 1; x >= 0; x--) {
	            answer->answer.add(twait->msg->answer[x]);
	            twait->msg->answer.removeindex(x, false);
	        }	
	        for (x = twait->msg->authority.count - 1; x >= 0; x--) {
	            answer->authority.add(twait->msg->authority[x]);
	            twait->msg->authority.removeindex(x, false);
	        }
            for (x = twait->msg->additional.count - 1; x >= 0; x--) {
	            answer->additional.add(twait->msg->additional[x]);
	            twait->msg->additional.removeindex(x, false);
	        }

            if (info->is_first) {
                if (zone->tunnel_config.no_aa)
                    answer->AA = false;
                else
                    answer->AA = twait->msg->AA;
            }

            answer_waiting_threads.removeitem(twait);
            answer->ID = query->ID;
            return TRET_CLOSED;
	    } else {
#ifdef __DEBUG_
            return TRET_OKBUSY;
#else
	        if (curtime > endtime) {
              char *tmpstr = addr_to_string(zone->tunnelservers[ourserver]);
              log_print(LOG_LEVEL_ERROR, "No answer from tunnel server %s", tmpstr);
              free(tmpstr);

              answer_waiting_threads.removeitem(twait);
    	        info->tr->trstate = THREAD_RUNNING;
	        } else {
	            return TRET_OKBUSY;
	        }
#endif
	    }
	}

	domtostring(thisdomain, zone->domain);

  	if (attemptix < zone->tunnel_config.retry) {
  		/* do a new attempt */
  		attemptix++;
  		ourserver = attemptix % zone->tunnelservers.count;

        addr = zone->tunnelservers[ourserver];

  		/* now send the message (we only support UDP currently) */

  		/* make a buffer from the query, and replace the ID by our one.
  		   the rest of the message can be kept completely */
        id = new_request_id();
        AgBuffer *ansbuff = new AgBuffer();
        ret = query->compile(ansbuff, UDP_MSG_SIZE);
        if (ret != 0) return TRET_ERROR;
        ansbuff->buff[0] = id / 256;
        ansbuff->buff[1] = id;

		child = new MessageOutThreadUDP(-1, *addr, ansbuff->buff, ansbuff->bufflen);

        twait = new ThreadWait(info->tr, id, *addr, ((MessageOutThreadUDP*)child)->sockid);
        answer_waiting_threads.add(twait);

        ansbuff->unreg();
        delete ansbuff; ansbuff = NULL;

        endtime = curtime + config.patience + 1;

        state = T_SENDING;
        return TRET_OKBUSY;
  	} else {
  		/* we've had enough */
  	    switch (zone->tunnel_config.recursive_answer_fail) {
  	        case T_RF_NODATA:
  	            /* return NODATA */
  	            return TRET_CLOSED;
  	        case T_RF_NXDOMAIN:
  	            /* domain not found */
  	            answer->RCODE = RCODE_NXDOMAIN;
  	            return TRET_CLOSED;
            case T_RF_HINT:
                /* hint */
                zone->HintZone::do_lookup_thread((AnsweringThread*)tmp, info, info->query->question[0]->QNAME);
  	        case T_RF_NONE:
                return TRET_CLOSED;
  	            /* do not answer at all */
  	            return TRET_ERROR;
  	        default:
  	        	return TRET_CLOSED;
  	    }
  	}

  	return TRET_ERROR;
}
