/*
    Posadis - A DNS Server
    The information on zone we're authoritive for
    Copyright (C) 2001  Meilof Veeningen <meilof@users.sourceforge.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef __ZONES_H
#define __ZONES_H

class Zone;
class DnsMessage;
class stdquery_info;


/*! The _thread_answer enum is the return value of the varios do_*_thread
    functions of the Zone class. These functions can either return
    error, meaning the query should be ignored, success, meaning the
    query has been answered, or thread, meaning a thread has been initialized
    to answer the query (e.g. to send a request) */

enum _thread_answer {
    //! The DNS query should be ignored (e.g. it is invalid)
    TA_ERROR,
    //! The DNS query has been successfully answered, the answer can be sent
    TA_SUCCESS,
    //! A Thread has been initialized to answer the query
    TA_THREAD,
};


#include "cachelookup.h"
#include "dnstypes.h"
#include "configuration.h"
#include "vector.h"
#include "threads.h"

/*!
  We support a few different forms of zones:
   o authoritive zones
     primary (master) ->  physically holds zone information in a dns master
                          file, usually also provides xfr
     secondary (slave) -> on startup, gets a copy of the zone using XFR, may
                          optionally also offer XFR itself
   o non-authoritive zones
     hint ->            all the nameserver do when information on such as zone
                        is requested, is pointing the client to some nameserver
                        which might know more about the zone
     tunnel ->			pass through queries to another nameserver and copy the
     					response

  Each zone has its own zone type, which is basically just a (unique) bitmask of
  the configuration groups it supports. For example, a master zone supports
  general zone configuration, master configuration (for serving as an XFR
  provider), and slave configuration (for serving as an XFR client). This
  bitmasking concept makes it easy to determine whether a specific
  configuration setting can be applied to the zone.

  Additionally, authoritive zones have the ZONE_MASK_AUTH bit set, so that it
  can be easily checked whether a zone is authoritive.

  Note also that the order in which the various zone configuration settings
  occur in the zone bitmask (least to most significant), is the same as in
  which they appear in the group and zone classes.
*/

#define ZONE_MASK_AUTH 64
#define ZONE_MASK_ZONE 1
#define ZONE_MASK_PRIMARY 2
#define ZONE_MASK_SECONDARY 4
#define ZONE_MASK_HINT 16
#define ZOME_MASK_TUNNEL 32

enum _Zonetype {
    ZONE_PRIMARY = ZONE_MASK_ZONE | ZONE_MASK_AUTH | ZONE_MASK_PRIMARY,
    ZONE_SECONDARY = ZONE_MASK_ZONE | ZONE_MASK_AUTH | ZONE_MASK_SECONDARY,
    ZONE_HINT = ZONE_MASK_ZONE | ZONE_MASK_HINT,
    ZONE_TUNNEL = ZONE_MASK_ZONE | ZOME_MASK_TUNNEL,
    ZONE_AUTH = ZONE_MASK_ZONE | ZONE_MASK_AUTH,
};

class ZoneGroup { public: char *name; _Zonetype type; zone_configuration_data config; };

extern Vector<ZoneGroup *> ZoneGroups;

/* Zones */

class Zone;

int read_zone_line(posadis_infile& file);
int read_group_line(posadis_infile& file);

int zones_init();

extern Vector<stdquery_info *> current_std_queries;

class stdquery_info {
  public:
    stdquery_info(DnsMessage *_query, DnsMessage *_answer, _addr *_querier, MainThread* _tr, int _QTYPE);
    ~stdquery_info();

    DnsMessage *query, *answer;
    _addr *querier;
    MainThread *tr;
    int QTYPE;
    bool is_first;
};

class xfr_info {
  public:
    xfr_info(DnsMessage *q, DnsMessage *a, _addr *q, MainThread *tr);
    ~xfr_info();

    DnsMessage *query, *answer;
    _addr *querier;
    MainThread *tr;
};

/*!
  The base Zone class contains information on one zone which is handeled by
  Posadis. A zone in this context is any domain mentioned in the configuration
  files, which can be a normal authoritive zone (e.g. master, slave), or a
  special kind of zone (e.g. cache, tunneled).

  The base class primarily contains virtual functions which are overrided by
  the various descendants of the class, such as the AuthoritiveZone class,
  which contains some behavior for classes that always return authoritive
  information, such as primaries, masters or slaves.
*/

class Zone {
  protected:
    Zone(_domain);                                      // constructor to be called by descendants
  public:
    virtual ~Zone();

    _Zonetype type;                                     // the zone type
    _domain domain;                                     // the domain name of the zone
    CacheItem *citem_root;								// the root cacheitem of the zone
    Vector<Zone*> subzones;								// the sub zones

    zone_configuration_data zone_config;                // some basic zone settings

    bool is_being_updated;								// specifies whether the zone is currently being loaded
    bool zone_is_loaded;								// if set to false, queries to this zone will return a server failure
    long next_update;                                   // when this zone needs updating next (0 if never)
    virtual void start_updating();                      // the zone is urged to update its information
    virtual void offer_glue(CacheItem *item);           // offer glue
    Vector<MainThread*> update_notify_threads;          // threads that like to be notified when it's ready
	void notify_threads();								// notifies all threads in the update_notify_threads
	
	//! adds a domain to the zone
	/*! This function creates a domain as a child of the citem_root item.
	    If the domain name already exists, the function also returns
	    succesfully. Note that this function does not say anything about the
	    ability of the zone to be editable; it is used for example by the
	    master file reading routines.
	
	    This function is implemented in cachelookup.cpp
	
	    The created domain name (dom) will be put in res, and 0 will be
	    returned if the operation was succesful, or < 0 on error.
	
	*/
	virtual int create_domain(CacheItem*& res, _domain dom);

    //! perform an internal lookup, sending information only if available
    /*! This function is used for internal lookups on the zone (for example,
        to resolve the hintzone->nameserver property. This function will
        always return immediately (e.g. do not contact other nameservers or
        attempt to read master files), giving back whatever information is
        directly available.

        The default implementation of do_internal_lookup is to lookup using
        the citem_root CacheItem (though zones are not obliged to implement
        this). This function is implemented in cachelookup.cpp
    */
    virtual int do_internal_lookup(CacheItem*& res, _domain dom, int flags = 0);

    // the following functions have a default implementation which returns
    // a "not supported" error.


    // asks the client to set up a dns lookup
    /*! This function asks the Zone to set up a dns lookup. The second
        argument, q, is a DnsMessage containing the query sent by the client.
        The next argument, a, is the DnsMessage which will be sent back to
        the client.

        IMPORTANT: The QNAME is free'd by global_do_lookup_thread, so:
         - the caller should domdup it
         - zones who make threads shoud also domdup it
    */

    virtual _thread_answer do_lookup_thread(AnsweringThread*& thr, stdquery_info *info, _domain QNAME);

    virtual _thread_answer do_axfr_thread(AnsweringThread*& thr, xfr_info *info);
    virtual _thread_answer do_ixfr_thread(AnsweringThread*& thr, xfr_info *info);
    virtual _thread_answer do_notify_thread(AnsweringThread*& thr, DnsMessage *q, _addr *requestor, DnsMessage *a, MainThread *tr);

    virtual _thread_answer do_nonsupported_thread(AnsweringThread*& thr, DnsMessage *q, DnsMessage *a, MainThread *tr);

};

_thread_answer global_do_lookup_thread(AnsweringThread*& thr, stdquery_info *info, _domain QNAME);
_thread_answer global_do_axfr_thread(AnsweringThread*& thr, xfr_info *info);
_thread_answer global_do_ixfr_thread(AnsweringThread*& thr, xfr_info *info);
_thread_answer global_do_notify_thread(AnsweringThread*& thr, DnsMessage *q, _addr *requestor, DnsMessage *a, MainThread *tr);

/*!
  The NotauthZone class is a direct descendant of the Zone class, and it
  contains some default behavior for zones that aren't normal authoritive
  zones, for example, cache, hint and tunnel.

  The class contains some virtual functions wich can be overrided by
  descendants to implement specific behavior if a domain name or RRset
  wasn't found, and recursion was desired.
*/

class NotauthZone : public Zone {
  protected:
	NotauthZone(_domain);                               // constructor to be called by descendants
    virtual ~NotauthZone();
};

extern Zone *rootzone;

/* the zones thread takes care of reloading zones when nessecary */
/* it gets executed each minute by default */
class ZonesThread : public MainThread {
  public:
    ZonesThread();                          // constructor
    virtual int execute();                  // overrided
    void calc_next();                       // calculates when to update next

    long nexttime;                          // the next time we need to reload some zone
    Zone *zone;                             // the zone which we need to reload then
};
extern ZonesThread *zones_thread;
#endif /* __ZONES_H */

